#ifndef _OME_BASE_TYPE_H_
//! \cond
#define _OME_BASE_TYPE_H_ "$Id: OMEbaseType.h 312 2020-03-21 21:25:35Z geoff $"
//! \endcond

/*! \file
 * \brief OME fundamental data type implementation.
 */

/** \addtogroup OMEtypes
 * @{
 */

#include <OMEmanifests.h>

#define FORCE_OME_EXTENDED_TYPE 0

#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
#define OME_EXT_TYPE(member) value.extType->extValue. member
#else
#define OME_EXT_TYPE(member) value. member
#endif

#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
// OMEextendedType is used internally by OMEbaseType
/*! \brief Storage record for large-sized elements on 32-bit hardware.
 */
class OME_DLL_EXPORT OMEextendedType
{
public:
    union {
        double df;
        int64_t i64;
        uint64_t ui64;
    } extValue;

    explicit OMEextendedType(double f)
    {
        extValue.df = f;
    }

    explicit OMEextendedType(int64_t i)
    {
        extValue.i64 = i;
    }

    explicit OMEextendedType(uint64_t ui)
    {
        extValue.ui64 = ui;
    }

    ~OMEextendedType() {}
}; // end class OMEextendedType
#endif

/*! \brief Utility class for type check operations on OMEtype interactions.
 */
class OME_TypeCheck
{
public:
    enum TypeCompatibility_t {
        IDENTICAL_TYPES = (1 << 0), //!< same exact types
        CAN_ASSIGN = (1 << 1),      //!< can assign into type
        SAME_BASE_TYPE = (1 << 2),  //!< exact same basic type
        DISSIMILAR_SIGN = (1 << 3), //!< signed vs. unsigned
        CAN_PROMOTE = (1 << 4),     //!< small size convertable to larger
        CAN_DEMOTE = (1 << 5),      //!< large size to small
        INCOMPATIBLE_TYPES = (1 << 6),  //!< no conversion possible
        UNKNOWN_COMPATIBILITY = (1 << 7), //!< right side is ANY
        ASSUME_COMPATIBLE = CAN_ASSIGN | UNKNOWN_COMPATIBILITY, //!< left side is ANY

        // used by OIL2 compiler
        TYPE_MASK = 255,
        OPTIONAL_TYPE_FLAG = 256,
        OME_CONST_POINTER = OME_POINTER | 0x1000
    };

    static uint32_t baseType(uint32_t t) OME_CONST_FUNCTION {
        return (OMEfundamentalType(t));
    }

    static bool isTypeANY(uint32_t t) OME_CONST_FUNCTION {
        uint32_t t1 = t & TYPE_MASK;
        if ((t1 == OME_ANY) || (t1 == OME_POINTER) || (t1 == OME_CONST_POINTER)
                || (t & OPTIONAL_TYPE_FLAG))
        {
            return (true);
        }
        return (false);
    }

    static constexpr bool is32BitInteger(uint32_t t) OME_CONST_FUNCTION {
        return ((t == OME_UINT32) || (t == OME_INT32));
    }

    static constexpr bool isHeldBy32BitInteger(uint32_t t) OME_CONST_FUNCTION {
        return ((t == OME_UINT32) || (t == OME_INT32) ||
                (t == OME_UINT16) || (t == OME_UINT8));
    }

    static constexpr bool is64BitInteger(uint32_t t) OME_CONST_FUNCTION {
        return ((t == OME_UINT64) || (t == OME_INT64));
    }

    static constexpr bool isInteger(uint32_t t) OME_CONST_FUNCTION {
        return (isHeldBy32BitInteger(t) || is64BitInteger(t));
    }

    static constexpr bool isContainer(uint32_t t) OME_CONST_FUNCTION {
        return ((t == OME_ARRAY) | (t == OME_ASSOC) | (t == OME_SET));
    }

    // NOTE: order matches signature of VISTA OME routine OMEokToAssignTypes() for performance
    static uint32_t typesAreAssignable(uint32_t srcType, uint32_t targetType) OME_CONST_FUNCTION {
        if (targetType == srcType)   // identical types
        {
            // TODO: if const supported, drop CAN_ASSIGN
            return (IDENTICAL_TYPES | SAME_BASE_TYPE | CAN_ASSIGN);
        }
        // runtime evaluation if srcType == OME_ANY or OME_POINTER
        if (isTypeANY(targetType))   // assume it will work
        {
            return (ASSUME_COMPATIBLE | CAN_ASSIGN);
        }
        if (targetType == OME_NIL)   // not initialized
        {
            return (CAN_ASSIGN);
        }
        if (isTypeANY(srcType))   // we don't know...but assume OK
        {
            return (UNKNOWN_COMPATIBILITY | CAN_ASSIGN);
        }
        // type-specific checks
        // OID's can be set to nil...
        bool isSigned = false;
        switch (targetType)
        {
        case OME_NIL: // uninitialized
            return (CAN_ASSIGN);
        case OME_OID:
            if (srcType == OME_NIL) {
                //cerr << "oid assigned to nil\n";
                return (CAN_ASSIGN);
            }
            return (INCOMPATIBLE_TYPES);
        case OME_STRING:
            return (INCOMPATIBLE_TYPES);
        case OME_DOUBLE:
            switch (srcType) {
            case OME_FLOAT: // can promote to double
                return (CAN_PROMOTE | SAME_BASE_TYPE | CAN_ASSIGN);
            case OME_INT32:
            case OME_INT64:
            case OME_UINT64:
            case OME_UINT32:
            case OME_UINT16:
            case OME_UINT8:
                return (CAN_PROMOTE | CAN_ASSIGN);
            default:
                return (INCOMPATIBLE_TYPES);
            } // end switch srcType
            break;
        case OME_FLOAT:
            switch (srcType) {
            case OME_DOUBLE: // can promote to double
                return (CAN_DEMOTE | SAME_BASE_TYPE | CAN_ASSIGN);
            case OME_INT32:
            case OME_INT64:
            case OME_UINT64:
            case OME_UINT32:
            case OME_UINT16:
            case OME_UINT8:
                return (CAN_PROMOTE | CAN_ASSIGN);
            default:
                return (INCOMPATIBLE_TYPES);
            } // end switch srcType
            break;
        case OME_INT64:
            isSigned = true;
        // fallthrough intentionally
        case OME_UINT64:
            switch (srcType) {
            // NOTE: identical types have been handled above
            case OME_INT64:
            case OME_UINT64:
                return (SAME_BASE_TYPE | DISSIMILAR_SIGN | CAN_ASSIGN);
            case OME_INT32:
                if (isSigned == true) {
                    return (SAME_BASE_TYPE | CAN_PROMOTE | CAN_ASSIGN);
                }
                return (SAME_BASE_TYPE | CAN_PROMOTE | DISSIMILAR_SIGN | CAN_ASSIGN);
            case OME_UINT32:
            case OME_UINT16:
            case OME_UINT8:
                if (isSigned == false) {
                    return (SAME_BASE_TYPE | CAN_PROMOTE | CAN_ASSIGN);
                }
                return (SAME_BASE_TYPE | CAN_PROMOTE | DISSIMILAR_SIGN | CAN_ASSIGN);
            case OME_FLOAT:
            case OME_DOUBLE:
                // TODO:  add CAN_ASSIGN?
                return (CAN_DEMOTE);
            default:
                return (INCOMPATIBLE_TYPES);

            } // end switch (srcType)
            break;
        case OME_INT32:
            isSigned = true;
        // fallthrough intentionally
        case OME_UINT32:
            switch (srcType) {
            case OME_INT64:
                if (isSigned == true) {
                    return (CAN_DEMOTE | CAN_ASSIGN);
                }
                return (CAN_DEMOTE | DISSIMILAR_SIGN | CAN_ASSIGN);
            case OME_UINT64:
                if (isSigned == false) {
                    return (CAN_DEMOTE | CAN_ASSIGN);
                }
                return (CAN_DEMOTE | DISSIMILAR_SIGN | CAN_ASSIGN);
            case OME_INT32:
            case OME_UINT32:
                return (SAME_BASE_TYPE | DISSIMILAR_SIGN | CAN_ASSIGN);
            case OME_UINT16:
            case OME_UINT8:
                if (isSigned == false) {
                    return (CAN_PROMOTE | CAN_ASSIGN);
                }
                return (CAN_PROMOTE | DISSIMILAR_SIGN | CAN_ASSIGN);
            default:
                return (INCOMPATIBLE_TYPES);
            }  /// end switch srcType
            break;
        case OME_UINT16:
            // TODO
            break;
        case OME_UINT8:
            // TODO
            break;
        } // end switch targetType
        // NOTREACHED
        return (INCOMPATIBLE_TYPES);
    }
}; // end class OME_TypeCheck

/*! \brief Fundamental ANY type for FARGOS/VISTA Object Management Environment.
 */
class OME_DLL_EXPORT OMEbaseType
{
public:
    /*! \brief Return text identifying the indicated OME type.
     * \param t specifies the type.
     * \return
     * A constant text label such as "int32", "array", "assoc", etc.
     * can be returned.
     */
    static constexpr const char *typeName(const unsigned int t) OME_CONST_FUNCTION {
        // ensure type is not beyond type range
        // 32 = POWER_OF_2_TYPE_TABLE_SIZE in OMEtype.cpp
        return (_OMEtypeTable[t & (32 - 1)].typeName);

    }

    /*! \brief Return text identifying the OME enum name for the indicated type.
     * \param t specifies the type.
     * \return
     * A constant text label such as "OME_INT32", "OME_ARRAY", "OME_ASSOC", etc.
     * can be returned.
     */
    static constexpr const char *enumName(const unsigned int t) OME_CONST_FUNCTION {
        // ensure type is not beyond type range
        // 32 = POWER_OF_2_TYPE_TABLE_SIZE in OMEtype.cpp
        return (_OMEtypeTable[t & (32 - 1)].enumName);

    }

    static OMEbaseType *decode(const class OMEstring *data, size_t *offset,
                               const uint32_t version = 0);

    int encode(class OMEencodeBuffer *) const; // NOT A STATIC FUNCTION
public:
    union {
        int32_t i;
        uint32_t ui;
        float f;
#if (FORCE_OME_EXTENDED_TYPE == 0) && (__SIZEOF_POINTER__ == 8)
        int64_t     i64;
        uint64_t    ui64;
        double      df;
#endif
        class OMEextendedType *extType;
        class OMEarray *array;
        class OMEarray *arrayPtr; // alias since gdb cannot deal with array
        class OMEassoc *assoc;
        class OMEoid *oid;
        class OMEset *set;
        class OMEstring *s;
        class OMEfixed *fixed;
        class OMEnlm *nlm;
        OMEbaseType *pointer;
    } value;
    uint32_t type;

private:
    void freeComplexData();

    inline void freeData() OME_ALWAYS_INLINE
    {
        if (OMEtypeClass(type) != OME_SIMPLE_TYPE) {
            freeComplexData();
        }
    }

    void makeUnique();

public:
    // destructor
    ~OMEbaseType()
    {
        freeData();
    }

    //! default initializer
    OMEbaseType()
    {
        type = OME_NIL;
    }

    //! copy constructor
    OMEbaseType(const OMEbaseType &org) {
        type = OME_NIL;
        *this = org;
    }

    //! initialize as default value of specific type
    explicit OMEbaseType(const enum OMEtypes_t t)
    {
        type = OME_NIL;
        initializeAsType(t);
    }

    explicit OMEbaseType(int32_t arg)
    {
        type = OME_INT32;
        value.i = arg;
    }

    explicit OMEbaseType(uint32_t arg)
    {
        type = OME_UINT32;
        value.ui = arg;
    }

    explicit OMEbaseType(uint16_t arg)
    {
        type = OME_UINT16;
        value.ui = arg;
    }

    explicit OMEbaseType(uint8_t arg)
    {
        type = OME_UINT8;
        value.ui = arg;
    }

    explicit OMEbaseType(float arg)
    {
        type = OME_FLOAT;
        value.f = arg;
    }

    explicit OMEbaseType(uint64_t arg)
    {
        type = OME_UINT64;
#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
        value.extType = new OMEextendedType(arg);
#else
        value.ui64 = arg;
#endif
    }

    explicit OMEbaseType(int64_t arg)
    {
        type = OME_INT64;
#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
        value.extType = new OMEextendedType(arg);
#else
        value.i64 = arg;
#endif
    }

    explicit OMEbaseType(double arg)
    {
        type = OME_DOUBLE;
#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
        value.extType = new OMEextendedType(arg);
#else
        value.df = arg;
#endif
    }

    explicit OMEbaseType(const OMEfixed &arg)
    {
        type = OME_FIXED;
        value.fixed = new OMEfixed(arg);
    }

    //! from string...
    explicit OMEbaseType(const OMEstring &s)
    {
        type = OME_STRING;
        value.s = new OMEstring(s);
    }

    //! take ownership of string...
    explicit OMEbaseType(OMEstring *s)
    {
        type = OME_STRING;
        value.s = s;
    }

    //! from read-only string
    explicit OMEbaseType(OMEstringInROM s, int len = -1) {
        type = OME_STRING;
        value.s = new OMEstring(s, len);
    }

    //! from array
    explicit OMEbaseType(const OMEarray &a)
    {
        type = OME_ARRAY;
        value.array = new OMEarray(a);
    }

    //! from assoc
    explicit OMEbaseType(const OMEassoc &a)
    {
        type = OME_ASSOC;
        value.assoc = new OMEassoc(a);
    }

    //! from set
    explicit OMEbaseType(const OMEset &s)
    {
        type = OME_SET;
        value.set = new OMEset(s);
    }

    void initializeAsType(const enum OMEtypes_t t);

    // Reference Operators
#    ifdef OBSOLETE
    void addReadOnlyReference();
    void dropReference();
    OMEbaseType *deepCopy() const;
#    endif

    void getUniqueReference() OME_ALWAYS_INLINE
    {
        if (OMEtypeClass(type) != OME_REFERENCE_TYPE) {
            return;
        }
        makeUnique();
    }
    // END REFERENCE OPERATORS
    template <typename TO_TYPE> TO_TYPE convertToNumericType() const;

    // Casts
    explicit operator int32_t() const;

    explicit operator uint32_t() const;

    explicit operator int64_t() const;

    explicit operator uint64_t() const;

    explicit operator float() const;

    explicit operator double() const;

    explicit operator const char *() const {
        switch (type) {
        case OME_STRING:
            return ((const char *) (*value.s));
        default:
            return ("[NOTSTRING]");
        }
    }

    // simple assignment operators
    OME_FAST_CALL OMEbaseType &operator=(const int16_t i)
    {
        freeData();
        type = OME_INT32; // TODO: support int16_t
        value.i = static_cast<int32_t>(i);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const uint16_t i)
    {
        freeData();
        type = OME_UINT16;
        value.ui = static_cast<uint32_t>(i);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const int32_t i)
    {
        freeData();
        type = OME_INT32;
        value.i = i;
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const uint32_t i)
    {
        freeData();
        type = OME_UINT32;
        value.ui = i;
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const float f)
    {
        freeData();
        type = OME_FLOAT;
        value.f = f;
        return (*this);
    }

    // ASSIGNMENT
    // extended assignment operators
    OME_FAST_CALL OMEbaseType &operator=(const double f) {
        freeData();
        type = OME_DOUBLE;
#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
        value.extType = new OMEextendedType(f);
#else
        value.df = f;
#endif
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const int64_t i) {
        freeData();
        type = OME_INT64;
#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
        value.extType = new OMEextendedType(i);
#else
        value.i64 = i;
#endif
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const uint64_t ui) {
        freeData();
        type = OME_UINT64;
#if (FORCE_OME_EXTENDED_TYPE == 1) || (__SIZEOF_POINTER__ < 8)
        value.extType = new OMEextendedType(ui);
#else
        value.ui64 = ui;
#endif
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const OMEfixedConstant *f) {
        freeData();
        type = OME_FIXED;
        value.fixed = new OMEfixed(f->d);
        return (*this);
    }

    // complex assignment operators
    // strings
    OME_FAST_CALL OMEbaseType &operator=(const char c) {
        freeData();
        type = OME_STRING;
        value.s = new OMEstring(((const unsigned char *) &c), 1);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const unsigned char c) {
        freeData();
        type = OME_STRING;
        value.s = new OMEstring(&c, 1);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const char *s) {
        freeData();
        type = OME_STRING;
        value.s = new OMEstring(s);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEstringInROM s) {
        freeData();
        type = OME_STRING;
        value.s = new OMEstring(s);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(const OMEstring &s) {
        freeData();
        type = OME_STRING;
        value.s = new OMEstring(s);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEstring *s) {  // take ownership
        freeData();
        type = OME_STRING;
        value.s = s; // take ownership...
        return (*this);
    }

    // fixed
    OME_FAST_CALL OMEbaseType &operator=(const OMEfixed &f) {
        freeData();
        type = OME_FIXED;
        value.fixed = new OMEfixed(f);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEfixed *f) { // take ownership
        freeData();
        type = OME_FIXED;
        value.fixed = f; // take ownership
        return (*this);
    }

    // oid
    OME_FAST_CALL OMEbaseType &operator=(const OMEoid &o) {
        freeData();
        type = OME_OID;
        value.oid = new OMEoid(o);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEoid *o) { // take ownership
        freeData();
        type = OME_OID;
        value.oid = o; // take ownership...
        return (*this);
    }

    // array
    OME_FAST_CALL OMEbaseType &operator=(const OMEarray &a) {
        freeData();
        type = OME_ARRAY;
        value.array = new OMEarray(a);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEarray *a) { // take ownership...
        freeData();
        type = OME_ARRAY;
        value.array = a; // take ownership...
        return (*this);
    }

    // assoc
    OME_FAST_CALL OMEbaseType &operator=(const OMEassoc &a) {
        freeData();
        type = OME_ASSOC;
        value.assoc = new OMEassoc(a);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEassoc *a) { // take ownership...
        freeData();
        type = OME_ASSOC;
        value.assoc = a; // take ownership...
        return (*this);
    }

    // set
    OME_FAST_CALL OMEbaseType &operator=(const OMEset &s) {
        freeData();
        type = OME_SET;
        value.set = new OMEset(s);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEset *s) { // take ownership...
        freeData();
        type = OME_SET;
        value.set = s; // take ownership...
        return (*this);
    }

    // native language message
    OME_FAST_CALL OMEbaseType &operator=(const OMEnlm &m) {
        freeData();
        type = OME_NLM;
        value.nlm = new OMEnlm(m);
        return (*this);
    }

    OME_FAST_CALL OMEbaseType &operator=(OMEnlm *m) { // take ownership...
        freeData();
        type = OME_NLM;
        value.nlm = m; // take ownership...
        return (*this);
    }

    // assignment by same type
    OME_FAST_CALL OMEbaseType &operator=(const OMEbaseType &);

    OME_FAST_CALL OMEbaseType &operator=(OMEbaseType *d); // take ownership...

    OME_FAST_CALL bool operator==(const char *) const;

    OME_FAST_CALL bool operator!=(const char *) const;

#if 0 // OPERATORS
    // COMPARISON

    OME_FAST_CALL bool operator==(const OMEbaseType &) const;

    OME_FAST_CALL bool operator==(const int32_t) const;

    OME_FAST_CALL bool operator==(const uint32_t) const;

    OME_FAST_CALL bool operator==(const float) const;

    OME_FAST_CALL bool operator==(const bool) const;


    OME_FAST_CALL bool operator==(const OMEstring &) const;

    OME_FAST_CALL bool operator==(const OMEoid &) const;

    OME_FAST_CALL bool operator!=(const OMEbaseType &) const;

    OME_FAST_CALL bool operator!=(const int32_t) const;

    OME_FAST_CALL bool operator!=(const uint32_t) const;

    OME_FAST_CALL bool operator!=(const float) const;

    OME_FAST_CALL bool operator!=(const bool) const;

    OME_FAST_CALL bool operator!=(const OMEoid &) const;

    OME_FAST_CALL bool operator!=(const char *) const;

    OME_FAST_CALL bool operator!=(const OMEstring &) const;

    OME_FAST_CALL bool operator<(const OMEbaseType &) const;

    OME_FAST_CALL bool operator<(const int32_t) const;

    OME_FAST_CALL bool operator<(const uint32_t) const;

    OME_FAST_CALL bool operator<(const float) const;

    OME_FAST_CALL bool operator>(const OMEbaseType &) const;

    OME_FAST_CALL bool operator>(const uint32_t) const;

    OME_FAST_CALL bool operator>(const int32_t) const;

    OME_FAST_CALL bool operator<=(const OMEbaseType &) const;

    OME_FAST_CALL bool operator<=(const int32_t) const;

    OME_FAST_CALL bool operator<=(const uint32_t) const;

    OME_FAST_CALL bool operator<=(const float) const;

    OME_FAST_CALL bool operator>=(const OMEbaseType &) const;

    OME_FAST_CALL bool operator>=(const int32_t) const;

    OME_FAST_CALL bool operator>=(const uint32_t) const;

    OME_FAST_CALL bool operator>=(const float) const;

    // ARITHMETIC
#    include <OMEtypeHdr1.genh>


    // Integer-only operators
#    include <OMEtypeHdr2.genh>

#endif // OPERATORS

    // SUBSCRIPTING
    OME_FAST_CALL READ_ONLY_OMEtype operator[](const uint64_t) const;

    OME_FAST_CALL OMEbaseType &operator[](const uint64_t);

    OME_FAST_CALL READ_ONLY_OMEtype operator[](const uint32_t i) const
    {
        uint64_t ui = static_cast<uint64_t>(i);
        READ_ONLY_OMEtype result = (*this)[ui];
        return (result);
    }

    OME_FAST_CALL READ_ONLY_OMEtype operator[](const int64_t i) const
    {
        uint64_t ui = static_cast<uint64_t>(i);
        READ_ONLY_OMEtype result = (*this)[ui];
        return (result);
    }

    OME_FAST_CALL READ_ONLY_OMEtype operator[](const int32_t i) const
    {
        uint64_t ui = static_cast<uint64_t>(i);
        READ_ONLY_OMEtype result = (*this)[ui];
        return (result);
    }

    OME_FAST_CALL OMEbaseType &operator[](const uint32_t i)
    {
        uint64_t ui = static_cast<uint64_t>(i);
        OMEbaseType &result = (*this)[ui];
        return (result);
    }

    OME_FAST_CALL OMEbaseType &operator[](const int64_t i)
    {
        uint64_t ui = static_cast<uint64_t>(i);
        OMEbaseType &result = (*this)[ui];
        return (result);
    }

    OME_FAST_CALL OMEbaseType &operator[](const int32_t i)
    {
        uint64_t ui = static_cast<uint64_t>(i);
        OMEbaseType &result = (*this)[ui];
        return (result);
    }

    // associative subscript with read-only access
    OME_FAST_CALL READ_ONLY_OMEtype operator[](const OMEbaseType &) const;

    OME_FAST_CALL READ_ONLY_OMEtype operator[](const OMEstring &) const;

    OME_FAST_CALL READ_ONLY_OMEtype operator[](const char *k) const NONNULL_CLASS_PARAMETERS(2)
    {
        OMEstring s(reinterpret_cast<OMEstringInROM>(k));

        READ_ONLY_OMEtype result = (*this)[s];
        return (result);
    }

    OME_FAST_CALL OMEbaseType &operator[](const OMEbaseType &);

    OME_FAST_CALL OMEbaseType &operator[](const OMEstring &);

    OME_FAST_CALL OMEbaseType &operator[](const char *k) NONNULL_CLASS_PARAMETERS(2)
    {
        OMEstring s(k);

        OMEbaseType &result = (*this)[s];
        return (result);
    }

    // simple chained operators...

    // output functions
    enum OMEtype_OutputFlags_t : uint8_t {
        OUTPUT_TYPENAME = (1 << 0),
        DISABLE_QUOTE_OUTPUT = (1 << 1),
        OUTPUT_NEWLINE = (1 << 2)
    };
    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputOnStream,
            int_fast16_t indent = 0, uint8_t includeTypePrefix = OUTPUT_TYPENAME) const;

    const char *typeName() const NONNULL_RETURN 
    {
        return (OMEbaseType::typeName(type));
    }
}; // end class OMEbaseType

#    ifndef DOXYGEN_ONLY
#    define OMEtype OMEbaseType
#    endif

template <typename TO_TYPE> inline TO_TYPE OMEbaseType::convertToNumericType() const
{
    TO_TYPE result;

    switch (type) {
    case OME_INT32:
        result = static_cast<TO_TYPE>(value.i);
        break;
    case OME_UINT32:
    case OME_UINT16:
    case OME_UINT8:
        result = static_cast<TO_TYPE>(value.ui);
        break;
    case OME_INT64:
        result = static_cast<TO_TYPE>(OME_EXT_TYPE(i64));
        break;
    case OME_UINT64:
        result = static_cast<TO_TYPE>(OME_EXT_TYPE(ui64));
        break;
    case OME_FLOAT:
        result = static_cast<TO_TYPE>(value.f);
        break;
    case OME_DOUBLE:
        result = static_cast<TO_TYPE>(OME_EXT_TYPE(df));
        break;
    default:
        result = static_cast<TO_TYPE>(0);
        break;
    }
    return (result);
}

inline OMEbaseType::operator int32_t() const
{
    return (convertToNumericType<int32_t>());
}

inline OMEbaseType::operator uint32_t() const
{
    return (convertToNumericType<uint32_t>());
}

inline OMEbaseType::operator int64_t() const
{
    return (convertToNumericType<int64_t>());
}

inline OMEbaseType::operator uint64_t() const
{
    return (convertToNumericType<uint64_t>());
}

inline OMEbaseType::operator float() const
{
    return (convertToNumericType<float>());
}

inline OMEbaseType::operator double() const
{
    return (convertToNumericType<double>());
}

#include <OMEtype_operators.hpp>

/** @}
 */

#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
