#ifndef _OME_CLASS_H
//! \cond
#define _OME_CLASS_H "$Id: OMEclass.h 347 2020-05-21 04:13:52Z geoff $"
//! \endcond

#include <map>

#include <OMEmethod.h>
#include <OMEtype.h>
#include <utils/base/atomic_values.h>

/** \defgroup OMEclass OME Class Description
 * @{
 */

typedef OME_CORE_CALL void *(*CreateInstanceFP)(const class OMEclass *cInfo);
typedef OME_CORE_CALL void (*DeleteInstanceFP)(const class OMEclass *cInfo, void *data);

OME_CORE_CALL void *OMEdefaultInstanceCreate(const class OMEclass *cInfo);
OME_CORE_CALL void OMEdefaultInstanceDelete(const class OMEclass *cInfo, void *data);



class OMEclass {
private:
	class OMEnameSpaceGroup	*nameSpaceInfo;
	class OMEinheritEntry		*inheritList;
	std::map<uint32_t, OMEmethodP>	methodTable;	// defined methods
	bool			resolvedFlag;	// all classes resolved
	bool			forceInheritUnique;
	uint32_t			instanceCount;
	// inheritance graph after processing...
	uint_fast16_t classInstanceTotal;
	class OMEinheritGraphNode **classInstances;
	uint8_t      		*searchOrderLists;
	// storage creation and deletion routines
	CreateInstanceFP	storageCreateRoutine;
	DeleteInstanceFP	storageDeleteRoutine;
	// Encode, decode functions...
	mutable struct OMEinstanceVarDescr const *varDescRec;
	mutable OMEarray	variableDescriptions;

	bool inheritsFrom(const OMEclass *aClass,
		const OMEstring *baseClassName = nullptr,
		const uint32_t *baseClassVerID = nullptr) const;

	void buildInheritanceGraph(uint_fast16_t *count, class OMEinheritGraphNode *l[]);

	uint8_t computeSearchOrder(uint8_t searchList[],
        const int_fast16_t startLevel = -1,
        const uint8_t currentCount = 0) const;

	int addMethodName(OMEstring &name, OMEmethodP methodRecord);
public:
	OMEstring		className;
	uint32_t			versionID;
	uint32_t			hashIndex;	//!< OMEhash(className)
	uint32_t			instanceSize;	//!< required storage in bytes
	OMEoid			*metaObject;

	OMEclass(OMEnameSpaceGroup *nameSpace, const OMEstring &name,
		const uint32_t verID, const uint32_t iSize,
		const bool unique = 0,
		const CreateInstanceFP csRoutine = OMEdefaultInstanceCreate,
		const DeleteInstanceFP dsRoutine = OMEdefaultInstanceDelete);

	~OMEclass();

	void setMetaObject(const OMEoid &obj) {
		delete metaObject;
		metaObject = new OMEoid(obj);
	}

	void setMetaObject(const OMEtype &obj) {
		delete metaObject;
		if (obj.type == OME_OID) {
			metaObject = new OMEoid(*obj.value.oid);
		} else {
			metaObject = nullptr;
		}
	}

	int addMethod(OMEmethodP); //!< Add method implementation.

	OME_FAST_CALL OMEmethodP findMethod(const OMEstring &methodName,
		const uint32_t indx=0) const;

	OME_FAST_CALL OMEinvocationData *findMethodImplementation(const uint32_t signature,
		const OMEstring &methodName, const uint32_t indx = 0) const;

	OME_FAST_CALL int listMethodInfo(OMEassoc &result) const;

	int setStorageDescription(const struct OMEinstanceVarDescr *desc);

	uint_fast16_t setStorageDescription(const OMEarray &oil2Data);

	int inheritFromClass(const OMEstring &nameSpace, const OMEstring &name,
		const uint32_t verID = 0, const bool unique = false);

    int inheritFromClass(const char *nameSpace, const char *name,
		const uint32_t verID = 0, const bool unique = false)
    {
        return (inheritFromClass(OMEstring(nameSpace), OMEstring(name),
                    verID, unique));
    }

	bool resolved() const { return (resolvedFlag); }

	bool permitInheritOfUniqueCopy() const { return (forceInheritUnique); }

	uint_fast16_t inheritListTotal() const { return (classInstanceTotal); }

	OME_FAST_CALL OMEclass *getClassAtLevel(uint_fast16_t level) const;

	OME_FAST_CALL const unsigned char *getSearchOrder(uint8_t *count,
		const int_fast16_t startLevel = -1) const NONNULL_CLASS_PARAMETERS(2);

	uint_fast32_t getNextInstanceID() {
        uint32_t result;
        result = atomicIncrement_uint32(&instanceCount, 1);
        return (result);
	}

	const OMEarray &getVariableDescriptions() const;

	const OMEstring &getNameSpaceName() const;

	bool doesInheritFrom(const OMEclass *aClass, 
		OMEstring *fromClassName = nullptr, uint32_t *fromClassVerID = nullptr) const;

	int resolveLinkages(OMEstring *missingClassName = nullptr,
		uint32_t *missingClassVersionID = nullptr);

	void *createInstanceData();

	void deleteInstanceData(void *iData);

	OME_FAST_CALL OMEtype *extractInstanceData(const void *iData, bool annotate) const;

	OME_FAST_CALL int importInstanceData(void *iData, const OMEtype &data) const;
	
	template <typename STREAMTYPE> STREAMTYPE & dumpClassData(STREAMTYPE &o = std::cout) const;
}; // end class OMEclass

typedef OMEclass *OMEclassP;

extern uint_fast16_t OMEconvertToStorageDescription(OMEarray &varDesc,
	const struct OMEinstanceVarDescr *desc);

extern uint_fast16_t OMEconvertToStorageDescription(OMEarray &varDesc, const OMEarray &data);

/** @}
 */

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
