#ifndef _OME_ENCODE_H
//! \cond
#define _OME_ENCODE_H "$Id: OMEencode.h 312 2020-03-21 21:25:35Z geoff $"
//! \endcond

#    include <OMEstring.h>

/*! \file
 * \brief OME type encoding routines.
 */

/** \addtogroup OMEtypes
 * @{
 */

/*! \brief Holds data for a single encoded OMEtype element.
 * Multiple OMEencodeBufferElement objects are linked together
 * to represent sequences of data.
 */
class OME_DLL_EXPORT OMEencodeBufferElement
{
    friend class OMEencodeBuffer;
protected:
    OMEencodeBufferElement *next;
    OMEstring *data;
    bool ownData;

public:
    OMEencodeBufferElement(OMEstring *d, bool own = true)
    {
        next = nullptr;
        data = own ? d : new OMEstring(*d);
        ownData = own;
    }

    ~OMEencodeBufferElement()
    {
        delete data;
    }

    size_t elementSize() const
    {
        return (data->length());
    }

    const OMEstring &getData() const
    {
        return (*data);
    }
}; // end class OMEencodeBufferElement

/*! \brief Buffer into which OMEtype data is encoded.
 */
class OME_DLL_EXPORT OMEencodeBuffer
{
protected:
    OMEencodeBufferElement *start;
    OMEencodeBufferElement *last;
    size_t totalSize;
    uint32_t encodeVersion;

public:
    OMEencodeBuffer(uint32_t version = 1)
    {
        start = nullptr;
        last = nullptr;
        totalSize = 0;
        encodeVersion = version;
    }

    // convenience function, create and initialize with contents of string
    OMEencodeBuffer(OMEstring *data, uint32_t version = 1)
    {
        OMEencodeBufferElement *elem =
            new OMEencodeBufferElement(data, 0); // don't own it
        start = elem;
        last = elem;
        totalSize = elem->elementSize();
        encodeVersion = version;
    }

    ~OMEencodeBuffer();

    void addEncodedElement(OMEencodeBufferElement *elem)
    {
        if (last != nullptr) {
            last->next = elem;
        }
        else {
            start = elem;
        }
        last = elem;
        totalSize += elem->elementSize();
    }

    uint32_t getEncodeVersion() const
    {
        return (encodeVersion);
    }

    size_t bufferSize() const
    {
        return (totalSize);
    }

    OMEstring *condenseIntoString(bool includeVersionID);
}; // end class OMEencodeBuffer

typedef int (*OMEencodeRoutineFP)(OMEencodeBuffer *, const class OMEtype &);

typedef OMEtype *(*OMEdecodeRoutineFP)(uint32_t version,
                                       const OMEstring *encodedData, size_t *offset);


/*! \brief Description for OME encoding routines for a specific
 * encoding version.
 */
struct OMEencodeRoutines {
    OMEencodeRoutines *next;
    uint32_t encodeVersion;
    OMEencodeRoutineFP encodeFunctions[OME_LAST_TYPE_INDEX];
    OMEdecodeRoutineFP decodeFunctions[OME_LAST_TYPE_INDEX];
};

extern OME_DLL_EXPORT void OMEdefineEncodeRoutinesForVersion(OMEencodeRoutines *routines);

extern OME_DLL_EXPORT void OMEloadVersion1Encodings();

extern OME_DLL_EXPORT uint_fast16_t OMElistDefinedEncodingVersions(class OMEarray &retList);

extern OME_DLL_EXPORT OMEstring *OMEcompressString(const OMEstring &data);

extern OME_DLL_EXPORT OMEstring *OMEuncompressString(const OMEstring &data);

extern OME_DLL_EXPORT OMEstring *OMEgzipString(const OMEstring &data);

extern OME_DLL_EXPORT OMEstring *OMEgunzipString(const OMEstring &data);

/** @}
 */

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
