#ifndef _OME_EVENT_H
//! \cond
#define _OME_EVENT_H "$Id: OMEevent.h 345 2020-05-21 01:03:07Z geoff $"
//! \endcond

#include <OMEthread.h>

/** \defgroup OMEevent OME Event Interfaces
 * @{
 */

#ifdef _WIN32
typedef int (*OMEwin32MessageLoopFP)();
#endif

typedef bool (*OMEeventCallbackFP)(class OMEevent *, OMEthread *);

/*!
\brief Base class for events within the FARGOS/VISTA
Object Management Environment.
*/
class OMEevent {
public:
	enum OMEeventType {
		timerExpired,	//!< elapsed time timer
		ioPossible,	//!< data available/can be written
		childDeath,	//!< child process terminated
		win32Event	//!< Microsoft Windows-specific event
	};
	uint64_t			eventHandle;	//!< Event-specific handle
	OMEeventCallbackFP	callback;	//!< Function to service event
	OMEthread		*waitingThread;	//!< OMEthread to signal
	void			*extraData;	//!< Any event-specific data
	OMEeventType		eventType;	//!< Fundamental event type
private:
	bool			posted;

public:
	OMEevent(OMEeventType t, uint32_t h, OMEeventCallbackFP func,
		OMEthread *thread = nullptr, void *extra = nullptr) {
		eventType = t;
		eventHandle = h;
		callback = func;
		waitingThread = thread;
		extraData = extra;
		posted = false;
	}

	virtual ~OMEevent()	{
	}

	//! Event equality
	bool operator==(const OMEevent &arg) const {
		if (eventHandle != arg.eventHandle) return (false);
		if (eventType != arg.eventType) return (false);
		return (true);
	}
	//! Event inequality
	bool operator!=(const OMEevent &arg) const {
		if (eventHandle != arg.eventHandle) return (true);
		if (eventType != arg.eventType) return (true);
		return (false);
	}
	//! Event ordering
	bool operator<(const OMEevent &arg) const {
		if (eventHandle > arg.eventHandle) return (false);
		if (eventHandle < arg.eventHandle) return (true);
		// eventHandles must be equal...
		if (eventType < arg.eventType) return (true);
		return (false);
	}

	/*!
	\brief Post event and call callback function.
	\retval 0 Event was handled and event should now be deleted.
	\retval 1 Event record should not be deleted.
	*/
	bool postEvent() {
		posted = true;
		bool rc = (*callback)(this, waitingThread);
		return (rc);
	}
	/*!
	\brief Check if event has been posted.
	\retval 0 Event has not yet been posted.
	\retval 1 Event has been posted.
	*/
	bool isPosted() const { return (posted); }
}; // end class OMEevent

typedef OMEevent *OMEeventP;

/*
class OMEwin32Event : public OMEevent {
public:
	OMEwin32Event(uint32_t h, OMEeventCallbackFP func);
	~OMEwin32Event()	{}
};
*/

void OMEaddEventToQueue(OMEevent *);

int OMEclearEventQueue(OMEevent::OMEeventType t);

void OMEpostEvent(OMEevent::OMEeventType t, uint64_t handle, uint64_t result = 0);

time_t OMEpostTimerEvent(time_t currentTime = 0);

time_t OMEnextEventTime(time_t currentTime = 0);	// probably not needed

/** @}
 */

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
