#ifndef _OME_MANIFESTS_H
//! \cond
#define _OME_MANIFESTS_H "$Id: OMEmanifests.h 312 2020-03-21 21:25:35Z geoff $"
//! \endcond

#    include <stdint.h>
#    include <utils/base/compiler_hints.h>

/*! \file
 * \brief OME constants and typedefs.
 */

#if __SIZEOF_POINTER__ == 4
#define OME_size_t uint32_t
#define OME_ssize_t int32_t
#else
#define OME_size_t uint64_t
#define OME_ssize_t int64_t
#endif

#    ifdef __GNUC__
/* GNU C++ Compiler */

/*
 GNU C/C++ includes bit-sized typedefs for most common value ranges.
 These should be safer to use than short, int and long
 as we can't be sure of the absolute sizes of these on all platforms
    (e.g. does sizeof(int)==32 or 64 on 64-bit systems?)
*/
#    include <sys/types.h>

#    if (__linux__ == 1) || (__OpenBSD__ == 1)
typedef int16_t int16;
typedef u_int16_t uint16; /* non-standard */
typedef int32_t int32;
typedef u_int32_t uint32; /* non-standard */
typedef int64_t int64;
#    endif

#    if __sun__ == 1 && __SVR4 == 1
// Sun Solaris uses ISO definitions...
typedef int16_t int16;
typedef uint16_t uint16;
typedef int32_t int32;
typedef uint32_t uint32;
typedef int64_t int64;
#    endif

#    if sun == 1 && __SVR4 == 1
#        include <sys/types.h>
/* Sun Solaris uses ISO definitions... */
typedef int16_t int16;
typedef uint16_t uint16;
typedef int32_t int32;
typedef uint32_t uint32;
typedef int64_t int64;
#    endif

#    ifdef _WIN32
typedef short int16;
typedef unsigned short uint16;
typedef long int32;
typedef unsigned long uint32;
typedef __int64 int64;
#    endif

/* end if __GNUC__ */
#    else

/* NOT __GNUC__ */
#    define __attribute__(arg) /* nothing */

#    endif /* end if not __GNUC__ */

#    define OME_NAMESPACE OMElibrary

enum OMEtypes_t {
    OME_NIL = 0,      /*!< nil is a special value, generally indicates no value */
    OME_INT32 = 1,    /*!< 32-bit integer */
    OME_INT = 1,      /*!< alias for 32-bit integers */
    OME_INT64 = 2,    /*!< 64-bit integer */
    OME_FLOAT = 3,    /*!< single precision floating point */
    OME_DOUBLE = 4,   /*!< double precision floating point */
    OME_OID = 5,      /*!< object Id */
    OME_STRING = 6,   /*!< string */
    OME_ARRAY = 7,    /*!< sparse array */
    OME_ASSOC = 8,    /*!< associative array */
    OME_ITERATOR = 9, /*!< internal iterator over set */
    OME_SET = 10,     /*!< set */
    OME_NLM = 11,     /*!< Native Language Message */
    OME_FIXED = 12,   /*!< fixed-point number */
    OME_POINTER = 13, /*!< internal pointer to OMEtype */
    OME_ANY = 14,     /*!< any value */
    _OME_UNUSED_15 = 15,
    _OME_UNUSED_16 = 16,
    OME_UINT32 = 17,  /*!< unsigned 32-bit integer */
    OME_UINT64 = 18,  /*!< unsigned 64-bit integer */
    OME_UINT16 = 19,  /*!< unsigned 16-bit integer */
    OME_UINT8 = 20,   /*!< unsigned 8-bit integer */

    OME_LAST_TYPE_INDEX /*!< Value one greater than last defined type */
};

enum OMEtype_OutputFlags_t
#if __cplusplus >= 201703
    : uint8_t
#endif
{
    OME_OUTPUT_TYPENAME = (1 << 0),
    OME_DISABLE_QUOTE_OUTPUT = (1 << 1),
    OME_OUTPUT_NEWLINE = (1 << 2),
    OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT = (1 << 3),
    OME_ESCAPE_UNPRINTABLE = (1 << 4),

    OME_DEFAULT_OUTPUT_MODE = OME_OUTPUT_TYPENAME,
    OME_DEFAULT_COMPLEX_OUTPUT_MODE = OME_OUTPUT_TYPENAME | OME_OUTPUT_NEWLINE | OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT
};


/* Constants to share between OMElistInterfaces() and OMEioDescriptor. */
#    define OME_IO_READ 1
#    define OME_IO_WRITE 2
#    define OME_IO_RESERVED0 4 /* ever used? */
#    define OME_IO_LISTEN 8
#    define OME_IO_CONNECT 16
#    define OME_IO_CREATE 32
#    define OME_IO_MUST_EXIST 64
#    define OME_IO_TRUNCATE 128
#    define OME_IO_APPEND 256
#    define OME_IO_NONBLOCKING 512
#    define OME_IO_DATAGRAM 1024
#    define OME_IO_RAW 2048
#    define OME_IO_MULTICAST 4096

/*
 Microsoft Visual C++ defines _STD as either std:: or ::
 HP/SGI's STL defines __STD (note two underscores vs. 1) as either std or blank
 We use the Microsoft convention and define _STD when using the HP/SGI STL
*/
#    ifndef _STD
#    ifdef __STD
/* g++-v2 */
#        define _STD __STD::
#    else
/* g++-v3 */
#        define _STD std::
#    endif
#    endif

/*! Common base-64 alphabet */
extern const unsigned char OMEbase64alphabet[];

/*! Common base-16 alphabet */
extern const unsigned char OMEhexDigits[];


#    ifndef O_CHECK_NULL_POINTER
/* NOTE: delete of null pointers is technically legal */
#    define OME_DELETE_OBJECT(ptr) delete ptr
#    define OME_DELETE_OBJECT_VECTOR(ptr) delete[] ptr
#    else
#    define OME_DELETE_OBJECT(ptr) \
    if (ptr) delete ptr
#    define OME_DELETE_OBJECT_VECTOR(ptr) \
    if (ptr) delete[] ptr
#    endif

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
