#ifndef _OME_METHOD_H
//! \cond
#define _OME_METHOD_H "$Id: OMEmethod.h 452 2020-07-23 20:19:23Z geoff $"
//! \endcond

/** \defgroup OMEmethod OME Method Classes
 * @{
 */

// Invocation Modes
#define OME_INVOKE_OIL2_INTERPRET	1
#define OME_INVOKE_OIL2_NATIVE		2
#define OME_INVOKE_KERNEL_THREAD	3
#define OME_INVOKE_TRIVIAL_NATIVE	4

#include <OMEtype.h>
#include <utils/base/atomic_values.h>

//! \brief Typedef of a native method
typedef OME_CORE_CALL int (*NativeMethodFP)(class OMEthread *);
//! \brief Typedef for stack creation routine associated with a native method
typedef OME_CORE_CALL void *(*CreateStackFP)(const class OMEinvocationData *mInfo);
//! \brief Typedef for stack delete routine associated with a native method
typedef OME_CORE_CALL void (*DeleteStackFP)(const class OMEinvocationData *mInfo, void *data);

//! \brief Default stack creation routine for a native method
extern OME_CORE_CALL void *OMEdefaultStackCreate(const class OMEinvocationData *mInfo);
//! \brief Default stack delete routine for a native method
extern OME_CORE_CALL void OMEdefaultStackDelete(const class OMEinvocationData *mInfo, void *data);


/*! \brief Record to describe the implementation of the method of an
 * OME class.
 */
class OMEinvocationData {
private:
	const struct OMEinstanceVarDescr	*varDescRec;
	OMEarray			variableDescriptions;
	mutable uint64_t	totalTicks;
	mutable uint32_t	totalSlices;
public:
	mutable uint32_t		totalInvocations;
	class OMEinvocationData		*next;
	uint32_t			signature;
	uint32_t			anyMask;	// precomputed value
	uint32_t			stackSize;
	uint32_t			invocationMode;
	union {
		NativeMethodFP	function;
		void 			*data;
	} code;
	CreateStackFP			createStackRoutine;
	DeleteStackFP			deleteStackRoutine;
	OMEarray			constantTable;
	OMEarray			localGlobals;


	OMEinvocationData();

	explicit OMEinvocationData(const OMEinvocationData &);

	~OMEinvocationData();

	int setStorageDescription(const struct OMEinstanceVarDescr *desc);

	int setStorageDescription(const OMEarray &oil2Data);

	OME_FAST_CALL void addTickCount(const int64_t tickCount) const OME_ALWAYS_INLINE {
		atomicIncrement_uint64(&totalTicks, tickCount);
		atomicIncrement_uint32(&totalSlices, 1);
	}

	OME_FAST_CALL int64_t getTickCount() const { return (totalTicks); }

	OME_FAST_CALL uint32_t getSliceCount() const { return (totalSlices); }

	OME_FAST_CALL uint32_t getInvocationCount() const { return (totalInvocations); }
}; // end class OMEinvocationData

/*! \brief Description of an OME method.
 */
class OMEmethod final : public OMEreferenceCount {
private:
	OMEinvocationData	*implementationList;
public:
	OMEstring		methodName;

	OMEmethod(const OMEstring &nm, OMEinvocationData *impl);

	~OMEmethod();

	void addImplementation(OMEinvocationData *impl);

	void addImplementation(OMEmethod *oldRec);

	OME_FAST_CALL OMEinvocationData *findImplementationWithSignature(const uint32_t sig) const;

	OME_FAST_CALL uint_fast32_t listMethodInfo(OMEarray &result, const uint_fast32_t count=0) const;

}; // end class OMEmethod

typedef OMEmethod *OMEmethodP;


/** @}
 */
#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
