#ifndef _OME_OUTPUT_ON_STREAM_H_
//! \cond
#define _OME_OUTPUT_ON_STREAM_H_ "$Id: OMEoutputOnStream.h 312 2020-03-21 21:25:35Z geoff $"
//! \endcond

/*! \file
 * \brief OME stream output functions.
 */
/** \addtogroup OMEtypes
 * @{
 */

#include <utils/base/text2int.h>


template <typename STREAMTYPE>
inline STREAMTYPE &OMEarrayStorage::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent, uint8_t includeTypePrefix) const
{
    if ((includeTypePrefix & (OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT | OME_DISABLE_QUOTE_OUTPUT)) == (OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT | OME_DISABLE_QUOTE_OUTPUT)) {
        includeTypePrefix &= ~OME_DISABLE_QUOTE_OUTPUT;
    }
    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        for (int_fast16_t count = 0; count < indent; ++count) {
            outputStream << "  ";
        }
        outputStream << "{\n";
    } else {
        outputStream << "{ ";
    }
    int_fast16_t newIndent = indent < 0 ? -indent : indent;
    newIndent += 1;
    for (SparseArrayType_t::const_iterator i = sparseArray.begin();
            i != sparseArray.end(); ++i) {
        for (int_fast16_t count = 0; count < newIndent; ++count) {
            outputStream << "  ";
        }
        outputStream << "[" << i->first << "] = ";
        i->second->outputOnStream(outputStream, -(newIndent + 1), includeTypePrefix);
    }
    newIndent -= 1;
    if (sparseArray.empty() == false) { // something was output
        for (int_fast16_t count = 0; count < newIndent; ++count) {
            outputStream << "  ";
        }
    }
    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        outputStream << "}\n";
    } else {
        outputStream << "}";
    }
    return (outputStream);
}

template <typename STREAMTYPE>
STREAMTYPE &OMEassocKeyPair::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent, uint8_t includeTypePrefix) const
{
    outputStream << "[\"" << *key << "\"] = ";
    value->outputOnStream(outputStream, indent, includeTypePrefix);
    return (outputStream);
}

template <typename STREAMTYPE>
inline STREAMTYPE &OMEassocStorage::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent, uint8_t includeTypePrefix) const
{
    if ((includeTypePrefix & (OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT | OME_DISABLE_QUOTE_OUTPUT)) == (OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT | OME_DISABLE_QUOTE_OUTPUT)) {
        includeTypePrefix &= ~OME_DISABLE_QUOTE_OUTPUT;
    }
    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        for (int_fast16_t count = 0; count < indent; ++count) {
            outputStream << "  ";
        }
        outputStream << "{\n";
    } else {
        outputStream << "{ ";
    }
    int_fast16_t newIndent = (indent < 0) ? -indent : indent;
    newIndent += 1;
    AssocArrayType_t::const_iterator i = assocArray.begin();
    while (i != assocArray.end()) {
        for (int_fast16_t count = 0; count < newIndent; ++count) {
            outputStream << "  ";
        }
        i->second->outputOnStream(outputStream, -newIndent - 1, includeTypePrefix);
//        outputStream << "\n";
        ++i; // get next element
    }
    newIndent -= 1;
    for (int_fast16_t count = 0; count < newIndent; ++count) {
        outputStream << "  ";
    }
    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        outputStream << "}\n";
    } else {
        outputStream << "}";
    }
    return (outputStream);
}

template <typename STREAMTYPE> inline STREAMTYPE &OMEnlmStorage::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent,
        uint8_t includeTypePrefix, const OMEstring *lang) const
{
    const char *fieldModifier;
    int fieldModifierLen;
    enum { NORMAL,
           SAW_ESCAPE,
           SAW_FIELD,
           SAW_FIELD_MODIFIER,
           SAW_FIELD_NUM
         } state;
    int sum = 0; // initialization not needed, but quiets compiler warnings

    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        for (int_fast16_t count = 0; count < indent; ++count) {
            outputStream << "  ";
        }
    }
    if (messageData == nullptr) { // we have no data!
        outputStream << "[NLM NULL]";
        return (outputStream);
    }
    const OMEstring &messProto = getMessageText(lang);
    if (messageArguments.isEmpty()) { // no arguments...
        // fast path, just copy data as-is...
        outputStream << messProto;
        return (outputStream);
    }
    // TO DO: create new string using messProto prototype
    const char *cp = static_cast<const char *>(messProto);
    const char *lastChar = cp + messProto.length();
    for (state = NORMAL; cp != lastChar; cp++) {
        switch (state) {
        case NORMAL:
            if (*cp == '\\') {
                state = SAW_ESCAPE;
            }
            else if (*cp == '$') {
                state = SAW_FIELD;
            }
            else {
                outputStream << *cp;
            }
            break;
        case SAW_ESCAPE:
            state = NORMAL;
            switch (*cp) {
            case 't':
                outputStream << '\t';
                break;
            case 'n':
                outputStream << '\n';
                break;
            case 'r':
                outputStream << '\r';
                break;
            default: // output as-is
                outputStream << *cp;
                break;
            }
            break;
        case SAW_FIELD:
            if ((*cp >= '0') && (*cp <= '9')) {
                sum = *cp - '0';
                state = SAW_FIELD_NUM;
                fieldModifier = 0;
                fieldModifierLen = 0;
            }
            else if (*cp == '%') {
                state = SAW_FIELD_MODIFIER;
                fieldModifier = cp; // save start...
            }
            else {
                outputStream << '$' << *cp; // $$
                state = NORMAL;
            }
            break;
        case SAW_FIELD_MODIFIER:
            if ((*cp >= '0') && (*cp <= '9')) { // end of modifier
                fieldModifierLen = cp - fieldModifier;
                sum = *cp - '0';
                state = SAW_FIELD_NUM;
            }
            break;
        case SAW_FIELD_NUM:
            if ((*cp >= '0') && (*cp <= '9')) { // end of modifier
                sum *= 10;
                sum = *cp - '0';
            }
            else {   // end of field...
                const OMEarray &args = messageArguments;
                // const OMEarray &info = argumentInfo;

                --sum; // $1 ==> argv[0]
                const OMEtype &data = args[sum];
                const OMEtype &anno = argumentInfo[sum];

                OMEnlm::outputFormattedData(outputStream,
                                            data, anno, fieldModifier,
                                            fieldModifierLen);
                state = NORMAL;
                cp -= 1; // reprocess the last char
            }
            break;
        default:
            std::cerr << "NLM format:  ILLEGAL state\n";
            break;
        }             // end switch
    }                 // end for-loop
    if (state == SAW_FIELD_NUM) { // handle last field
        const OMEarray &args = messageArguments;
        // const OMEarray &info = argumentInfo;

        --sum; // $1 ==> argv[0]
        const OMEtype &data = args[sum];
        const OMEtype &anno = argumentInfo[sum];

        OMEnlm::outputFormattedData(outputStream,
                                    data, anno, fieldModifier,
                                    fieldModifierLen);
    }
    return (outputStream);
}

template <typename STREAMTYPE> inline STREAMTYPE &OMEnlm::outputFormattedData(STREAMTYPE &outputStream,
        const class OMEbaseType &data,
        const class OMEbaseType &annotation,
        const char *fieldModifier, int modifierLen)
{
    // TO DO:  actually pay attention to modifiers, annotations
    data.outputOnStream(outputStream, 0, OME_DISABLE_QUOTE_OUTPUT);
    return (outputStream);
}

#if 0
template <typename STREAMTYPE> inline STREAMTYPE &OMEoidStorageInternal::outputOnStream(STREAMTYPE &o,
        int_fast16_t indent, bool fullInfo) const
{
    o << "OID INTERNAL REFEFENCES NOT SUPPORTED IN EXTERNAL LIBRARIES\n";
    o.flush();
    return (o);
}
#endif

template <typename STREAMTYPE> inline STREAMTYPE &OMEoidStorageExternal::outputOnStream(STREAMTYPE &outputStream,
        int_fast16_t indent, bool fullInfo) const
{
    outputStream << oidData;
    if (fullInfo) {
        outputStream << "EXT ACL=" << acl << "\n";
    }
    return (outputStream);
}

template <typename STREAMTYPE> inline STREAMTYPE &OMEoidStorage::outputOnStream(STREAMTYPE &outputStream,
        int_fast16_t indent, bool fullInfo) const
{
    switch (referenceType) {
    case OMErefType::internal: {
        const OMEoidStorageInternal *intPtr = dynamic_cast<const OMEoidStorageInternal *>(storageRef);
        intPtr->outputOnStream(outputStream, indent, fullInfo);
    }
    break;
    case OMErefType::external: {
        const OMEoidStorageExternal *extPtr = dynamic_cast<const OMEoidStorageExternal *>(storageRef);
        extPtr->outputOnStream(outputStream, indent, fullInfo);
    }
    break;
    default:
        break;
    }
    return (outputStream);
}

template <typename STREAMTYPE> inline STREAMTYPE &OMEsetStorage::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent, uint8_t includeTypePrefix) const
{
    if ((includeTypePrefix & (OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT | OME_DISABLE_QUOTE_OUTPUT)) == (OME_FORCE_QUOTES_IN_CONTAINER_OUTPUT | OME_DISABLE_QUOTE_OUTPUT)) {
        includeTypePrefix &= ~OME_DISABLE_QUOTE_OUTPUT;
    }

    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        for (int_fast16_t count = 0; count < indent; ++count) {
            outputStream << "  ";
        }
        outputStream << "{\n";
    } else {
        outputStream << "{ ";
    }

    int_fast16_t newIndent = (indent < 0) ? -indent : indent;
    bool notFirst = false;
    for (_STD deque<OMEbaseTypeP>::const_iterator i = dataSet.begin();
            i != dataSet.end(); ++i) {
        if (notFirst == false) {
            notFirst = true;
            if ((includeTypePrefix & OME_OUTPUT_NEWLINE) == 0) {
                outputStream << ", ";
            }
        }
        (*i)->outputOnStream(outputStream, newIndent + 1, includeTypePrefix);
    }
    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        for (int_fast16_t count = 0; count < newIndent; ++count) {
            outputStream << "  ";
        }
        outputStream << "}\n";
    } else {
        outputStream << "}";
    }
    return (outputStream);
}

template <typename STREAMTYPE> inline STREAMTYPE &OMEbaseType::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent,
        uint8_t includeTypePrefix) const
{
    uint32_t workBfr[6];

    if (includeTypePrefix & OME_OUTPUT_NEWLINE) {
        for (int_fast16_t i = 0; i < indent; ++i) {
            outputStream << "  ";
        }
    }
    int_fast16_t negIndent = (indent < 0) ? indent : -indent;

    if (includeTypePrefix & OUTPUT_TYPENAME) {
        outputStream << typeName() << " ";
    }
    switch (type) {
    case OME_NIL:
        if ((includeTypePrefix & OUTPUT_TYPENAME) == 0) {
            outputStream << "nil";
        }
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_INT32:
        outputStream << int_to_ascii(workBfr, sizeof(workBfr), value.i);
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_UINT32:
    case OME_UINT16:
    case OME_UINT8:
        outputStream << uint_to_ascii(workBfr, sizeof(workBfr), value.ui);
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_FLOAT:
        //      outputStream.setf(0, std::ios::floatfield);
        outputStream << float_to_ascii((char *) workBfr, sizeof(workBfr), value.f);
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_DOUBLE:
        //      outputStream.setf(0, std::ios::floatfield);
        outputStream << float_to_ascii((char *) workBfr, sizeof(workBfr), OME_EXT_TYPE(df));
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_INT64:
        outputStream << int_to_ascii(workBfr, sizeof(workBfr), OME_EXT_TYPE(i64));
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_UINT64:
        outputStream << uint_to_ascii(workBfr, sizeof(workBfr), OME_EXT_TYPE(ui64));
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_OID:
        value.oid->outputOnStream(outputStream, 0);
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_STRING:
        value.s->outputOnStream(outputStream, 0, includeTypePrefix);
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_ARRAY:
        value.array->outputOnStream(outputStream, negIndent, includeTypePrefix);
        break;
    case OME_ASSOC:
        value.assoc->outputOnStream(outputStream, negIndent, includeTypePrefix);
        break;
    case OME_SET:
        value.set->outputOnStream(outputStream, negIndent, includeTypePrefix);
        break;
    case OME_POINTER:
        if (includeTypePrefix & OUTPUT_TYPENAME) {
            outputStream << "-> ";
        }
        value.pointer->outputOnStream(outputStream, negIndent, includeTypePrefix);
        break;
    case OME_NLM:
        if ((includeTypePrefix & DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        value.nlm->outputOnStream(outputStream, 0, includeTypePrefix);
        if ((includeTypePrefix & DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    case OME_FIXED:
        value.fixed->outputOnStream(outputStream, 0);
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    default:
        outputStream << "UnsupportedType(" << type << ")";
        if (includeTypePrefix & OUTPUT_NEWLINE) {
            outputStream << "\n";
        }
        break;
    }
    return (outputStream);
}

template <typename STREAMTYPE> inline STREAMTYPE &operator<<(STREAMTYPE &outputStream, const OMEbaseType &data)
{
    uint8_t mode;
    switch (data.type) {
    case OME_ARRAY:
    case OME_ASSOC:
    case OME_SET:
        mode = OME_DEFAULT_COMPLEX_OUTPUT_MODE;
        break;
    default:
        mode = OME_DEFAULT_OUTPUT_MODE;
        break;
    }

    data.outputOnStream(outputStream, 0, mode);
    return (outputStream);
}

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
