#ifndef _OME_REFCOUNT_H
//! \cond
#define _OME_REFCOUNT_H "$Id: OMErefCount.h 406 2020-07-12 05:07:33Z geoff $"
//! \endcond

/*! \file
 * \brief OME reference count implementation.
 */
/** \addtogroup OMEtypes
 * @{
 */

#    include <OMEmanifests.h>
#    include <OMEmutex.h>

#    ifdef OME_USE_NAMESPACES
namespace OME_NAMESPACE
{
#    endif

//#define BOGUS_VALUE ((C *) (0xbadbad00))

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"

/*! \brief Base class for reference-counted data.
 *
 * OMEreference serves as the base class
 * for data that may be referenced by multiple entities.
 */
class OMEreferenceCount
{
protected:
    mutable int32_t references;
    int32_t         mutexID;

    void lockMe();

    void unlockMe();

public:
    // initializes with references=1
    OMEreferenceCount();

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    virtual ~OMEreferenceCount() {} // = delete;
#pragma GCC diagnostic pop

    void addReadOnlyReference();

    int_fast32_t dropReference();

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    virtual OMEreferenceCount *deepCopy() const { return (nullptr); }
#pragma GCC diagnostic pop

    OMEreferenceCount *getUniqueReference();

    int_fast32_t totalReferences() const OME_ALWAYS_INLINE
    {
        return (references);
    }
}; // end class OMEreferenceCount

#pragma GCC diagnostic pop

/*! \brief Templated type-specific reference to a reference-counted object.
 *
 * \param C specifies the class of the object being referenced.
 *
 * OMEreferenceToData<C> is a template to provide a type-specific handle
 * to reference-counted data and enable support for copy-on-write.
 *
 * The class argument \em C should be the name of a class that
 * both inherits from class OMEreferenceCount and holds the data.
 */
template <class C> class OME_DLL_EXPORT OMEreferenceToData
{
protected:
    C *data;

    inline void addReadOnlyReference() OME_ALWAYS_INLINE
    {
        if (OME_EXPECT_TRUE(data != nullptr)) {
            data->addReadOnlyReference();
        }
    }

    inline void dropReference(C *newData=nullptr) OME_ALWAYS_INLINE
    {
        if (OME_EXPECT_TRUE(data != nullptr)) {
            data->dropReference();
        }
        data = newData;
    }

    // default initializer
    inline OMEreferenceToData()
    {
        data = nullptr;
    }

    explicit OMEreferenceToData(C *existingData) NONNULL_CLASS_PARAMETERS(2)
    {
        existingData->addReadOnlyReference();
        data = existingData;
    }

    // copy initializer
    inline explicit OMEreferenceToData(const OMEreferenceToData<C> &org)
    {
        org.data->addReadOnlyReference();
        data = org.data;
    }

    virtual ~OMEreferenceToData()
    {
        dropReference();
    }

public:
    /*!  \brief Force unique reference to the data, which will trigger
     * duplication if necessary (copy-on-write).
     */
    inline void getUniqueReference()
    {
        if (data != nullptr) {
            C *newData = reinterpret_cast<C *>(data->getUniqueReference());
            if (newData != data) { // don't write if not needed
                data = newData;
            }
        }
    }

    /*!  \brief Get current reference total.
     *
     * \retval -1234 indicates improper initialization.
     */
    inline int_fast32_t totalReferences() const OME_ALWAYS_INLINE
    {
#    ifndef DONT_INCLUDE_VALIDITY_CHECKS
        if (OME_EXPECT_FALSE(data == nullptr)) {
            return (-1234);
        }
#    endif
        return (data->totalReferences());
    }
}; // end class OMEreferenceToData<>


#    ifdef OME_USE_NAMESPACES
} // end namespace OME_NAMESPACE
#    endif

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
