#ifndef _OME_STRING_H
#    define _OME_STRING_H "$Id: OMEstring.h 406 2020-07-12 05:07:33Z geoff $"

/*! \file
 * \brief OME string implementation.
 */
/** \addtogroup OMEtypes
 * @{
 */

#    include <string>

#    ifdef __CWTYPE__
#    define HAVE_WIDE_CHARS
#    endif

#    include <OMErefCount.h>

enum OME_character_set_t 
#if __cplusplus >= 201703
    : uint_fast8_t 
#endif
{
    OME_CHARSET_UNDEFINED = 0,
    OME_CHARSET_MASK_CHARSET = 0x1f,
    OME_CHARSET_STRING_IN_ROM = 0x20,
    OME_CHARSET_MASK_WIDE_CHARS = 0x10,
    OME_CHARSET_ASCII = 1,
    OME_CHARSET_EBCDIC = 2,
    OME_CHARSET_BINARY = 3,
    OME_CHARSET_UNICODE = (4 | OME_CHARSET_MASK_WIDE_CHARS),
#    ifdef SYSTEM390
    OME_DEFAULT_CHARSET = OME_CHARSET_EBCDIC
#    else
    OME_DEFAULT_CHARSET = OME_CHARSET_ASCII
#    endif
};

/*! \brief Return the number of bytes needed to represent the UTF character.
 */
inline constexpr uint_fast8_t UTF8characterLength(uint_fast8_t firstByte)
{
    return ( (OME_EXPECT_TRUE((firstByte & 0x80) == 0) ? 1 /* regular ASCII */ :
        ( ((firstByte & 0xe0) == 0xc0) ? 2 :
        ( ((firstByte & 0xf0) == 0xe0) ? 3 :
          4 ))));
#if 0
    if ((firstByte & 0xf8) == 0xf0) {
        return (4);
    }
#endif
}

extern uint_fast8_t UTF8characterLength(const void *charPtr) NONNULL_PARAMETERS(1);
/*! \brief Return the number of bytes needed to represent the UTF
 * character at the indicated address.
 */
inline uint_fast8_t UTF8characterLength(const void *charPtr)
{
    const unsigned char *ucp = reinterpret_cast<const unsigned char *>(charPtr);
    uint_fast8_t firstByte = *ucp;
    return (UTF8characterLength(firstByte));
}

// OMEstringStorage is used internally by OMEstring, thus all of its
// members are private and OMEstring is listed as a friend.

typedef const void *OMEstringInROM;

/*! \brief Reference-counted storage for OMEstring objects.
 */
class OMEstringStorage final : public OMEreferenceCount
{
    friend class OMEstring;
    friend class OMEreferenceToData<OMEstringStorage>;

    typedef uint_fast32_t OMEstringCharSet_t;

private:
    union {
        _STD string *str;
        OMEstringInROM romString;
#    ifdef HAVE_WIDE_CHARS
        _STD wstring *wstr;
#    endif
    } stringData;
    OMEstringCharSet_t  charSet;

    // default initializer
    OMEstringStorage();

    explicit OMEstringStorage(const char *s, uint_fast8_t impliedCharSet = OME_DEFAULT_CHARSET) NONNULL_CLASS_PARAMETERS(2);

    explicit OMEstringStorage(OMEstringInROM textSegmentString, ssize_t l = -1, uint_fast8_t impliedCharSet = OME_DEFAULT_CHARSET);

    explicit OMEstringStorage(const unsigned char *s, ssize_t l = -1, uint_fast8_t impliedCharSet = OME_CHARSET_UNDEFINED) NONNULL_CLASS_PARAMETERS(2);

    explicit OMEstringStorage(size_t l, uint_fast8_t impliedCharSet = OME_CHARSET_UNDEFINED);

    // destructor
    ~OMEstringStorage();

    static constexpr bool stringIsInROM(OMEstringCharSet_t c)
    {
        return ((c & OME_CHARSET_STRING_IN_ROM) != 0);
    }

    static constexpr bool stringIsMBCS(OMEstringCharSet_t c)
    {
        return ((c & OME_CHARSET_MASK_WIDE_CHARS) != 0);
    }

    static constexpr OMEstringCharSet_t ROMstringLen(OMEstringCharSet_t c)
    {
        return ((c >> 8) & 0xffffff);
    }

    void noteCharacterSet(uint_fast8_t c)
    {
        // retain existing flag for string-in-ROM
        charSet = (c & OME_CHARSET_MASK_CHARSET) |
                  (charSet & ~(OME_CHARSET_MASK_CHARSET));
    }

    void determineCharacterSet();

    uint_fast8_t getCharacterSet() const
    {
        return (charSet & OME_CHARSET_MASK_CHARSET);
    }

    size_t length() const OME_PURE_FUNCTION;

    size_t lengthInCharacters() const OME_PURE_FUNCTION; // deal with multi-byte character sets

    // CONVERSION
    operator const char *() const OME_PURE_FUNCTION;

    operator const unsigned char *() const OME_PURE_FUNCTION;

    // COMPARISON
    bool operator==(const OMEstringStorage &arg) const;

    bool operator!=(const OMEstringStorage &arg) const;

    bool operator<(const OMEstringStorage &arg) const;

    bool operator>(const OMEstringStorage &arg) const;

    bool operator<=(const OMEstringStorage &arg) const;

    bool operator>=(const OMEstringStorage &arg) const;

    bool operator==(const char *arg) const;

    bool operator!=(const char *arg) const;

//    OMEstringStorage &operator+=(const OMEstringStorage &rop);

    OMEstringStorage *operator+(const OMEstringStorage &rop) const;


    virtual OMEreferenceCount *deepCopy() const override;

    ssize_t substr(const unsigned char *sub, size_t len, size_t startOffset = 0) const;

    ssize_t substr(const char *sub, size_t len, size_t startOffset = 0) const
    {
        return (substr(reinterpret_cast<const unsigned char *>(sub), len, startOffset));
    }

    ssize_t substr(const OMEstringStorage &sub, size_t startOffset = 0) const;

    ssize_t lastSubstr(const unsigned char *sub, size_t len, size_t startOffset = ~static_cast<size_t>(0)) const;

    ssize_t lastSubstr(const OMEstringStorage &sub, size_t startOffset = ~static_cast<size_t>(0)) const;


    unsigned char midchar(size_t offset) const;

    uint32_t midMBCSchar(size_t offset) const;

    class OMEstring *midstr(size_t offset, size_t len) const;

    class OMEstring *midMBCSstr(size_t offset, size_t len) const;

    class OMEstring *reverse();

    class OMEstring *reverseMBCS();

    void truncateToLength(size_t newLen);

    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0, uint8_t includeTypePrefix = 0) const;

    template <typename STREAMTYPE> STREAMTYPE &outputOnStreamAsHex(STREAMTYPE &outputStream, int_fast16_t indent = 0,
            bool dense = false) const;
}; // end class OMEstringStorage


template <typename STREAMTYPE> inline STREAMTYPE &OMEstringStorage::outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent, uint8_t includeTypePrefix) const
{
    for (int_fast16_t count = 0; count < indent; ++count) {
        outputStream << "  ";
    }
    if (charSet == OME_CHARSET_BINARY) {
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        if (includeTypePrefix & OME_ESCAPE_UNPRINTABLE) {
            size_t l = this->length();
            const unsigned char *d = static_cast<const unsigned char *>(*this);
            for (size_t i = 0; i < l; ++i) {
                unsigned char c = d[i];
                if (isprint(c)) {
                    outputStream << c;
                }
                else {
                    char bfr[8];
                    snprintf(bfr, sizeof(bfr), "\\%o", c);
                    outputStream << bfr;
                }
            }
        }
        else {
            outputStream << "[binary data]";
        }
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        return (outputStream);
    }
    if (stringIsInROM(charSet)) {
        // TO DO:  deal with Multi-Byte Character Set
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        outputStream << static_cast<const char *>(stringData.romString);
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        return (outputStream);
    }

    if (stringIsMBCS(charSet)) {
#ifdef HAVE_WIDE_CHARS
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        outputStream << stringData.wstr->c_str();
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
#else
        outputStream << "[WIDESTRING]";
#endif
    }
    else {
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        outputStream << stringData.str->c_str();
        if ((includeTypePrefix & OME_DISABLE_QUOTE_OUTPUT) == 0) {
            outputStream << "\"";
        }
        //      outputStream << *(stringData.str);
    }
    return (outputStream);
}


template <typename STREAMTYPE> inline STREAMTYPE &OMEstringStorage::outputOnStreamAsHex(STREAMTYPE &outputStream,
        int_fast16_t indent, bool dense) const
{
    for (int_fast16_t count = 0; count < indent; ++count) {
        outputStream << "  ";
    }

    const size_t len = length();
    const unsigned char *ucp = static_cast<const unsigned char *>(*this);
    for (size_t i = 0; i < len; ++i) {
        if (dense == false) {
            if ((i & 0xf) == 0) {
                if (i != 0) {
                    outputStream << "\n";
                    for (int_fast16_t count = 0; count < indent; ++count) {
                        outputStream << "  ";
                    }
                }
            }
            else {
                outputStream << " ";
            }
        }
        uint8_t c = *(ucp++);
        uint8_t b = (c >> 4) & 0xf;
        outputStream << OMEhexDigits[b] << OMEhexDigits[c & 0xf];
    }
    if (dense == false) {
        outputStream << "\n";
    }
    return (outputStream);
}


class OME_DLL_EXPORT OMEstring final : public OMEreferenceToData<OMEstringStorage>
{
    friend class OMEstringStorage;

private:
    explicit OMEstring(OMEstringStorage *s)
    {
        // used by deepCopy()
        data = s;
    }

public:
    static const OMEstring nullOMEstring;

    // default initializer
    OMEstring()
    {
        data = new OMEstringStorage();
    }

    // copy initializer
    OMEstring(const OMEstring &org) : OMEreferenceToData<OMEstringStorage>(org.data)
    {
//        data = org.data;
//        addReadOnlyReference();
    }

    explicit OMEstring(const char *s, uint_fast8_t impliedCharSet = OME_DEFAULT_CHARSET)
    {
        data = new OMEstringStorage(s, impliedCharSet);
    }

    explicit OMEstring(const unsigned char *s, ssize_t l = -1, uint_fast8_t impliedCharSet = OME_CHARSET_UNDEFINED)
    {
        data = new OMEstringStorage(s, l, impliedCharSet);
    }

    explicit OMEstring(OMEstringInROM s, ssize_t l = -1, uint_fast8_t impliedCharSet = OME_DEFAULT_CHARSET)
    {
        data = new OMEstringStorage(s, l, impliedCharSet);
    }

    explicit OMEstring(size_t l, uint_fast8_t impliedCharSet = OME_CHARSET_UNDEFINED)
    {
        data = new OMEstringStorage(l, impliedCharSet);
    }

    // assignment
    OMEstring &operator=(const OMEstring &arg)
    {
        if (this->data == arg.data) {
            return (*this); // assignment to self...
        }
        arg.data->addReadOnlyReference(); // make sure it doesn't disappear...
        dropReference(arg.data);
        return (*this);
    }

    ~OMEstring() {}

    OMEstring *deepCopy() const
    {
        OMEstringStorage *storageCopy = static_cast<OMEstringStorage *>(data->deepCopy());
        OMEstring *copy = new OMEstring(storageCopy);
        return (copy);
    }

    OMEstring &operator=(const char *s)
    {
        OMEstringStorage *newStorage = new OMEstringStorage(s);
        dropReference(newStorage);
        return (*this);
    }

    OMEstring &operator=(const unsigned char *s)
    {
        OMEstringStorage *newStorage = new OMEstringStorage(s, -1);
        dropReference(newStorage);
        return (*this);
    }

    inline void noteCharacterSet(uint_fast8_t c)
    {
        data->noteCharacterSet(c);
    }

    inline void determineCharacterSet()
    {
        data->determineCharacterSet();
    }

    inline uint_fast8_t getCharacterSet() const
    {
        return (data->getCharacterSet());
    }

    inline size_t length() const
    {
        return (data->length());
    }

    // CONVERSION
    inline operator const char *() const
    {
        // cast to character string
        return (static_cast<const char *>(*data));
    }

    inline operator const unsigned char *() const
    {
        // cast to character string
        return (static_cast<const unsigned char *>(*data));
    }

    OMEstring *printableCopy() const;


    // COMPARISON
    inline bool operator==(const OMEstring &arg) const
    {
        if (data == arg.data) {
            return (true); // fast path
        }
        if (data->length() != arg.data->length()) {
            return (false);
        }
        return (*data == *arg.data);
    }

    inline bool operator!=(const OMEstring &arg) const
    {
        if (data == arg.data) {
            return (false); // fast path
        }
        if (data->length() != arg.data->length()) {
            return (true);
        }
        return (*data != *arg.data);
    }

    inline bool operator<(const OMEstring &arg) const OME_ALWAYS_INLINE
    {
        return (*data < *arg.data);
    }

    inline bool operator>(const OMEstring &arg) const OME_ALWAYS_INLINE
    {
        return (*data > *arg.data);
    }

    inline bool operator<=(const OMEstring &arg) const OME_ALWAYS_INLINE
    {
        return (*data <= *arg.data);
    }

    inline bool operator>=(const OMEstring &arg) const OME_ALWAYS_INLINE
    {
        return (*data >= *arg.data);
    }

    inline bool operator==(const char *arg) const OME_ALWAYS_INLINE
    {
        return (*data == arg);
    }

    inline bool operator!=(const char *arg) const OME_ALWAYS_INLINE
    {
        return (*data != arg);
    }

    inline bool operator==(const unsigned char *arg) const OME_ALWAYS_INLINE
    {
        return (*data == (const char *) arg);
    }

    inline bool operator!=(const unsigned char *arg) const OME_ALWAYS_INLINE
    {
        return (*data != (const char *) arg);
    }

    // CONCATENATION
    OMEstring *operator+(const OMEstring &rop) const;

    OMEstring &operator+=(const OMEstring &rop);

    inline ssize_t substr(const OMEstring &sub, size_t startOffset = 0) const
    {
        return (data->substr(*(sub.data), startOffset));
    }

    inline ssize_t lastSubstr(const OMEstring &sub, size_t startOffset = ~static_cast<size_t>(0)) const
    {
        return (data->lastSubstr(*(sub.data), startOffset));
    }

    inline ssize_t substr(const unsigned char *sub, size_t len = ~static_cast<size_t>(0), size_t startOffset = 0) const
    {
        return (data->substr(sub, len, startOffset));
    }

    inline ssize_t substr(const char *sub, size_t len = ~static_cast<size_t>(0), size_t startOffset = 0) const
    {
        return (data->substr(reinterpret_cast<const unsigned char *>(sub), len, startOffset));
    }

    inline ssize_t lastSubstr(const unsigned char *sub, size_t len = ~static_cast<size_t>(0), size_t startOffset = ~static_cast<size_t>(0)) const
    {
        return (data->lastSubstr(sub, len, startOffset));
    }

    // character-set aware access
    inline unsigned char midchar(size_t offset) const
    {
        return (data->midchar(offset));
    }

    inline OMEstring *midstr(size_t offset, size_t len) const
    {
        return (data->midstr(offset, len));
    }

    inline OMEstring *reverse() const
    {
        return (data->reverse());
    }

    inline void truncateToLength(size_t newLen)
    {
        getUniqueReference();
        data->truncateToLength(newLen);
    }

    /*! \brief Output an OMEstring to an output stream.
    * \param outputStream is a reference to an output stream.
    * \param indent specifies the number of spaces to indent before
    * beginning output.  If negative, no indent is performed
    * for the first line, but subsequent lines are indented
    * based on the absolute value of \a indent.
    * \param includeTypePrefix is a mask of flags indicating if type names
    * should prefix data elements, if strings should be enclosed in quotes,
    * etc.
     */
    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream,  int_fast16_t indent = 0, uint8_t includeTypePrefix = 0) const
    {
        data->outputOnStream(outputStream, indent, includeTypePrefix);
        return (outputStream);
    }

    /*! \brief Output the contents of an OMEstring object to an output stream
     * as a collection of hexadecimal characters.
    * \param outputStream is a reference to an output stream.
    * \param indent specifies the number of spaces to indent before
    * beginning output.  If negative, no indent is performed
    * for the first line, but subsequent lines are indented
    * based on the absolute value of \a indent.
    * \param dense is a Boolean flag indicating if the content should be
    * output as one contiguous sequence of hexadecimal characters or broken
    * across multiple lines as needed.
    */
    template <typename STREAMTYPE> STREAMTYPE &outputOnStreamAsHex(STREAMTYPE &outputStream, int_fast16_t indent = 0,
            bool dense = false) const
    {
        data->outputOnStreamAsHex(outputStream, indent, dense);
        return (outputStream);
    }
}; // end class OMEstring


/*! \brief Output an OMEstring to an output stream.
 */
template <typename STREAMTYPE> inline STREAMTYPE &operator<<(STREAMTYPE &outputStream, const OMEstring &data)
{
    data.outputOnStream(outputStream, 0, OME_DISABLE_QUOTE_OUTPUT);
    return (outputStream);
}

ssize_t findSubstring(const unsigned char *source, size_t srcLen,
                      const unsigned char *subString, size_t subLen);

ssize_t findLastSubstring(const unsigned char *source, size_t srcLen,
                          const unsigned char *subString, size_t subLen);

OMEstring *convertCaseOfString(const OMEstring &str, uint_fast8_t toLower = 0);

int caseInsensitiveCompare(const OMEstring &str1, const OMEstring &str2,
                           bool fuzzy = false);

OMEstring *substituteText(const OMEstring &srcStr, const class OMEassoc &v);

OMEstring *OMEbinaryToBase64(const OMEstring &source,
                             bool breakIntoLines = true);

OMEstring *OMEbase64ToBinary(const OMEstring &source);

size_t OMEhexToBinary(const unsigned char *src, size_t srcLen, unsigned char *dest, size_t destLen);

OMEstring *OMEhexToBinary(const OMEstring &source);

OMEstring *OMEbinaryToHex(const OMEstring &source);

OMEstring *OMEbinaryToHex(const unsigned char *source, size_t sourceLen);

OMEstring *OMEconvertURIescapes(const OMEstring &source);

OMEstring *OMEmakeSafeURI(const OMEstring &source);

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
