#ifndef _OME_THREAD_CNTRL_H
//! \cond
#define _OME_THREAD_CNTRL_H "$Id: OMEthreadCntrl.h 453 2020-07-23 20:20:44Z geoff $"
//! \endcond

#ifdef _WIN32
#include <winsock2.h>
//#include <windows.h>
#else
// POSIX Threads
#ifndef DONT_USE_PTHREADS
#include <pthread.h>
#endif
#endif

/** \addtogroup OMEthread
 * @{
 */

/*! \brief Abstract interface for an OME thread controller.
 */
class OMEthreadController {
public:
	class OMEthread	*threadInfo;
	uint32_t		exitStatus;

	explicit OMEthreadController(OMEthread *t) {
		exitStatus = ~0;
		threadInfo = t;
	}
	
	virtual ~OMEthreadController()	{}

    //! \brief Release a suspended thread
	virtual void releaseThread() = 0;
    //! \brief Suspend an active thread
	virtual void suspendThread() = 0;
    //! \brief Terminate a thread
	virtual void terminateThread() = 0;
    //! \brief Yield time slice currently being used by an active thread
	virtual void yieldSlice() = 0;
    //! \brief Execute a time slice for an active thread
	virtual void executeSlice() = 0;

    //! \brief Return label identifying the type of the thread controller.
	virtual const char *controllerType() const OME_CONST_FUNCTION = 0;
}; // end class OMEthreadController

/*! \brief Implementation of OME threads for compiled
 * OIL2 source.
 */
class OMEthreadNativeOIL2 : public OMEthreadController {
private:
public:
	explicit OMEthreadNativeOIL2(OMEthread *t);

	~OMEthreadNativeOIL2();

	virtual void releaseThread() override;
	virtual void suspendThread() override;
	virtual void terminateThread() override;
	virtual void yieldSlice() override;
	virtual void executeSlice() override;
	virtual const char *controllerType() const override;
}; // class OMEthreadNativeOIL2

/*! \brief Specialization of OMEthreadNativeOIL2 for OIL2 source
 * compiled to OIL2 Architecture Native Format.
 */
class OMEthreadInterpretOIL2 final : public OMEthreadNativeOIL2 {
public:
	explicit OMEthreadInterpretOIL2(OMEthread *t);

	~OMEthreadInterpretOIL2();

#ifdef PICK_UP_FROM_NATIVE
	virtual void releaseThread() override;
	virtual void suspendThread() override;
	virtual void terminateThread() override;
	virtual void yieldSlice() override;
	virtual void executeSlice() override;
#endif
	virtual const char *controllerType() const override;
}; // class OMEthreadInterpretOIL2

/*! \brief Implementation of OME thread controller for
 * native threads.
 */
class OMEthreadNative final : public OMEthreadController {
#ifdef _WIN32
	HANDLE			threadHandle;
#else
#ifndef DONT_USE_PTHREADS
	static void *OMEpthreadPrologue(void *);
	// POSIX Threads
	pthread_mutex_t	threadMutex;
	pthread_cond_t	suspendCondition;
	pthread_t		threadID;
#endif
#endif
public:
	explicit OMEthreadNative(OMEthread *t);

	~OMEthreadNative();

	virtual void releaseThread() override;
	virtual void suspendThread() override;
	virtual void terminateThread() override;
	virtual void yieldSlice() override;
	virtual void executeSlice() override {}	// never called for this form...

	virtual const char *controllerType() const override;
}; // end class OMEthreadNative

/*! \brief Statistics record to track amount of time used by
 * an OMEthread.
 */
class OMEtickInterval {
public:
	int64_t		    startTick;
	int64_t		    stopTick;
	int64_t		    pendingTotal;
	class OMEthread *activeThread;

	explicit OMEtickInterval(class OMEthread *thread) NONNULL_CLASS_PARAMETERS(2);

	~OMEtickInterval();

	inline void startInterval() OME_ALWAYS_INLINE {
		startTick = OMEgetCPUticks();
	}

//	void startIntervalForThread(OMEthread *thread);

	void stopInterval() {
		stopTick = OMEgetCPUticks();
		int64_t delta = stopTick - startTick;
		pendingTotal += delta;
	}

	void noteThreadDead(const class OMEthread *deadThread);

	int64_t intervalLength() const {
		return (pendingTotal);
	}

	void addIntervalCounts(const OMEinvocationData *method);
}; // class OMEtickInterval

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
