/* Copyright (C) 2010 - 2019 FARGOS Development, LLC
 * Author: Geoff Carpenter - http://www.fargos.net/gcc.html
 */

#ifndef _CIRCULAR_BFR_HPP_
//! \cond
#define _CIRCULAR_BFR_HPP_ "$Id: circular_bfr.hpp 454 2020-07-23 20:22:23Z geoff $"
//! \endcond

#include <stdint.h>
#include <time.h>
#include <utils/base/compiler_hints.h>
#ifndef _WIN32
#include <sched.h>
#endif

/** \addtogroup FARGOSbfr FARGOS Buffer Classes
 *
 * \brief These classes manage a region of memory, represented by a
 * BufferRegion, into a collection of blocks.  The CircularBufferManager
 * turns the region into strict a FIFO circular buffer of blocks.
 * The FixedBufferManager allocates blocks without imposing a strict
 * order.
 * @{
 */

/*! \brief Magic number to identify a shared memory circular buffer */
#define DEFAULT_CIRCULAR_BFR_MAGIC_NUMBER "CIRCBFR"

/*! \brief Magic number to identify a fixed-size memory buffer */
#define DEFAULT_FIXED_BFR_MAGIC_NUMBER "FIXDBFR"

/* NOTE:  structures prefixed with SharedBuffer are used in the
 * context of a shared memory segment or file; they must NOT contain
 * pointers, as these would have no meaning outside the address space
 * in which they were created.
 */

typedef uint32_t   SharedBuffer_Offset_t;  // default

/*! \brief Region header for a 32-bit shared memory segment.
 */
struct SharedBufferRegionHeader_32 {
    char        magicNumber[8];
    uint8_t     headerLength;
    char        componentId[47];
	uint32_t	byteOrderTag;	/* 0x11223344; last nybble indicates word length */
    uint32_t    _pad;
    // should be new cache line
	uint32_t	blockDimensions;
	uint32_t	offset_startFreeList;
	uint32_t	updatingThread;
	uint32_t	offset_startActiveList;
	uint32_t	offset_endActiveList;


    void initialize(const char *component, const char *magic,
        size_t regionLen);

    constexpr bool isNativeByteOrder() const {
        return (byteOrderTag == 0x11223344);
    }
};

/*! \brief Region header for a 64-bit shared memory segment.
 */
struct SharedBufferRegionHeader_64 {
    char        magicNumber[8];
    uint8_t     headerLength;
    char        componentId[47];
	uint32_t	byteOrderTag;	/* 0x18283848; last nybble indicates word length */
    uint32_t    _pad;
    // should be new cache line
	uint64_t	blockDimensions;
	uint32_t	offset_startFreeList; // should be 64 bits
	uint32_t	updatingThread;
	uint64_t	offset_startActiveList;
	uint64_t	offset_endActiveList;

    void initialize(const char *component, const char *magic,
        size_t regionLen);

    constexpr bool isNativeByteOrder() const {
        return (byteOrderTag == 0x11223344);
    }
};

/*! \brief Indicates if segment uses native byte order
 */
inline bool sharedBufferSegmentIsInNativeByteOrder(const void *segment)
{
    return (reinterpret_cast<const SharedBufferRegionHeader_32 *>(segment)->isNativeByteOrder());
}

/*! \brief Indicates if segment uses 64-bit offsets.
 */
inline bool sharedBufferSegmentSizeIs64bit(const void *segment)
{
    return (reinterpret_cast<const SharedBufferRegionHeader_32 *>(segment)->headerLength != sizeof(SharedBufferRegionHeader_32));
}

/*! \brief Allocation record for chains in a 32-bit shared memory buffer */
struct SharedBufferAllocRecord_32 {
    uint32_t    offset_allocRecord;
	uint32_t    offset_block;
	uint32_t    blockLen;
	uint32_t    usedLen;
	uint32_t    offset_nextInChain;
};

/*! \brief Allocation record for chains in a 64-bit shared memory buffer */
struct SharedBufferAllocRecord_64 {
    uint64_t    offset_allocRecord;
	uint64_t    offset_block;
	uint64_t    blockLen;
	uint64_t    usedLen;
	uint64_t    offset_nextInChain;
};

//! \brief Default SharedBufferRegionHeader
#if __SIZEOF_POINTER__ == 8
typedef SharedBufferRegionHeader_64 SharedBufferRegionHeader;
#define SHARED_BUFFER_OFFSET_SIZE 4 /* should be 8 */
#else
typedef SharedBufferRegionHeader_32 SharedBufferRegionHeader;
#define SHARED_BUFFER_OFFSET_SIZE 4
#endif

//! \brief Default SharedBufferAllocRecord
#if __SIZEOF_POINTER__ == 8
typedef SharedBufferAllocRecord_64 SharedBufferAllocRecord;
#else
typedef SharedBufferAllocRecord_32 SharedBufferAllocRecord;
#endif


/*! \brief Verify record is prepared and, if needed, wait until it
 * is prepared.
 *
 * \param rec points to the SharedBufferAllocRecord to be examined.
 *
 * Since many usage scenarios involve separate consumer and processing
 * threads, it is possible to a processing thread to become aware of
 * the existence of a buffer record before it is completed prepared.
 * The last step of record preparation is to set the usedLen field to the
 * proper length.
 */
extern void waitForBufferAllocRecordToBeReady(SharedBufferAllocRecord *rec);

#ifndef _WIN32
// This apparently orphaned inline is intentional
inline
#include <utils/base/circular_wait.hpp>
#endif


/* Layout in shared memory segment:
 * 	SharedBufferRegionHeader
 * 	SharedBufferAllocRecord[maxPackets]
 * 	...
 * 	data block
 * 	data block
 * 	....
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"
/*! \brief Interface to a buffer region.  This is an abstract class.
 *
 * A BufferRegion object manages a portion of address space which is typically
 * a shared memory segment.  In contrast to conventional memory allocation
 * routines, a BufferRegion works with \ref SharedBufferAllocRecord objects
 * that use only offsets from the base of the region and thus are able to be
 * correctly interpreted if the region is mapped at a different address by
 * a different process or the file is processed offline.
 */
class BufferRegion {
public:
    enum OwnershipState {
        NOT_OWNER = 0,  //!< indicates region is not owned by this BufferRegion
        OWN_BLOCK = 1,  //!< indicates region is owned by the BufferRegion and should be recovered when deleted.
        OWN_MAP = 2,    //!< indicates region is a memory map owned by the BufferRegion and should be unmapped when deleted.
        OWN_RECORD = 3
    };
    typedef uint_fast8_t BufferRegion_Overhead_t;
protected:
	unsigned char   *region;
	size_t          regionLength;
    size_t          blockSize;
    size_t          availableBlocks;
    OwnershipState  ownership;
public:
    /* Create a buffer region at the specified address.
     * \param ptr points at the base of the region of memory to be used
     * \param len indicate the number of bytes available in the region
     * \param blk_size indicate the maximum individual allocation unit.
     * \param ownerBy indicates the administrative ownershp state
     * of the region.
     *
     * If the administrative ownership state is set to OWN_BLOCK,
     * the region block will be deleted by the BufferRegion destructor.
     * If set to OWN_MAP, the region will be unmapped.
     */
	BufferRegion(unsigned char *ptr, size_t len, size_t blk_size,
            OwnershipState ownedBy=NOT_OWNER);

    /*! \brief Destructor for a BufferRegion.
     *
     * See the discussion in the BufferRegion constructor 
     * as to the effect of the administrative ownership state.
     */
	virtual ~BufferRegion();

    /*! \brief Reconstruct the buffer using a new region.
     *
     * \param ptr points at the base of the region of memory to be used.
     * \param len indicates the number of bytes available in the region.
     * \param blk_size indicates the maximum individual allocation unit.
     * \param ownedBy specifies the initial ownership of the block
     *
     * NOTE: This function should be used with caution.  Many usage cases
     * would be surprised to have an existing region be invalidated
     * and replaced with an alternate region, potentially of different size.
     */
    void replaceRegion(unsigned char *ptr, size_t len, size_t blk_size,
            OwnershipState ownedBy=NOT_OWNER);

    /*! \brief Change administrative ownership of region.
     *
     * \param newOwnership specifies the new ownership state.
     */
    void setOwnership(OwnershipState newOwnership) {
        ownership = newOwnership;
    }

    /*! \brief Return the administrative ownership state of the region.
     */
    OwnershipState getOwnership() const {
        return (ownership);
    }

    //! \brief Return the number of bytes in the region.
    inline size_t getRegionLength() const OME_ALWAYS_INLINE {
        return (regionLength);
    }

    //! \brief Return the block size set for the region.
    inline size_t getBlockSize() const OME_ALWAYS_INLINE {
        return (blockSize);
    }

    //! \brief Return the address of the buffer region.
    inline unsigned char *getBufferBase() const OME_ALWAYS_INLINE {
        return (region);
    }

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"

    //! \brief Return first active allocation record.
    virtual SharedBufferAllocRecord *getActiveListHead();

    //! \brief Return first free allocation record.
    virtual SharedBufferAllocRecord *getFreeListHead();

    /*! \brief Traverse to next allocation record on current chain.
     *
     * \param record points to the current block.  The first
     * record of the active or free list is obtained by
     * a call to getActiveListHead() or getFreeListHead(), respectively.
     *
     * \retval nullptr indicates the end of the chain has been reached.
     */
    virtual SharedBufferAllocRecord *traverseNextBlock(SharedBufferAllocRecord *record);

    /*! Convert a buffer allocation record into a physical address in the
     * context of the local process' address space.
     *
     * \param record points to a SharedBufferAllocRecord that should be
     * translated in the context of this buffer region.
     */
    virtual unsigned char *blockAddress(const SharedBufferAllocRecord *record) {
        return (region + record->offset_block);
    }

#pragma GCC diagnostic pop

    /*! Interface to allocate a block of storage from the buffer region.
     *
     * \param len specifies the number of bytes needed for the block.
     *
     * \retval nullptr indicates a block could not be allocated.
     */
    virtual SharedBufferAllocRecord *allocateBlock(size_t len) = 0;

    /*! Interface to return a block of storage previously obtained
     * by an allocateBlock() call.
     *
     * \param record points to the SharedBufferAllocRecord
     * representing the block to be returned.
     *
     * NOTE: not all subclasses can do something useful with
     * a returned block.
     */
    virtual void returnBlock(SharedBufferAllocRecord *record) = 0;

    /*! \brief Interface to return the per-block overhead associated with
     * an allocator.
     */
    virtual BufferRegion_Overhead_t getBlockOverhead() const = 0;

    /*! \brief Debug routine used to verify integrity of block lists.
     *
     * \param displayFlag specifies the amount of verbosity.
     * A value of 0 only reports corruption in the chain links.
     * A value of 1 provides a summary of the number of free
     * and active blocks and bytes.
     * A value of 2 displays information about each active
     * and free block.
     *
     * \retval 0 indicates no corruption was detected.
     * \retval 1 indicates some corruption was detected.
     */
    int checkLists(int displayFlag);

}; // end class BufferRegion


/*! \brief Impose a circular buffer on a \ref BufferRegion. */
class CircularBufferManager : public BufferRegion {
public:
    static int32_t TOTAL_allocation_wait_spins;
    static int32_t TOTAL_return_wait_spins;

	CircularBufferManager(unsigned char *ptr, size_t len, size_t blk_size,
        const char *component="",
        const char *magicNumber=DEFAULT_CIRCULAR_BFR_MAGIC_NUMBER);

    CircularBufferManager(unsigned char *ptr, size_t len); // existing segment

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
	~CircularBufferManager() {}
#pragma GCC diagnostic pop

    virtual SharedBufferAllocRecord *allocateBlock(size_t len) VIRTUAL_OVERRIDE ;

    virtual void returnBlock(SharedBufferAllocRecord *record) VIRTUAL_OVERRIDE;

    virtual BufferRegion_Overhead_t getBlockOverhead() const VIRTUAL_OVERRIDE {
        return (0);
    }
}; // end class CircularBufferManager

/*! \brief Impose a fixed-size buffer on a \ref BufferRegion */
class FixedBufferManager : public BufferRegion {
public:
    static int32_t TOTAL_allocation_wait_spins;

    FixedBufferManager(unsigned char *ptr, size_t len, size_t blk_size,
        const char *component="",
        const char *magicNumber=DEFAULT_FIXED_BFR_MAGIC_NUMBER);

    FixedBufferManager(unsigned char *ptr, size_t len); // existing segment

    ~FixedBufferManager();

    virtual SharedBufferAllocRecord *allocateBlock(size_t len) VIRTUAL_OVERRIDE;

    virtual void returnBlock(SharedBufferAllocRecord *record) VIRTUAL_OVERRIDE;

    virtual BufferRegion_Overhead_t getBlockOverhead() const VIRTUAL_OVERRIDE {
        return (sizeof(SharedBufferAllocRecord));
    }
}; // end class FixedBufferManager

#pragma GCC diagnostic pop

/** @} */
#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
