#ifndef _FAST_LOOKUP_TABLE_HPP_
//! \cond
#define _FAST_LOOKUP_TABLE_HPP_ "$Id: fastLookupTable.hpp 454 2020-07-23 20:22:23Z geoff $";
//! \endcond

// get definition of BinaryBlock_struct
#include <utils/logging/logging_api.hpp>
#include <utils/base/atomic_values.h>
#include <map>

/** \addtogroup FARGOSbase
 * @{
 */

/*! \defgroup FARGOSfastTable Fast Key Lookup Table
 *
 * \brief Enables very fast retrieval of a records associated with a string key
 * by trading space for time.
 *
 * Best used with keys that are case-insensitive and usually unique within 4
 * characters.
 * @{
 */


//! \brief Namespace to qualify ltBinaryBlock routine
namespace FastLookupUtils {
// namespace exists to force unique name for function

// NOTE: BinaryBlock_struct is taken from implementation in logging_api.hpp 
//! \brief Less-than comparison for BinaryBlock_struct
struct ltBinaryBlock {
    /*! \brief Performs less-than comparison on BinaryBlock_struct elements.
     *
     * These are passed by value in 4 registers rather than requiring
     * them to be passed by reference and maintained in memory.
     */
    bool operator()(const BinaryBlock_struct arg1, const BinaryBlock_struct arg2) const {
#if 0
        /* NOTE: this optimization fails if some form of string pooling is in use
         * and arg="abcdef" with length 6 and arg2="abcdef" with length 3.
         */
        if (OME_EXPECT_FALSE(arg1.dataBfr == arg2.dataBfr)) { // equal since they point at the same element
            return (false); // so not less-than
        }
#endif
        /* NOTE: with arbitrary 3 possibilities (<, =, >), we test the
         * condition that combines 2 first (probability: 2:3 vs. 1:3)
         */
        if (arg1.bfrLen >= arg2.bfrLen) { // arg2 equal length or shorter
            int rc = memcmp(arg1.dataBfr, arg2.dataBfr, arg2.bfrLen);
            if (rc < 0) { // arg1 is less
                return (true);
            }
            return (false); // arg1 is greater or matches but is equal or longer
        }
        // arg1 is shorter or same
        int rc = memcmp(arg1.dataBfr, arg2.dataBfr, arg1.bfrLen);
        if (rc <= 0) { // arg1 is less or matches but is shorter
            return (true);
        }
        return (false);
    }
}; // end ltBinaryBlock

}; // end namespace FastLookupUtils


/*! \brief Fast map lookup table.
 *
 * \param RECORD_TYPE identifies the type of the records to be referenced
 * by the map.
 * \param MAX_LOOKAHEAD specifies how many characters are to be used for
 * fixed fanout tables.
 * \param SMALLEST_CHAR identifies the smallest expected character value
 * to be seen in a key.
 * \param LARGEST_CHAR identifies the largest expected character value
 * to be seen in a key.
 * \param LT_KEY_COMPARE is the comparison routine for the key.
 * The default routine does a binary comparison and would rarely need to
 * be replaced; however, this can be useful to handle special modifiers
 * embedded within a key.
 *
 * \note
 *
 * If a character appears outside the range of [SMALLEST_CHAR,LARGEST_CHAR],
 * the key will still be accepted.  It will be added to the local
 * out-of-bounds list, which incurs O(\a ln \a N) additional overhead
 * where \a N is the number of exceptional elements recorded at that level.
 *
 * If a key is longer than the \a MAX_LOOKAHEAD limit,
 * it will still be accepted.
 * It is added to the out-of-bounds list at the leaf element of the fixed
 * fanout tables.
 * The default comparison routine honors the key length.
 *
 * The \a MAX_LOOKAHEAD parameter has the greatest effect on trading space
 * for time:  the amount of space required is proportional to (\a LARGEST_CHAR -
 * \a SMALLEST_CHAR) raised to the \a MAX_LOOKAHEAD power.
 * Simply put: the amount of space required increases exponentially
 * as \a MAX_LOOKAHEAD increases and
 * it becomes increasingly infeasible to fit all of the data into a CPU's cache.
 * If the keys are not uniformly distributed, many of the fixed fanout tables
 * will be sparse, which lowers the effectiveness of the CPU's cache with
 * respect to whatever data was retrieved.
 *
 * Arbitrary binary keys can be handled using \a SMALLEST_CHAR set to 0 and
 * \a LARGEST_CHAR set to 255; however, cache efficiency goes down and storage
 * requirements go up as the range is widened.
 * With \c MAX_LOOKAHEAD=2, about
 * 1 megabyte will be required for the data structures.  Increasing
 * \a MAX_LOOKAHEAD to 3 raises the storage requirements to about 256 megabytes.
 */
template <typename RECORD_TYPE, const uint_fast8_t MAX_LOOKAHEAD=4,
         const unsigned char SMALLEST_CHAR='A', const unsigned char LARGEST_CHAR='Z',
         typename LT_KEY_COMPARE=FastLookupUtils::ltBinaryBlock> class FastLookupTable {
protected:
    enum { ELEMENT_RANGE = (LARGEST_CHAR - SMALLEST_CHAR) + 1 };
    //! \brief pointer to records that end at this position
    RECORD_TYPE                                                 *record[ELEMENT_RANGE];
    //! \brief pointer to chain of records that had character in this position
    FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE>    *nextLevel[ELEMENT_RANGE];
    //! \brief Map to handle entries that do not fit: character out of bounds or too long
    std::map<BinaryBlock_struct,RECORD_TYPE *,LT_KEY_COMPARE>   outOfBoundEntries;  

    RECORD_TYPE *findRecordInSpillover(const BinaryBlock_struct key) const {
        typename std::map<BinaryBlock_struct,RECORD_TYPE *,LT_KEY_COMPARE>::const_iterator i = outOfBoundEntries.find(key);
        if (i == outOfBoundEntries.end()) {
            return (nullptr);
        }
        return (i->second);
    }

    RECORD_TYPE *findRecordInSpillover(const unsigned char *key, const uint_fast32_t len) const OME_ALWAYS_INLINE {
        return (findRecordInSpillover(BinaryBlock_struct(key, len)));
    }

    void addRecordToSpillover(const unsigned char *key, const uint_fast32_t keyLen, RECORD_TYPE *rec) OME_ALWAYS_INLINE {
        outOfBoundEntries.emplace(BinaryBlock_struct(key, keyLen), rec);
    }

public:
    /*! \brief Iterator for FastLookupTable.
     *
     * Allows traversal of entries in a FastLookupTable<>.
     *
     * \note Entries tend to be retrieved in lexiographic order, with the
     * exception of those entries that need special handling.  For example,
     * adding an entry that begins with a digit into a table with a range of A-Z
     * will create an exception record. Those records that begin with a
     * digit will be retrieved after all of those that began with letters.
     */
    class FastLookupTableIterator {
    private:
        //! \brief Internal class for iterator state
        class IteratorState {
        public:
            FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *levelPtr;
            int_fast16_t    currentCharIndex;

            enum SEARCH_STATES {
                FIND_RECORD, IN_CHILDREN, IN_EXTRAS, AT_END 
            } searchState;

            typename std::map<BinaryBlock_struct,RECORD_TYPE *,LT_KEY_COMPARE>::iterator   extra_it;  

            void reset(FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *level) {
                levelPtr = level;
                currentCharIndex = -1;
                searchState = FIND_RECORD;
                extra_it = levelPtr->outOfBoundEntries.begin();
            }

            IteratorState()     {}

            ~IteratorState()    {}
        }; // end internal class IteratorState

        IteratorState   stateStack[MAX_LOOKAHEAD];
        int_fast16_t    currentLookaheadDepth;

    public: 
        /*! \brief Reset the iterator so that the next call to nextElement()
         * returns the first record.
         *
         * \param level points to the table to be associated with the iterator.
         */
        void resetIterator(FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *level) {
            stateStack[0].reset(level);
            currentLookaheadDepth = 0;
        }

        /*! \brief Returns a Boolean indicator if the end of the
         * table has been seen.
         */
        inline bool atEnd() const OME_ALWAYS_INLINE {
            return (stateStack[0].searchState == IteratorState::AT_END);
        }

        /*! \brief Returns a pointer to the next element located by the iterator.
         *
         * \retval 0 indicates the end of the table has been reached.
         */
        RECORD_TYPE *nextElement() {
            // in case called again after reaching end
            if (OME_EXPECT_FALSE(atEnd())) {
                return (nullptr);
            }
            while (OME_EXPECT_TRUE(currentLookaheadDepth >= 0)) {
                IteratorState &curState = stateStack[currentLookaheadDepth]; // alias
                if (curState.searchState != IteratorState::IN_EXTRAS) {
                    // search through range, handling records and children
                    while (curState.currentCharIndex < ELEMENT_RANGE) {
                        if (curState.searchState == IteratorState::FIND_RECORD) {
                            curState.currentCharIndex += 1; // move to next entry
                            if (OME_EXPECT_FALSE(curState.currentCharIndex == ELEMENT_RANGE)) { // at end of symbols
                                curState.searchState = IteratorState::IN_EXTRAS;
                            } else {
                                curState.searchState = IteratorState::IN_CHILDREN;
                                if (curState.levelPtr->record[curState.currentCharIndex] != nullptr) { // record at this level
                                    return (curState.levelPtr->record[curState.currentCharIndex]);
                                }
                            }
                        }
                        if (curState.searchState == IteratorState::IN_CHILDREN) {
                            curState.searchState = IteratorState::FIND_RECORD;
                            if (curState.levelPtr->nextLevel[curState.currentCharIndex] != nullptr) {
                                currentLookaheadDepth += 1;
                                stateStack[currentLookaheadDepth].reset(curState.levelPtr->nextLevel[curState.currentCharIndex]);
                                RECORD_TYPE *result = nextElement();
                                if (OME_EXPECT_TRUE(result != nullptr)) {
                                    return (result);
                                }
                                // This would be very, very, very unexpected
                                // No actual data was provided underneath child node
                                currentLookaheadDepth -= 1;
                            }
                        }
                    } // end while entries in range to examine
                    // exhausted range of characters
                    curState.searchState = IteratorState::IN_EXTRAS; // process exception cases
                } // end if curState.searchState != IN_EXTRAS
                // To reach here, must be processing exception cases
                if (curState.extra_it != curState.levelPtr->outOfBoundEntries.end()) {
                    RECORD_TYPE *result = curState.extra_it->second;
                    ++(curState.extra_it);
                    return (result);
                }
                // reached end of all exception cases
                currentLookaheadDepth -= 1;
            }
            currentLookaheadDepth = 0; // reset back to 0
            stateStack[0].searchState = IteratorState::AT_END; // flag at end
            return (nullptr);
        }

        /*! \brief Constructor for FastLookupTable iterator
         *
         * \param levelPtr points to the table to be associated with the iterator.
         */
        FastLookupTableIterator(FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *levelPtr) {
            resetIterator(levelPtr);
        }

        //! \brief Destructor for FastLookupTable iterator
        ~FastLookupTableIterator() {}

        //! \brief Return first element in table
        RECORD_TYPE *firstElement() {
            resetIterator(stateStack[0].levelPtr);
            RECORD_TYPE *result = nextElement();
            return (result);
        }

    }; // end support class FastLookupTableIterator


    //! \brief Construct a FastLookupTable object
    FastLookupTable() {
        memset(record, 0, sizeof(record));
        memset(nextLevel, 0, sizeof(nextLevel));
    }

    // not really expected to be used via multiple inheritence...but who knows?
    virtual ~FastLookupTable() {
        for (unsigned i = 0; i < ELEMENT_RANGE; ++i) {
            delete nextLevel[i];
        }
    }

    /*! \brief Find record in mapping table.
     *
     * \param key points to the start of the key
     * \param keyLen indicates the length of the key; it must be greater than 0.
     *
     * \retval 0 indicates the record was not found
     */
    RECORD_TYPE *findRecord(const unsigned char *key, const uint_fast32_t keyLen) const OME_ALWAYS_OPTIMIZE("-O3") {
        OME_PREFETCH(key, 0, 0); // read-only, only for this period of time
        LOG_COMPONENT_CERR(app,debug) << "table=" << this << " Find key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) << "\"" << LOG_ENDLINE;
        RECORD_TYPE *result;
        int_fast32_t offsetInKey = 0; // could be only 8 bits
        const int_fast32_t maxLen = OME_EXPECT_TRUE((keyLen <= MAX_LOOKAHEAD)) ? static_cast<int32_t>(keyLen) - 1 : MAX_LOOKAHEAD - 1;
        const FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *levelPtr = this;

        // traverse to either length of key or MAX_LOOKAHEAD
        while (offsetInKey  < maxLen) {
            int_fast16_t iOffset = static_cast<int_fast16_t>(key[offsetInKey]) - SMALLEST_CHAR;
            /* It is relatively rare for the character to be outside the bounds
            * we are tracking.
            */
            if (OME_EXPECT_FALSE(iOffset < 0) || OME_EXPECT_FALSE(iOffset >= ELEMENT_RANGE)) {
                result = levelPtr->findRecordInSpillover(key, keyLen);
                return (result);
            }
//            OME_PREFETCH(&(levelPtr->nextLevel[iOffset]), 0, 3); // read-only, retain
            levelPtr = levelPtr->nextLevel[iOffset];
            if (OME_EXPECT_FALSE(levelPtr == nullptr)) {
                LOG_COMPONENT_CERR(app,debug) << "table=" << this <<
                    " intermediate Record at depth=" << offsetInKey <<
                    " key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) <<
                    "\"" << " is null" << LOG_ENDLINE;
                return (nullptr);
            }
            offsetInKey += 1;
        }
        if (offsetInKey == (static_cast<int_fast32_t>(keyLen) - 1)) { // made it to end of key
            int_fast16_t iOffset = static_cast<int_fast16_t>(key[offsetInKey]) - SMALLEST_CHAR;
            // last character in key
            if (OME_EXPECT_TRUE(iOffset >= 0) && OME_EXPECT_TRUE(iOffset < ELEMENT_RANGE)) {
                result = levelPtr->record[iOffset];
                LOG_COMPONENT_CERR(app,debug) << "table=" << this <<
                    " Record at depth=" << offsetInKey <<
                    " key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) <<
                    "\" is " << result << LOG_ENDLINE;
                return (result);
            }
        }
        result = levelPtr->findRecordInSpillover(key, keyLen);
        return (result);
    }

    /*! \brief Convenience interface for working with C-string keys.
     *
     * \param key points to the null-terminated string.
     */
    inline RECORD_TYPE *findRecord(const char *key) const OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        RECORD_TYPE *result = findRecord(reinterpret_cast<const unsigned char *>(key),
                (uint_fast32_t) strlen(key));
        return (result);
    }

    /*! \brief Convenience interface for working with std::string keys
     *
     * \param key references to std::string holding the key.
     */
    inline RECORD_TYPE *findRecord(const std::string &key) const OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        RECORD_TYPE *result = findRecord(reinterpret_cast<const unsigned char *>(key.data()),
                (uint_fast32_t) key.length());
        return (result);
    }

    /*! \brief Add a record to mapping table.
     *
     * \param key points to the start of the key
     * \param keyLen indicates the length of the key; it must be greater than 0.
     * \param rec is a pointer to the record which is to be stored
     *
     * \retval 1 indicates it was added within the \c MAX_LOOKAHEAD tables.
     * \retval 2 indicates it had characters outside the range 
     * \retval 3 indicates it was longer than \c MAX_LOOKAHEAD.
     *
     * \warning The contents of the \a key are not copied and must
     * be maintained at the specified location for the lifetime of
     * the record.  An ideal solution is for the key to be located
     * within the record itself and thus share its fate.
     */
    int addRecord(const unsigned char *key, const uint_fast32_t keyLen, RECORD_TYPE *rec) {
        LOG_COMPONENT_CERR(app,debug) << "table=" << this << " Add key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) << "\"" << LOG_ENDLINE;
        uint_fast32_t offsetInKey = 0;
        int_fast16_t iOffset = static_cast<int_fast16_t>(key[offsetInKey]) - SMALLEST_CHAR;
        if (OME_EXPECT_FALSE((iOffset < 0)) || OME_EXPECT_FALSE((iOffset >= ELEMENT_RANGE))) {
            LOG_COMPONENT_CERR(app,debug) << "table=" << this << " Add to spillover key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) << "\"" << LOG_ENDLINE;
            addRecordToSpillover(key, keyLen, rec);
            return (2);
        }
        FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *levelPtr = this;
        OME_PREFETCH(&(levelPtr->nextLevel[iOffset]), 1, 3); // probably updated-for-write, retain
        const uint_fast32_t maxLen = OME_EXPECT_TRUE(keyLen <= MAX_LOOKAHEAD) ? keyLen : MAX_LOOKAHEAD;
        // traverse to either length of key or MAX_LOOKAHEAD
        uint_fast32_t nextOffset = 1; // always 1 more than offsetInKey
        while (nextOffset < maxLen) {
            FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE> *nextLevelPtr = levelPtr->nextLevel[iOffset];
            if (OME_EXPECT_FALSE(nextLevelPtr == nullptr)) { // create entry for level
                nextLevelPtr = new FastLookupTable<RECORD_TYPE,MAX_LOOKAHEAD,SMALLEST_CHAR,LARGEST_CHAR,LT_KEY_COMPARE>();
                levelPtr->nextLevel[iOffset] = nextLevelPtr;
            }
            levelPtr = nextLevelPtr;
            offsetInKey = nextOffset;
            iOffset = static_cast<int_fast16_t>(key[offsetInKey]) - SMALLEST_CHAR;
            if (OME_EXPECT_FALSE((iOffset < 0)) || OME_EXPECT_FALSE((iOffset >= ELEMENT_RANGE))) {
            LOG_COMPONENT_CERR(app,debug) << "table=" << this << " Add to spillover key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) << "\"" << LOG_ENDLINE;
                levelPtr->addRecordToSpillover(key, keyLen, rec);
                return (2);
            }
            OME_PREFETCH(&(levelPtr->nextLevel[iOffset]), 1, 3); // probably updated-for-write, retain
            nextOffset += 1;
        }
        if (OME_EXPECT_TRUE(nextOffset == keyLen)) { // this was the last character
            LOG_COMPONENT_CERR(app,debug) << "table=" << this << " Add to record at depth=" << nextOffset << " key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) << "\"" << LOG_ENDLINE;
            levelPtr->record[iOffset] = rec;
            return (1);
        }
        LOG_COMPONENT_CERR(app,debug) << "table=" << this << " Add to spillover key=\"" << AS_TEXT_BUFFER((const char *) key, keyLen) << "\"" << LOG_ENDLINE;
        levelPtr->addRecordToSpillover(key, keyLen, rec);
        return (3);
    }

    /*! \brief Convenience interface for working with C-string keys.
     *
     * \param key points to the null-terminated string key
     * \param rec points to the record to be added
     *
     * \warning The key contents are not copied and must be retained
     * at the location specified by \a key.
     */
    inline int addRecord(const char *key, RECORD_TYPE *rec) OME_ALWAYS_INLINE {
        int result = addRecord(reinterpret_cast<const unsigned char *>(key),
                (uint_fast32_t) strlen(key), rec);
        return (result);
    }

    /*! \brief Convenience interface for working with \c std::string keys.
     *
     * \param key is a reference to a \c std::string holding the string key
     * \param rec points to the record to be added
     *
     * \warning The \a key contents are not copied and must be maintained
     * at the current location for the lifetime of the entry.
     * Caution is recommended to ensure that the data is not
     * subsequently deleted.
     */
    int addRecord(const std::string &key, RECORD_TYPE *rec) OME_ALWAYS_INLINE {
        int result = addRecord(reinterpret_cast<const unsigned char *>(key.data()),
                (uint_fast32_t) key.length(), rec);
        return (result);
    }

    /*! \brief Get initial iterator.
     */
    FastLookupTableIterator begin() {
        FastLookupTableIterator it(this);
        return (it);
    }

}; // end class FastLookupTable<>

/** @} */

/** @} */
#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
