#ifndef _RATE_COUNTER_HPP_
//! \cond
#define _RATE_COUNTER_HPP_ "$Id: rate_counter.hpp 454 2020-07-23 20:22:23Z geoff $"
//! \endcond
// Copyright (C) 2008 - 2019 FARGOS Development, LLC

/* O(1) rate counter
 */

#include <utils/time/tick_time.h>
#include <utils/base/assignToZero.hpp>
#include <memory.h>

/** \addtogroup FARGOSbase
 * @{
 */

/*! \defgroup FARGOSrate O(1) Rate Counter
 *
 * \brief O(1) sliding window rate counter supporting custom time windows
 * and counter types.
 * @{
 */


/*! \brief Sliding window rate counter that takes O(1) time.
 *
 * \param EVENT_COUNT_TYPE specifies the type to be used to hold
 * the number of events within the interval.  Typically this is
 * an \c int32_t, but it might be a floating point or fixed-point
 * value in some applications.
 * \param UNITS_PER_SECOND defines the granularity of the time units to
 * be tracked.
 *
 * \note \a EVENT_COUNT_TYPE must be able to represent
 * negative values as this is used by the return
 * value from add_events_to_slot().
 */
template <typename EVENT_COUNT_TYPE=int32_t, uint64_t UNITS_PER_SECOND=1> class RateCounterForInterval {
private:
    enum {
        ONE_BILLION = 1000000000UL,
        NS_PER_UNITS = ONE_BILLION / UNITS_PER_SECOND
    };
    EVENT_COUNT_TYPE    *event_slots;       //!< storage for events-per-interval counters
    EVENT_COUNT_TYPE    events_in_interval; //!< total events currently in the interval
    EVENT_COUNT_TYPE    rate_threshold; //!< maximum permitted events in the interval
    uint64_t            interval_in_units;  //!< interval duration in UNITS_PER_SECOND
    uint64_t            interval_in_nanoseconds; //!< interval duration in nanoseconds
    uint64_t            interval_threshold_in_seconds; //!< coarse interval duration in seconds
    struct timespec     last_time_used; //!< time of last recorded event
    uint32_t            last_slot_used; //!< round-robin subscript for slot table
    bool                changed_slot; //!< enables optimization if events are widely spaced

public:
    /** \brief Set the size of the interval window in units of
     * UNITS_PER_SECOND.
     *
     * \param units specifies the duration of the interval
     * in UNITS_PER_SECOND.
     *
     * \note resets any existing event counts.
     */
    void set_interval(uint64_t units) OME_ALWAYS_INLINE {
        interval_in_units = units;
        interval_in_nanoseconds = units * NS_PER_UNITS;
        if (UNITS_PER_SECOND == 1) {
            interval_threshold_in_seconds = units;
        } else {
            interval_threshold_in_seconds = (interval_in_nanoseconds / ONE_BILLION) + 1;
        }
        delete[] event_slots;
        event_slots = new EVENT_COUNT_TYPE[interval_in_units];
        for(uint_fast32_t i=0;i<interval_in_units;i+=1) {
            assignToZero(event_slots[i]);
        }
        last_time_used.tv_sec = 0;
        last_time_used.tv_nsec = 0;
        last_slot_used = 0;
        changed_slot = false;
        assignToZero(events_in_interval);
    }

    /*! \brief Construct an O(1) RateCounterForInterval object.
     *
     *  \param max_events specifies the maximum permissible events
     *         within an interval.
     *  \param units specifies the length of the interval in units of
     *  UNITS_PER_SECOND.
     */
    RateCounterForInterval(EVENT_COUNT_TYPE max_events, uint64_t units)
    {
        rate_threshold = max_events;
        event_slots = nullptr;
        set_interval(units);
    }

    ~RateCounterForInterval() {
        delete[] event_slots;
    }

    /** \brief Get the length of the interval in units
     * of \a UNITS_PER_SECOND.
     */
    inline uint32_t get_interval() const OME_ALWAYS_INLINE {
        return (interval_in_units);
    }

    /** \brief Get the length of the interval in nanoseconds.
     */
    inline uint32_t get_interval_in_nanoseconds() const OME_ALWAYS_INLINE {
        return (interval_in_nanoseconds);
    }

    /** \brief Get total number of events in the interval.
     */
    inline EVENT_COUNT_TYPE get_events_in_interval() const OME_ALWAYS_INLINE {
        return (events_in_interval);
    }

    /** \brief Get the maximum permitted events per interval.
     */
    inline EVENT_COUNT_TYPE get_maximum_rate() const OME_ALWAYS_INLINE {
        return (rate_threshold);
    }

    /** \brief Set the maximum permitted events per interval.
     */
    inline void set_maximum_rate(EVENT_COUNT_TYPE max_events) OME_ALWAYS_INLINE {
        rate_threshold = max_events;
    }


    /** \brief Add events to an interval.
     *
     * \param count specifies the number of events to be added.
     * \param now indicates the time associated with the events; if
     *        a null pointer is passed, the current time is assumed.
     *
     * \return The amount by which the rate was exceeded
     *         is returned; negative values indicate how many
     *         more events can be added while 0 indicates the
     *         last event was added.
     */
    EVENT_COUNT_TYPE add_events_to_slot(EVENT_COUNT_TYPE count=1,
        const struct timespec *now = nullptr)
    {
        struct timespec system_time;
        const struct timespec *current_time;

        if (now != nullptr) {
            current_time = now;
        } else {
            clock_gettime_via_tick(&system_time, false);
            current_time = &system_time;
        }

        int64_t delta_secs = current_time->tv_sec - last_time_used.tv_sec;
        int64_t delta_ns = current_time->tv_nsec - last_time_used.tv_nsec;
        last_time_used = *current_time;
        // fast path for initialization and events beyond the interval window
        // First test avoids possible overflow,
        // while second test updates delta_ns as a side-effect
        if ((delta_secs >= static_cast<int64_t>(interval_threshold_in_seconds)) ||
            ((delta_ns += (ONE_BILLION * delta_secs)) >= static_cast<int64_t>(interval_in_nanoseconds))) {
            if (changed_slot == true) {
                memset(event_slots, 0, sizeof(EVENT_COUNT_TYPE) * interval_in_units);
                changed_slot = false;
            }
            last_slot_used = 0;
            event_slots[0] = count; // really event_slots[last_slot_used] = count;
            events_in_interval = count;
            return (events_in_interval - rate_threshold);
        }
        if (OME_EXPECT_FALSE(delta_ns < 0)) { // very unexpected, but treat as now
            delta_ns = 0;
        }
        uint64_t delta = delta_ns / NS_PER_UNITS;
        uint_fast32_t slot = last_slot_used;
        EVENT_COUNT_TYPE newTotal = events_in_interval;
        for(uint_fast32_t i = static_cast<uint_fast32_t>(delta); i > 0; --i) {
            changed_slot = true;
            // everything prior to this was a no-op...
            if (++slot >= interval_in_units) { // wrap to next round-robin slot
                slot = 0;
            }
            newTotal -= event_slots[slot];
            assignToZero(event_slots[slot]);    // fade it away...
        }
        newTotal += count;
        events_in_interval = newTotal;
        event_slots[slot] += count;
        last_slot_used = slot;
        return (events_in_interval - rate_threshold);
    }
}; // end class RateCounterForInterval<>

// mandatory typedef for compatibility
typedef RateCounterForInterval<int32_t> RateCounter;

// optional typedef for convenience 
typedef RateCounterForInterval<int64_t> RateCounter64;

/** @} */

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
