#ifndef _TEXT2INT_H_
//! \cond
#define _TEXT2INT_H_ "$Id: text2int.h 463 2020-07-24 17:42:01Z geoff $"
//! \endcond

/* Copyright (C) 2010 - 2020 FARGOS Development, LLC
 * Author:  Geoff Carpenter -- http://www.fargos.net/gcc.html
 */


#include <stdint.h>
#include <memory.h>
#include <math.h>
#include <utils/base/compiler_hints.h>
#include <utils/base/text2int_tbl.h>
#ifndef _WIN32
#include <netinet/in.h>
#endif

/*! \file
 * \brief Extreme-performance text/number conversion routines.
 * \author Geoff Carpenter
 * \copyright 2010 - 2020 FARGOS Development, LLC
 */

/** \defgroup FARGOSbase FARGOS Base Utility Functions
 *
 * \brief The base routines in the FARGOS Development utility library
 * perform functions such as fast ASCII-to-integer conversions,
 * Bloom filters, fast lookup tables, O(1) rate counters, etc.
 * @{
 */

/*! \defgroup FARGOSnumeric FARGOS Fast Numeric Conversion Functions
 *
 * \brief These functions quickly convert between ASCII and integer,
 * floating-point and fixed-point values.
 * @{
 */

/* The "trick" here is that we add 0 to the value defined OR the keywords.
 * Assuming nothing is defined (WIN32), we end up with __BIG_ENDIAN0 and
 * __BYTE_ORDER0.  If the values are defined, we end up with 43210 and
 * either 43210, 12340 or 34120 (for the DEC PDP case).
 */
//! \cond
#define __BIG_ENDIAN0 43210
#define __BYTE_ORDER0 99990
#define __BYTE_ORDER_APPEND2(x,y) x ## y
#define __BYTE_ORDER_APPEND1(x,y) __BYTE_ORDER_APPEND2(x,y)
#define _BE_PATTERN __BYTE_ORDER_APPEND1(__BIG_ENDIAN,0)
#define _HB_PATTERN __BYTE_ORDER_APPEND1(__BYTE_ORDER,0)

#define __PERMIT_UNALIGNED_READS 1 /* since we are running on x86_64, unaligned reads are permitted */

#define _ENABLE_TABLE_PREFETCH_ 0
//! \endcond

#ifdef __cplusplus
extern "C" {
#endif

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-attribute=pure"

extern uint_fast32_t text2uint32(const char *textString, uint_fast8_t text_len) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a sequence of text characters into an
 * unsigned integer as quickly as possible. Depending on string length,
 * this routine is between 2-3 times faster than standard routines such
 * as strtol(). 
 * \b Note: text2int32() can be used if negative numbers will be seen.
 *
 * \sa For C++ code, textField2uint() is an even faster routine that can be used
 * in the rare cases that the field length is known at compile time.
 * The generic fastTextField2int() is also provides a faster implementation
 * for C++ source but drops the requirement of knowing the field length
 * at compile time.
 */
inline uint_fast32_t text2uint32(const char *textString, uint_fast8_t text_len)
{

    /* we want to read from the string, but don't intend to use it again */
    OME_PREFETCH(textString, 0, 0);
    /* The following all reference addresses determined at link time. */
    enum { baseOffset = '0' * 10 * sizeof(uint32_t) }; /* offset of first row */

#if _ENABLE_TABLE_PREFETCH_ == 1
    /* We want to read from the conversion table and will be using it
     * again, so keep it cached if possible.  The table is intentionally
     * laid out so as to keep the interesting section densely packed.
     *
     * 10 digits, max 10 characters in length, 4 bytes per digit =
     * 400 bytes total.
     * Precache 4 lines to cover up to 6 digits.
     */
    OME_PREFETCH(((const char *) text2int_table) + baseOffset + 3 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int_table) + baseOffset + 2 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int_table) + baseOffset + CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int_table) + baseOffset, 0, 3);
#endif
    const unsigned char *text = (const unsigned char *) textString;
    uint_fast32_t result = 0;
    switch (text_len) {
    case 10:
        result += text2int_table[text[text_len - 10]][9];
    case 9:
        result += text2int_table[text[text_len - 9]][8];
    case 8:
        result += text2int_table[text[text_len - 8]][7];
    case 7:
        result += text2int_table[text[text_len - 7]][6];
    case 6:
        result += text2int_table[text[text_len - 6]][5];
    case 5:
        result += text2int_table[text[text_len - 5]][4];
    case 4:
        result += text2int_table[text[text_len - 4]][3];
    case 3:
        result += text2int_table[text[text_len - 3]][2];
    case 2:
        result += text2int_table[text[text_len - 2]][1];
    case 1:
        result += text2int_table[text[text_len - 1]][0];
    default:
        break;
    }
    return (result);
}

extern int_fast32_t text2int32(const char *textString, uint_fast8_t text_len) NONNULL_PARAMETERS(1) OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a sequence of text characters into a 32-bit signed integer as
 * quickly as possible. Optional leading "-" or "+" supported.
 * \note use text2uint32() when number is known to be positive.
 */
inline int_fast32_t text2int32(const char *textString, uint_fast8_t text_len)
{
    OME_PREFETCH(textString, 0, 0);
    bool isNegative = false;
    /* find first digit, "-" or "+" */
    unsigned int i = 0;
    do { /* for performance, we assume a non-null string */
        char c = textString[i];
        /* Note:  could use isdigit(), but it would read from
         * main memory at a minimum; as written, the test is done
         * against immediate constants.
         * space, "+", and "-" all precede the digits in the ASCII
         * character set, so we test against '0' first as it has
         * the highest probability of failing.
         */
        if (OME_EXPECT_TRUE((c >= '0') && (c <= '9'))) { /* have a digit */
            break;
        }
        /* If not a digit, then a "-" is the next most probable
         * special character we would expect to encounter
         */
        if (c == '-') {
            i += 1; /* skip "-" */
            isNegative = true;
            break;
        }
        /* Rare, but possible:  a leading "+" is simply ignored. */
        if (c == '+') {
            i += 1; /* skip "+" */
            break;
        }
        i += 1; /* look at next character (we're assuming whitespace) */
    }
    while (i < text_len);
    uint32_t val = text2uint32(textString + i, text_len - i);
    /* most of the time, we have positive numbers */
    if (OME_EXPECT_TRUE(isNegative == false)) {
        return (static_cast<int_fast32_t>(val));
    }
    return (0 - static_cast<int_fast32_t>(val));
}

extern uint64_t text2uint64(const char *textString, uint_fast8_t text_len) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a sequence of text characters into
 * a 64-bit unsigned integer as quickly as possible.
 * \note text2int64() can be used if negative numbers will be seen.
 */
inline uint64_t text2uint64(const char *textString, uint_fast8_t text_len)
{
    /* we want to read from the string, but don't intend to use it again */
    OME_PREFETCH(textString, 0, 0);
    /* The following all reference addresses determined at link time. */
    enum { baseOffset = '0' * 20 * sizeof(uint64_t) }; /* offset of first row */
#if _ENABLE_TABLE_PREFETCH_ == 1
    /* We want to read from the conversion table and will be using it
     * again, so keep it cached if possible.  The table is intentionally
     * laid out so as to keep the interesting section densely packed.
     * 10 digits, max 20 characters in length, 8 bytes per digit =
     * 1600 bytes total.
     * Precache 10 lines to cover up to 8 digits.
     */
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 9 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 8 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 7 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 6 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 5 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 4 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 3 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + 2 * CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset + CACHE_LINE_LENGTH, 0, 3);
    OME_PREFETCH(((const char *) text2int64_table) + baseOffset, 0, 3);
#endif
    const unsigned char *text = (const unsigned char *) textString;
    uint64_t result = 0;
    switch (text_len) {
    case 20:
        result += text2int64_table[text[text_len - 20]][19];
    case 19:
        result += text2int64_table[text[text_len - 19]][18];
    case 18:
        result += text2int64_table[text[text_len - 18]][17];
    case 17:
        result += text2int64_table[text[text_len - 17]][16];
    case 16:
        result += text2int64_table[text[text_len - 16]][15];
    case 15:
        result += text2int64_table[text[text_len - 15]][14];
    case 14:
        result += text2int64_table[text[text_len - 14]][13];
    case 13:
        result += text2int64_table[text[text_len - 13]][12];
    case 12:
        result += text2int64_table[text[text_len - 12]][11];
    case 11:
        result += text2int64_table[text[text_len - 11]][10];
    case 10:
        result += text2int64_table[text[text_len - 10]][9];
    case 9:
        result += text2int64_table[text[text_len - 9]][8];
    case 8:
        result += text2int64_table[text[text_len - 8]][7];
    case 7:
        result += text2int64_table[text[text_len - 7]][6];
    case 6:
        result += text2int64_table[text[text_len - 6]][5];
    case 5:
        result += text2int64_table[text[text_len - 5]][4];
    case 4:
        result += text2int64_table[text[text_len - 4]][3];
    case 3:
        result += text2int64_table[text[text_len - 3]][2];
    case 2:
        result += text2int64_table[text[text_len - 2]][1];
    case 1:
        result += text2int64_table[text[text_len - 1]][0];
    default:
        break;
    }
    return (result);
}

extern int64_t text2int64(const char *textString, uint_fast8_t text_len) NONNULL_PARAMETERS(1) OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a sequence of text characters into a 64-bit
 * signed integer as quickly as possible. Optional leading "-" or
 * "+" supported.
 * \note use text2uint64() when number is known to be positive.
 *
 * \sa textField2int() is an even faster routine suitable for use by
 * C++ source.
 */
inline int64_t text2int64(const char *textString, uint_fast8_t text_len)
{
    OME_PREFETCH(textString, 0, 0);
    bool isNegative = false;
    /* find first digit, "-" or "+" */
    unsigned int i = 0;
    do { /* for performance, we assume a non-null string */
        char c = textString[i];
        if (OME_EXPECT_TRUE((c >= '0') && (c <= '9'))) { /* have a digit */
            break;
        }
        if (c == '-') {
            i += 1; /* skip "-" */
            isNegative = true;
            break;
        }
        if (c == '+') {
            i += 1; /* skip "+" */
            break;
        }
        i += 1; /* look at next character */
    }
    while (i < text_len);
    uint64_t val = text2uint64(textString + i, text_len - i);
    /* most of the time, we have positive numbers */
    if (OME_EXPECT_TRUE(isNegative == false)) {
        return (static_cast<int64_t>(val));
    }
    return (0 - static_cast<int64_t>(val));
}


extern uint32_t n2to_uint32(const void *byteData) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a network byte order 2-byte integer into a
 * native unsigned 32-bit value. */
inline uint32_t n2to_uint32(const void *byteData)
{
    const unsigned char *bytes = reinterpret_cast<const unsigned char *>(byteData);

    uint32_t result;
#if (__PERMIT_UNALIGNED_READS == 1) && defined(__GNUC__)
    const uint16_t *u16p = reinterpret_cast<const uint16_t *>(bytes);
#if _HB_PATTERN == _BE_PATTERN
    result = *u16p;
#else
    result = __builtin_bswap16(*u16p);
#endif
#else // unaligned reads not permitted or not gcc
    result = bytes[0] * 0x100 + bytes[1];
#endif
    return (result);
}

extern uint32_t n4to_uint32(const void *byteData) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a network byte order 4-byte integer into a
 * native unsigned 32-bit value. */
inline uint32_t n4to_uint32(const void *byteData)
{
    const unsigned char *bytes = reinterpret_cast<const unsigned char *>(byteData);
    uint32_t    result;
#if (__PERMIT_UNALIGNED_READS == 1) && defined(__GNUC__)
    const uint32_t *u32p = reinterpret_cast<const uint32_t *>(bytes);
#if _HB_PATTERN == _BE_PATTERN
    result = *u32p;
#else
    result = __builtin_bswap32(*u32p);
#endif
#else // unaligned reads not permitted or not gcc
    result =    bytes[0] * static_cast<uint32_t>(0x1000000) +
                bytes[1] * static_cast<uint32_t>(0x10000) +
                bytes[2] * static_cast<uint32_t>(0x100) +
                bytes[3];
#endif
    return (result);
}

extern uint64_t n8to_uint64(const void *byteData) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Convert a network byte order 8-byte integer into a
 * native unsigned 64-bit value. */
inline uint64_t n8to_uint64(const void *byteData)
{
    const unsigned char *bytes = reinterpret_cast<const unsigned char *>(byteData);
    uint64_t    result;
#if (__PERMIT_UNALIGNED_READS == 1) && defined(__GNUC__)
    const uint64_t *u64p = reinterpret_cast<const uint64_t *>(bytes);
#if _HB_PATTERN == _BE_PATTERN
    result = *u64p;
#else
    result = __builtin_bswap64(*u64p);
#endif
#else // unaligned reads not permitted or not gcc
    result =    bytes[0] * static_cast<uint64_t>(0x100000000000000) +
                bytes[1] * static_cast<uint64_t>(0x1000000000000) +
                bytes[2] * static_cast<uint64_t>(0x10000000000) +
                bytes[3] * static_cast<uint64_t>(0x100000000) +
                bytes[4] * static_cast<uint64_t>(0x1000000) +
                bytes[5] * static_cast<uint64_t>(0x10000) +
                bytes[6] * static_cast<uint64_t>(0x100) +
                bytes[7];
#endif
    return (result);
}

/* We only handle big-endian and little endian, not DEC PDP order.
 * For the time being, we assume if not known to be big endian...it's little
 * endian.
 * These functions are evaluated at COMPILE time, not runtime,
 * so we don't care about any specialized support from the target CPU,
 * like a swap instruction.
 * Typical usage is with switch-statement case labels to make decisions
 * based on PDU types encoded in network byte order without first needing
 * to do any byte-order manipulation.
 */
/*! \brief Equivalent of \b htons(), but evaluated at compile time. */
extern CONSTEXPR uint16_t const_uint16_ton2(const uint16_t arg) OME_ALWAYS_INLINE OME_PURE_FUNCTION OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Equivalent of \b htonl(), but evaluated at compile time. */
extern CONSTEXPR uint32_t const_uint32_ton4(const uint32_t arg) OME_ALWAYS_INLINE OME_PURE_FUNCTION OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Equivalent of \b htonll(), but evaluated at compile time. */
extern CONSTEXPR uint64_t const_uint64_ton8(const uint64_t arg) OME_ALWAYS_INLINE OME_PURE_FUNCTION OME_ALWAYS_OPTIMIZE("-O3");
#if _HB_PATTERN == _BE_PATTERN
/* we believe host byte order is big endian, so same as network byte order */
inline CONSTEXPR uint16_t const_uint16_ton2(const uint16_t arg)
{
    return (arg);
}

inline CONSTEXPR uint32_t const_uint32_ton4(const uint32_t arg)
{
    return (arg);
}

inline CONSTEXPR uint64_t const_uint64_ton8(const uint64_t arg)
{
    return (arg);
}
#else
/* we assume host byte order is little endian */
inline CONSTEXPR uint16_t const_uint16_ton2(const uint16_t arg)
{
    return (((arg & 0xff) << 8) | ((arg >> 8) & 0xff));
}

inline CONSTEXPR uint32_t const_uint32_ton4(const uint32_t arg)
{
    return (((arg & 0xff) << 24) |
            (((arg >> 8) & 0xff) << 16) |
            (((arg >> 16) & 0xff) << 8) |
            ((arg >> 24) & 0xff));
}

inline CONSTEXPR uint64_t const_uint64_ton8(const uint64_t arg)
{
    return (((arg & 0xff) << 56) |
            (((arg >> 8) & 0xff) << 48) |
            (((arg >> 16) & 0xff) << 40) |
            (((arg >> 24) & 0xff) << 32) |
            (((arg >> 32) & 0xff) << 24) |
            (((arg >> 40) & 0xff) << 16) |
            (((arg >> 48) & 0xff) << 8) |
            ((arg >> 56) & 0xff));
}
#endif

/*! \brief Quickly convert a binary integer into ASCII decimal text.
 *
 * \param workBfr provides space into which the ASCII text will be placed.
 * An array of uint32_t is requested in order to ensure proper alignment
 * for performance (and correctness on machines where unaligned access
 * results in a bus error).  A minimum of a 6-element uint32_t array is
 * required.
 * \param bfrLen specifies the size of the provided work buffer in bytes;
 * at least 24 bytes should be provided.  This value should always be
 * a multiple of sizeof(uint32_t), which is 4.
 * \param value specifies the 64-bit unsigned value to be output.
 * \param retStrLen is an optional argument that points at an
 * uint32_t that will be provided with the length of the generated string.
 *
 * \return a pointer to the start of the generated null-terminated text
 * string is returned.
 *
 * \note This reentrant routine is more than an order of magnitude
 * faster than sprintf().
 *
 * \sa CONVERT_UINT_TO_TEXT()
 */
extern char *uint_to_ascii(uint32_t workBfr[], uint_fast8_t bfrLen,
    uint64_t value, uint_fast8_t *retStrLen = nullptr) NONNULL_RETURN;

/*! \brief Identical to uint_to_ascii(), except that negative values are
 * accepted.
 *
 * \sa CONVERT_INT_TO_TEXT()
 */
extern char *int_to_ascii(uint32_t workBfr[], uint_fast8_t bfrLen,
    int64_t value, uint_fast8_t *retStrLen = nullptr) NONNULL_RETURN;

/*! \brief Convenience macro to declare a work buffer and convert
 * an unsigned value into text.
 * \param varName specifies the name of the new variable to be declared.
 * It will point to the text result.
 * \param sourceVal specifies the integer value to be converted
 */
#define CONVERT_UINT_TO_TEXT(varName, sourceVal) uint32_t _workBfr_ ## varName [6]; char *varName = uint_to_ascii(_workBfr_ ## varName, sizeof(_workBfr_ ## varName), sourceVal)

/*! \brief Convenience macro to declare a work buffer and convert
 * a signed value into text.  If the value is known to be positive,
 * CONVERT_UINT_TO_TEXT() is faster.
 * \param varName specifies the name of the new variable to be declared.
 * It will point to the text result.
 * \param sourceVal specifies the integer value to be converted
 */
#define CONVERT_INT_TO_TEXT(varName, sourceVal) uint32_t _workBfr_ ## varName [6]; char *varName = int_to_ascii(_workBfr_ ## varName, sizeof(_workBfr_ ## varName), sourceVal);

/*! \brief Convert a binary integer into right-justified ASCII decimal
 * text and pad on the left with a specified fill character.
 * \param workBfr provides space into which the ASCII text will be placed.
 * An array of uint32_t is requested in order to ensure proper alignment
 * for performance (and correctness on machines where unaligned access
 * results in a bus error).  A minimum of a 6-element uint32_t array is
 * required.
 * \param bfrLen specifies the size of the provided work buffer in bytes;
 * at least 24 bytes should be provided.
 * \param value specifies the 64-bit unsigned value to be output.
 * \param fieldWidth specifies the desired width of the field.
 * \param fillChar specifies the fill character to pad unused positions
 * on the left.  If not specified, it defaults to a space.
 *
 * \note Uses uint_to_ascii() for conversion.
 *
 * \return a pointer to the start of the generated null-terminated text
 * string is returned.
 */
extern char *uint_to_ascii_right_justified(uint32_t workBfr[], uint_fast8_t bfrLen,
        uint64_t value, uint_fast8_t fieldWidth, char fillChar = ' ') NONNULL_RETURN;

/*! \brief Format a fixed-point value with indicated decimal places.
 * \param resultBfr provides space into which the ASCII decimal text
 * will be placed.
 * \param bfrLen specifies the length of the provided buffer.
 * \param value specifies the fixed-point value to be output.
 * \param decimals specifies the implied number of decimal places of
 * precision.
 * \param retStrLen is an optional argument that points to a location
 * into which the size of the result string will be placed.
 * \param outputPrecision is an optional parameter that specifies the
 * desired output precision of the formatted result.  The default value
 * of -2 is used to indicate the precision specified by \a decimals
 * should be used as-is.
 * A value of -3 is used to request truncation to the minimum
 * required precision.
 * A value of -1 indicates truncation to an
 * integer with no decimal point whereas a value of 0 will truncate
 * to an integer result but retain a trailing decimal point.
 */
extern char *fixedpoint_to_ascii(char *resultBfr, uint_fast8_t bfrLen,
    int64_t value, uint_fast8_t decimals, uint_fast8_t *retStrLen = nullptr,
    int_fast8_t outputPrecision = -2) NONNULL_RETURN;

/*! \brief Return value structure for representing fixed-point values
 * with variable precision
 */
// NOTE: this is suitable for C usage
struct VariableFixedPointValue {
    int64_t     value;
    int_fast8_t precision;
};

/*! \brief Convert ASCII decimal text string to fixed point representation.
 *
 * \param str points to the text string to be converted
 * \param len is an optional argument that indicates the length of the
 * source string.  If specified as 0, strlen() will be used to determine
 * the length.
 * \param desiredPrecision is an optional parameter that specifies the
 * desired precision of the result; it will be scaled to meet the
 * constraint.  A value of -2 indicates to use the precision implied
 * by the source string.
 *
 * \note:
 */
extern VariableFixedPointValue fast_ascii_to_fixedpoint(const char *str,
    uint_fast8_t len = 0, int_fast8_t desiredPrecision = -2) NONNULL_PARAMETERS(1);

/*! \brief Convert ASCII decimal text string to fixed point representation.
 *
 * \param str points to the text string to be converted
 * \param resultPrecision is an optional argument that points at
 * a location into which the apparent precision of the source string
 * will be stored.
 * \param len is an optional argument that indicates the length of the
 * source string.  If specified as 0, strlen() will be used to determine
 * the length.
 * \param desiredPrecision is an optional parameter that specifies the
 * desired precision of the result; it will be scaled to meet the
 * constraint.  A value of -2 indicates to use the precision implied
 * by the source string.
 *
 * \sa fast_ascii_to_fixedpoint()
 */

extern int64_t ascii_to_fixedpoint(
    const char *str, uint_fast8_t *resultPrecision = nullptr,
    uint_fast8_t len = 0, int_fast8_t desiredPrecision = -2) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE;

inline int64_t ascii_to_fixedpoint(const char *str,
    uint_fast8_t *resultPrecision, uint_fast8_t len,
    int_fast8_t desiredPrecision)
{
    VariableFixedPointValue result = fast_ascii_to_fixedpoint(str, len,
            desiredPrecision);
    if (resultPrecision != nullptr) {
        *resultPrecision = static_cast<uint_fast8_t>(result.precision);
    }
    return (result.value);
}


/*! \brief Convert ASCII decimal text string to a double.
 *
 * \param str points to the text string to be converted
 * \param len is an option argument that indicates the length of the
 * source string.  If specified as 0, strlen() will be used to determine
 * the length.
 *
 * \note scientific notation is tolerated; however, values below 10^-10
 * will be treated as 0 and values above 10^10 will be calculated using
 * the pow() library function.
 */
extern double text2double(const char *str, uint_fast8_t len = 0) NONNULL_PARAMETERS(1);


/*! \brief Format a double-precision value with indicated decimal places.
 * \param resultBfr provides space into which the ASCII text will be
 * placed.
 * \param bfrLen specifies the length of the provided buffer.
 * \param value specifies the floating-point value to be output.
 * \param retStrLen is an optional argument that points to a location
 * into which the size of the result string will be placed.
 * \param outputPrecision is an optional parameter that specifies the
 * desired output precision of the formatted result.  The value of -2 is
 * used to indicate a minimal amount of precision should be used,
 * but at least one decimal place will be displayed. A value of -1
 * indicates truncation to an integer with no decimal point whereas a
 * value of 0 will truncate to an integer result but retain a trailing
 * decimal point.
 * \param roundValue is an optional Boolean flag that indicates if the
 * floating-point value should be rounded to the nearest integer or
 * simply truncated (the default operation).
 *
 * \note This reentrant routine is more than an order of magnitude
 * faster than sprintf() and accomplishes its work without any division.
 */
extern char *float_to_ascii(char *resultBfr, uint_fast8_t bfrLen,
    double value, uint_fast8_t *retStrLen = nullptr,
    int_fast8_t outputPrecision = -2,
    bool roundValue = false) NONNULL_PARAMETERS(1) NONNULL_RETURN;

extern uint_fast8_t bytesInUTFcharacter(const char val) OME_PURE_FUNCTION OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");

/*! \brief Return number of bytes required to represent UTF-encoded
 * character as indicated by value of leading byte.
 */
inline uint_fast8_t bytesInUTFcharacter(const char val)
{
    if (OME_EXPECT_TRUE((val & 0x80) == 0)) {
        return (1);
    }
    if ((val & 0xe0) == 0xc0) {
        return (2);
    }
    if ((val & 0xf0) == 0xe0) {
        return (3);
    }
    if ((val & 0xf8) == 0xf0) {
        return (4);
    }
    if ((val & 0xfc) == 0xf8) {
        return (5);
    }
    // assume...
    return (6);
}

extern uint_fast8_t bytesInUTFchar(const void *s) NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");

/*! \brief Return number of bytes required to represent UTF-encoded
 * character pointed to by argument.
 */
inline uint_fast8_t bytesInUTFchar(const void *s)
{
    const char *ptr = (const char *) s;
    return (bytesInUTFcharacter(*ptr));
}

#pragma GCC diagnostic pop

#ifdef __cplusplus
}; /* close extern "C" scope */
#endif

#ifdef __cplusplus
/* NOTE: this C++ template cannot be used in C source and must be outside
 * the extern "C" scope
 */

template <unsigned int EXPONENT> extern CONSTEXPR uint64_t powerOf10Constant() OME_CONST_FUNCTION OME_ALWAYS_INLINE;

/*! \brief Generate compile-time constant for integral power of 10
 *
 * NOTE: the return type is a 64-bit unsigned, which permits values
 * of up to 10-to-the-19th to be represented.
 * Currently EXPONENT values greater than 19 are accepted without complaint.
 */
template <unsigned int EXPONENT> inline CONSTEXPR uint64_t powerOf10Constant()
{
    return (powerOf10Constant < EXPONENT - 1 > () * 10);
}

//! \cond
// specialization of powerOf10Constant<>() to stop recursion at 0
template <> inline CONSTEXPR uint64_t powerOf10Constant<0>()
{
    return (static_cast<uint64_t>(1));
}
//! \endcond

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-attributes"

// declare to permit attributes to be assigned
template <unsigned int LEN> extern uint_fast32_t right_padded_strlen(const char *s, const char PADDING_CHAR = ' ') NONNULL_PARAMETERS(1) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");

/*! \brief Templated convenience routine to return the length of a
 * possibly padded string.  The assumption is that the string is padded on
 * the right with a consistent padding character up to a predefined length.
 *
 * \arg LEN specifies the apparent length of the string; typically this
 * is the field width.
 * \param s points to the string to be examined.
 * \param PADDING_CHAR specifies the padding character; it defaults to a space.
 *
 * \return A value between 0 and LEN will be returned, representing the offset+1
 * of the last character in the string before the padding characters appear.
 *
 * \note Although this is not declared as a \c constexpr,
 * the compiler can determine the length at compile time if
 * the source was compiled with optimization and the function
 * was passed a string constant.
 */
template <unsigned int LEN> inline uint_fast32_t right_padded_strlen(const char *s, const char PADDING_CHAR)
{
    return ((s[LEN - 1] != PADDING_CHAR) ? LEN : right_padded_strlen < LEN - 1 > (s, PADDING_CHAR));
}

template <> inline uint_fast32_t right_padded_strlen<0>(const char *s, const char PADDING_CHAR) NONNULL_PARAMETERS(1);
//! \brief specialization to stop recursion at length 0
template <> inline uint_fast32_t right_padded_strlen<0>(const char *s, const char PADDING_CHAR)
{
    return (0);
}

/*! \brief Templated ASCII-to-integer conversion routine that very quickly
 * converts fixed-length fields of positive integers.
 *
 * \a FIELD_LEN is the template parameter and specifies the field length.
 * Unlike text2uint(), the length of the field is must be known at compile
 * time.
 * \param field points at the start of the text field.
 *
 * This routine can be more than 5 to 8 times faster than standard routines
 * such as strtol().  Unfortunately, the opportunity for a field's length
 * to be known at compile time is rarely encountered.  One compromise
 * is to add the overhead of a test on field length and then select the
 * custom call for that particular length.
 *
 * \note In contrast to text2uint(), this routine does not tolerate the
 * field containing anything but the digits 0 through 9.
 * With higher levels of optimization, the compiler can evaluate this
 * code at compile time if the text field content itself is constant.
 *
 * \sa text2uint(), textField2int()
 */
template <uint_fast8_t FIELD_LEN> inline uint64_t textField2uint(const char *field)
{
    return ( ((field[0] - '0') * powerOf10Constant<FIELD_LEN - 1>()) +
        textField2uint<FIELD_LEN - 1>(field + 1) );
    
}

//! \cond
// specialization for single character string
template<> inline uint64_t textField2uint<1>(const char *field)
{
    return (field[0] - '0');
}

// specialization for a null string
template<> inline uint64_t textField2uint<0>(const char *field)
{
    return (0);
}
//! \endcond

/*! \brief Templated ASCII-to-integer conversion routine that converts
 * fixed-length fields.
 *
 * \a FIELD_LEN is the template parameter and specifies the field length.
 * Unlike text2uint(), the length of the field is must be known at compile
 * time.
 * \param field points at the start of the text field.
 *
 * \note In contrast to textField2uint(), this routine recognizes a
 * leading minus sign, accepts a leading plus sign, and tolerates
 * leading spaces.
 *
 * \sa text2uint(), textField2int()
 */
template <uint_fast8_t FIELD_LEN> int64_t textField2int(const char *field)
{
    return ((field[0] == '-') ?
        (0 - static_cast<int64_t>(textField2uint<FIELD_LEN - 1>(field + 1))) :
        ((field[0] == '+') ?  static_cast<int64_t>(textField2uint<FIELD_LEN - 1>(field + 1)) :
        ((field[0] == ' ') ? textField2int<FIELD_LEN - 1>(field + 1) :
        static_cast<int64_t>(textField2uint<FIELD_LEN>(field))) ));
}

//! \cond
// specialization for a null string
template<> inline int64_t textField2int<0>(const char *field)
{
    return (0);
}

inline int64_t fastTextField2int(const char *field, uint_fast8_t len=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Uses a field length determined at runtime to select
 * the appropriate implementation of textField2int().
 *
 * \param field points at the start of the text field.
 * \param len specifies the length of the field; if 0, strlen()
 * is used to determine the length.
 *
 * \sa text2int() - in contrast to text2int(), which reuses the
 * same instruction sequences for smaller string lengths,
 * this routine expands to distinct implementations for each possible
 * string length. It can only be used in C++ code due to the underlying
 * use of templates, whereas text2int() can be used by C source.
 */
inline int64_t fastTextField2int(const char *field, uint_fast8_t len)
{
    if (len == 0) {
        len = strlen(field);
    }
    switch (len) {
    case 20:
        return (textField2int<20>(field));
    case 19:
        return (textField2int<19>(field));
    case 18:
        return (textField2int<18>(field));
    case 17:
        return (textField2int<17>(field));
    case 16:
        return (textField2int<16>(field));
    case 15:
        return (textField2int<15>(field));
    case 14:
        return (textField2int<14>(field));
    case 13:
        return (textField2int<13>(field));
    case 12:
        return (textField2int<12>(field));
    case 11:
        return (textField2int<11>(field));
    case 10:
        return (textField2int<10>(field));
    case 9:
        return (textField2int<9>(field));
    case 8:
        return (textField2int<8>(field));
    case 7:
        return (textField2int<7>(field));
    case 6:
        return (textField2int<6>(field));
    case 5:
        return (textField2int<5>(field));
    case 4:
        return (textField2int<4>(field));
    case 3:
        return (textField2int<3>(field));
    case 2:
        return (textField2int<2>(field));
    case 1:
        return (textField2int<1>(field));
    default:
        return (0);
    } // end switch len
}
//! \endcond

#pragma GCC diagnostic pop

extern CONSTEXPR double cleanedNumericValue(double v, double maxResolution) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Return value clipped to 0 if not above a threshold of precision
 *
 * \param v specifies the value
 * \param minPrecision specifies the minimum precision.
 *
 * \return The value of zero will be returned if v is within
 * the range (-minPrecision,+minPrecision); otherwise v will be returned.
 */
inline CONSTEXPR double cleanedNumericValue(double v, double minPrecision)
{
    // would use fabs(v) but it is not constexpr on all platforms
    return ((((v >= 0) ? v : -v) >= minPrecision) ? v : 0);
}

extern CONSTEXPR double cleanedPositiveNumericValue(double v, double maxResolution) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");
/*! \brief Return value clipped to 0 if not above a threshold of precision
 *
 * \param v specifies the value that is supposed to be positive.
 * \param minPrecision specifies the minimum precision.
 *
 * The value of zero will be returned if v is not at least minPrecision.
 * Note that this means that any values below minPrecision will be treated as 0.
 */
inline CONSTEXPR double cleanedPositiveNumericValue(double v, double minPrecision)
{
    return ((v >= minPrecision) ? v : 0);
}

template <typename STREAMTYPE> STREAMTYPE &operator<<(STREAMTYPE &os, const VariableFixedPointValue arg)
{
    char    bfr[64];
    os << fixedpoint_to_ascii(bfr, sizeof(bfr), arg.value, arg.precision);
    return (os);
}


#endif /* end C++-specific functions */

/** @} */

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
