#ifndef _TIMED_MUTEX_HPP_
//! \cond
#define _TIMED_MUTEX_HPP_ "$Id: timed_mutex.hpp 454 2020-07-23 20:22:23Z geoff $"
//! \endcond
/* Copyright (C) 2010 - 2019 FARGOS Development, LLC */

#include <string>
#include <stdint.h>

#ifdef _WIN32
#include <winsock2.h>
#include <windows.h>
#define OME_MUTEX_TYPE CRITICAL_SECTION
#define OME_CONDITION_TYPE HANDLE
#else
#include <pthread.h>
#define OME_MUTEX_TYPE pthread_mutex_t
#define OME_CONDITION_TYPE pthread_cond_t
#endif /* _WIN32 */

#include <utils/base/compiler_hints.h>
#include <utils/shared_vars/shared_variable.hpp>

/** \addtogroup FARGOSbase
 * @{
 */

/** \defgroup FARGOSmutex Timed Mutex and Condition Variables
 *
 * \brief Portable, timed critical section and condition variable support.
 * @{
 */

//! \cond
#define _TOKEN_TO_STRING(x) #x
#define _EXPAND_TOKEN_TO_STRING(x) _TOKEN_TO_STRING(x)
#define THIS_FILE_AND_LINE(f,l)  f ":" _EXPAND_TOKEN_TO_STRING(l)

#ifndef HIDE_TIMED_MUTEX_STATISTICS
  #if BLDFLAG_optimized == 1
    /* hide when optimized */
    #define HIDE_TIMED_MUTEX_STATISTICS 1
  #else
    #define HIDE_TIMED_MUTEX_STATISTICS 0
  #endif
#endif
//! \endcond

/*! \brief Generic mutex implementation that supports timing statistics.
 */
class TimedMutex {
friend class TimedCondition;    // expose statistics
protected:
    //! \brief Statistics for a TimedMutex object.
    class TimedMutexStatistics {
    public:
#if HIDE_TIMED_MUTEX_STATISTICS == 0
        SharedMemoryVariableNode        namingNode;
        SMV_StandaloneNumeric<uint64_t> lockRequestedAt;
        SMV_StandaloneNumeric<uint64_t> totalDelayTicks;
        SMV_StandaloneNumeric<uint64_t> lockObtainedAt;
        SMV_StandaloneNumeric<uint64_t> totalHeldTicks;
        SMV_StandaloneNumeric<uint64_t> lockHeldBy; // thread id
        SMV_StandaloneNumeric<int32_t>  locksWanted;
        SMV_StandaloneNumeric<int32_t>  locksObtained;
        SMV_StandaloneString<128>       lockLocation;
#else
        uint64_t        lockRequestedAt;
        uint64_t        totalDelayTicks;
        uint64_t        lockObtainedAt;
        uint64_t        totalHeldTicks;
        uint64_t        lockHeldBy;
        uint32_t        locksWanted;
        uint32_t        locksObtained;
        const char      *lockLocation;
#endif
        TimedMutexStatistics(const char *name, SharedMemoryVariableNode *parentNode);

        TimedMutexStatistics(const char *name, SharedMemoryVariableManager *mgr);

        ~TimedMutexStatistics() {}

    }; // end internal class TimedMutexStatistics

    TimedMutexStatistics    *statistics;
    uint64_t                obtainedAt;
    OME_MUTEX_TYPE          mutex;

    int64_t acquireLock();

public:
    /*! \brief Implements generic mutex with performance counters.
     *
     * \param lockName is a user-provided name to label
     * the lock.  If a null pointer, no statistics will be maintained.
     * \param parentNode specifies a root under which any SharedMemoryVariable
     * counters should be registered.
     * \param recursive indicates if a recursive lock is desired.
     */
    explicit TimedMutex(const char *lockName, SharedMemoryVariableNode *parentNode,
        bool recursive=false);

    /*! \brief Implements generic mutex with performance counters.
     *
     * \param lockName is a user-provided name to label
     * the lock.  If a null pointer, no statistics will be maintained.
     * \param mgr specifies SharedMemoryVariableManager to which
     * the counters should be registered.
     * \param recursive indicates if a recursive lock is desired.
     */
    explicit TimedMutex(const char *lockName=nullptr,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager,
        bool recursive=false);

    ~TimedMutex();

    /*! \brief Lock the mutex.
     * \param place specifies the location in readable form.
     *
     * \note See \ref LOCK_HERE().
     *
     * \return The number of nanoseconds required to
     * obtain the lock is returned.
     *
     * \note Timings are still provided even if statistics are not maintained.
     * Use untimedLock() for the variant which does not provide
     * any timing data.
     */
    int64_t lock(const char *place=nullptr) OME_ALWAYS_INLINE {
        int64_t result = acquireLock();
#if HIDE_TIMED_MUTEX_STATISTICS == 0
        if ((statistics != nullptr) && (place != nullptr)) {
            statistics->lockLocation = place;
        }
#endif
        return (result);
    }

    /*! \brief Lock the mutex, but maintain
     * neither usage nor timing statistics.
     *
     * \retval 0 is always returned.
     */
    inline int64_t untimedLock() OME_ALWAYS_INLINE {
#ifdef _WIN32
        EnterCriticalSection(&mutex);
#else
        pthread_mutex_lock(&mutex);
#endif
        return (0);
    }

    /*! \brief Release a previously locked mutex.
     *
     * \return The duration in nanoseconds that
     * the lock was held is returned.
     *
     * \note Timings are still provided even if statistics are not maintained.
     * Use untimedUnlock() for the variant which does not provide
     * any timing data.
     */
    int64_t unlock();

    /*! \brief Unlock the mutex, but maintain
     * neither usage nor timing statistics.
     *
     * \retval 0 is always returned.
     */
    inline int64_t untimedUnlock() OME_ALWAYS_INLINE {
#ifdef _WIN32
        LeaveCriticalSection(&mutex);
#else
        pthread_mutex_unlock(&mutex);
#endif
        return (0);
    }

    /*! \brief Get address of underlying native mutex.
     */
    inline OME_MUTEX_TYPE *mutex_address() OME_ALWAYS_INLINE NONNULL_RETURN {
        return (&mutex);
    }

    /*! \brief Return name of mutex.
     *
     * \retval "TimedMutex" will be returned if the name
     * was never set; this will never return a null pointer.
     */
    const char *getMutexName() const NONNULL_RETURN {
#if HIDE_TIMED_MUTEX_STATISTICS == 0
        if (statistics != nullptr) {
            return (statistics->namingNode.getName());
        }
#endif
        return ("TimedMutex");
    }

    /*! \brief Return thread id of thread holding lock.
     *
     * \retval 0 indicates either no thread is holding the lock
     * or it was not remembered because statistics are not available.
     */
    inline uint64_t getThreadIdOfCurrentOwner() const OME_ALWAYS_INLINE {
#if HIDE_TIMED_MUTEX_STATISTICS == 0
        if (statistics != nullptr) {
            return (statistics->lockHeldBy);
        }
#endif
        return (0);
    }
}; // end class TimedMutex


/*! \brief Generic condition variable for use with the \ref TimedMutex class.
 */
class TimedCondition {
protected:
    //! \brief Statistics for a TimedCondition object.
    class TimedConditionStatistics {
    public:
#if HIDE_TIMED_MUTEX_STATISTICS == 0
        SharedMemoryVariableNode        namingNode;
        SMV_StandaloneNumeric<uint64_t> waitStartedAt;
        SMV_StandaloneNumeric<uint64_t> waitCompletedAt;
        SMV_StandaloneNumeric<uint64_t> totalWaitTicks;
        SMV_StandaloneNumeric<uint64_t> totalPostEvents;
        SMV_StandaloneNumeric<int32_t>  waitsStarted;
        SMV_StandaloneNumeric<int32_t>  waitsCompleted;
        SMV_StandaloneString<128>       waitLocation;
#else
        uint64_t        waitStartedAt;
        uint64_t        waitCompletedAt;
        uint64_t        totalWaitTicks;
        uint64_t        totalPostEvents;
        uint32_t        waitsStarted;
        uint32_t        waitsCompleted;
        const char      *waitLocation;
#endif
        TimedConditionStatistics(const char *name, SharedMemoryVariableNode *parentNode);

        TimedConditionStatistics(const char *name, SharedMemoryVariableManager *mgr);

        ~TimedConditionStatistics() {}

    }; // end internal class TimedConditionStatistics


    TimedConditionStatistics    *statistics;
    uint64_t                    startedAt;
    TimedMutex                  *mutex;
    OME_CONDITION_TYPE          condition;
    bool                        active;
public:
    /*! \brief Implements generic condition variable with performance counters.
     *
     * \param m is an existing mutex to be associated with the condition
     * variable.  It is expected to be locked before calling waitForCondition()
     * or waitForConditionOrUntil().
     * \param conditionName is a user-provided name to label
     * the condition.  If a null pointer, no statistics will be maintained.
     * \param parentNode specifies a root under which any SharedMemoryVariable
     * counters should be registered.
     */
    explicit TimedCondition(TimedMutex *m, const char *conditionName,
        SharedMemoryVariableNode *parentNode) NONNULL_CLASS_PARAMETERS(2);

    /*! \brief Implements generic condition variable with performance counters.
     *
     * \param m is an existing mutex to be associated with the condition
     * variable.  It is expected to be locked before calling waitForCondition()
     * or waitForConditionOrUntil().
     * \param conditionName is a user-provided name to label
     * the condition.  If a null pointer, no statistics will be maintained.
     * \param mgr specifies SharedMemoryVariableManager to which
     * the counters should be registered.
     */
    explicit TimedCondition(TimedMutex *m, const char *conditionName=nullptr,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager) NONNULL_CLASS_PARAMETERS(2);

    ~TimedCondition();

    /*! \brief Return address of mutex object associated with the condition
     * variable.
     */
    TimedMutex *getMutex() const OME_ALWAYS_INLINE NONNULL_RETURN {
        return (mutex);
    }

    /*! \brief Get address of underlying native condition variable.
     */
    inline OME_CONDITION_TYPE *condition_address() OME_ALWAYS_INLINE NONNULL_RETURN {
        return (&condition);
    }

    /*! \brief Wait indefinitely for the condition
     * to be posted.
     *
     * \param waitingHere specifies the location of the call.
     * See \ref WAIT_HERE().
     *
     * \retval The duration of the wait in nanoseconds is returned.
     *
     * \note Timings are still provided even if statistics are not maintained.
     */
    int64_t waitForCondition(const char *waitingHere=nullptr);

    /*! \brief Wait for the condition to be posted or
     * until a specified point in time.
     *
     * \param timeout specifies a point in time after which the routine will
     * return.  This is not a difference in time relative to the invocation
     * of the function call, but instead an absolute point in time.
     *
     * \param waitingHere specifies the location of the call.
     * See \ref WAIT_HERE().
     *
     * \retval The duration of the wait in nanoseconds is returned.
     *
     * \note Timings are still provided even if statistics are not maintained.
     */
    int64_t waitForConditionOrUntil(const struct timespec *timeout,
            const char *waitingHere=nullptr) NONNULL_CLASS_PARAMETERS(2);

    /*! \brief Notify any sleeping threads that the condition
     * has occurred.
     *
     * \note Normally the associated mutex should have been locked,
     * but if one is aware of what they are doing, it is possible to
     * use this routine without having locked the mutex.
     *
     * \retval 0 is always returned
     */
    int postCondition();

    /*! \brief Return a Boolean indication of
     * whether or not a thread is sleeping.
     */
    inline bool isSleeping() const OME_ALWAYS_INLINE {
        return (active);
    }

    /*! \brief Return name of condition variable.
     *
     * \retval "TimedCondition" is returned if the name
     * was never set; this routine will never return a
     * null pointer.
     */
    const char *getConditionName() const NONNULL_RETURN {
#if HIDE_TIMED_MUTEX_STATISTICS == 0
        if (statistics != nullptr) {
            return (statistics->namingNode.getName());
        }
#endif
        return ("TimedCondition");
    }

    /*! \brief Convenience routine to lock the associated mutex.
     *
     * \param place specifies the location in readable form.
     *
     * \note See \ref LOCK_HERE().
     *
     * \return The number of nanoseconds required to
     * obtain the lock is returned.
     */
    int64_t lock(const char *place=nullptr) OME_ALWAYS_INLINE {
        return (mutex->lock(place));
    }

    /*! \brief Convenience routine to lock the associated mutex, but maintain
     * neither usage nor timing statistics.
     *
     * \retval 0 is always returned.
     */
    inline int64_t untimedLock() OME_ALWAYS_INLINE {
        return (mutex->untimedLock());
    }

    /*! \brief Convenience routine to release previously locked mutex
     * associated with the condition variable.
     *
     * \return The duration in nanoseconds that
     * the lock was held is returned.
     */
    int64_t unlock() OME_ALWAYS_INLINE {
        return (mutex->unlock());
    }

    /*! \brief Convenience routine to unlock the mutex associated with
     * the condition variable, but maintain neither usage nor timing statistics.
     *
     * \retval 0 is always returned.
     */
    inline int64_t untimedUnlock() OME_ALWAYS_INLINE {
        return (mutex->untimedLock());
    }

    /*! \brief Convenience routine to get address of underlying native mutex
     * that is associated with the condition variable.
     */
    inline OME_MUTEX_TYPE *mutex_address() OME_ALWAYS_INLINE NONNULL_RETURN {
        return (mutex->mutex_address());
    }
}; // end class TimedCondition

/*! \brief Convenience macro to lock a TimedMutex and attribute
 * the location of the source file and line number.
 */
#define LOCK_HERE() lock(THIS_FILE_AND_LINE(__FILE__,__LINE__))

/*! brief Convenience macro to call waitForCondition() and
 * note source file and line number.
 */
#define WAIT_HERE() waitForCondition(THIS_FILE_AND_LINE(__FILE__,__LINE__))

/*! brief Convenience macro to call waitForConditionOrUntil()
 * and note source file and line number.
 */
#define WAIT_HERE_UNTIL(timeout) waitForConditionOrUntil(timeout, THIS_FILE_AND_LINE(__FILE__,__LINE__))

/** @} */

/** @} */
#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
