#ifndef _HTTP_EMBEDDED_SERVER_HPP_
//! \cond
#define _HTTP_EMBEDDED_SERVER_HPP_ "$Id: HTTPembeddedServer.hpp 411 2020-07-13 03:02:58Z geoff $"
//! \endcond

#include <map>
#include <set>

#include <utils/io/io_processor.hpp>
#include <utils/io/read_file.hpp>

/*! \file
 * \brief FARGOS Embedded HTTP server
 */

/** \defgroup FARGOShttp FARGOS Embedded HTTP server
 * @{
 */


/*! \brief Protocol-independent representation of
 * a socket address. IPv4, IPv6 and file domain
 * sockets are supported.
 */
class GenericSocketAddress {
public:
    union GenericSocketAddress_union {
        unsigned char       bytes[63];
        struct sockaddr     genericAddr;
        struct sockaddr_in  ipv4;
        struct sockaddr_in6 ipv6;
#ifndef _WIN32
        struct sockaddr_un  filename;
#endif
    } socketAddress;
    uint8_t                 addressLength;

    /*! \brief Assign from an URL representation of the address
     * and an integer value of the port.  The following URL prefixes
     * are recognized:  "ipv4:", "ipv6:" and "file:".  If no
     * prefix is specified, the default of "ipv4:" is assumed.
     * 
     * \param targetAddress points to a string containing the
     * text representation of the address, which may be prefixed
     * with the relevant type. 
     * For IPv4 addresses, this is expected to be a dotted-decimal string.
     * For IPv6 addresses, this is expected to be a colon-separated string
     * of hexadecimal values.
     * For files, this is expected to be a path.
     * \param port specifies the port.
     *
     * Use setFromURL() if the address text also contains the port.
     */
    int setFromHostAndPort(const char *targetAddress, uint16_t port) {
        const char *address = targetAddress;
        int l = (int) strlen(address);
        const char *colon = (const char *) memchr(address, ':', l);

        int selectedFamily = AF_INET;

        if (colon != nullptr) {
            uint_fast32_t prefixLen = (uint_fast32_t) (colon - address);
            if (prefixLen == 4) {
                if (memcmp(address, "ipv4:", 5) == 0) { // IP V4
                    selectedFamily = AF_INET;
                    address += 5;
                    if ((*address == '/') && (*(address + 1) == '/')) {
                        address += 2; // skip leading "/"
                    }
                } else if (memcmp(address, "ipv6:", 5) == 0) { // IP V6
                    selectedFamily = AF_INET6;
                    address += 5;
                    if ((*address == '/') && (*(address + 1) == '/')) {
                        address += 2; // skip leading "/"
                    }
#ifndef _WIN32
                } else if (memcmp(address, "file:", 5) == 0) { // file domain
                    selectedFamily = AF_LOCAL; // AF_UNIX
                    address += 5;
                    if ((*address == '/') && (*(address + 1) == '/')) {
                        address += 2; // skip leading "/"
                    }
                    socketAddress.filename.sun_family = AF_LOCAL;
                    strncpy(socketAddress.filename.sun_path, address, sizeof(socketAddress.filename.sun_path));
                    addressLength = strlen(socketAddress.filename.sun_path);
                    return (addressLength);
#endif
                }
            }
        }
        int rc;
        switch (selectedFamily) {
        case AF_INET:
            memset(&socketAddress.ipv4, 0, sizeof(socketAddress.ipv4));
            socketAddress.ipv4.sin_family = AF_INET;
            rc = inet_pton(selectedFamily, address, &socketAddress.ipv4.sin_addr);
            socketAddress.ipv4.sin_port = htons(port);
            addressLength = sizeof(struct sockaddr_in);
            break;
        case AF_INET6:
            memset(&socketAddress.ipv6, 0, sizeof(socketAddress.ipv6));
            socketAddress.ipv6.sin6_family = AF_INET6;
            rc = inet_pton(selectedFamily, address, &socketAddress.ipv6.sin6_addr);
            socketAddress.ipv6.sin6_port = htons(port);
            addressLength = sizeof(struct sockaddr_in6);
            break;
        default:
            rc = -1;
            addressLength = 0;
            break;
        } // end switch selected address family
        return (addressLength);
    }

    /*! \brief Convenience cover to setFromHostAndPort() that
     * also converts the port.
     *
     * \param addressString points to the text string to be parsed
     * and converted.
     */
    int setFromURL(const char *addressString)
    {
        enum { MAX_COLONS = 16 };
        char    workBfr[2048];
        char    *colonAddr[MAX_COLONS];

        uint_fast8_t colonsSeen = 0;

        const char *src = addressString;
        char *dest = workBfr;
    
        while (*src != '\0') {
            if (*src == ':') {
                if (colonsSeen < MAX_COLONS) {
                    colonAddr[colonsSeen] = dest;
                    colonsSeen += 1;
                }
            }
            *(dest++) = *(src++);
        }
        int port = 0;
        if (colonsSeen != 0) { // port specified 
            uint32_t offsetOfLastColon = (uint32_t) (colonAddr[colonsSeen - 1] - workBfr);
            if (offsetOfLastColon > 5) { // not ipv4: or similar prefix
                *colonAddr[colonsSeen - 1] = '\0'; // terminate host portion
                uint32_t portLen = (uint32_t) (dest - colonAddr[colonsSeen - 1]);
                if (portLen < 7) { // :65535 + null
                    port = text2uint32(colonAddr[colonsSeen - 1] + 1, portLen - 1);
                }
            }
        }
        int result = setFromHostAndPort(workBfr, port);
        return (result);
    }

    GenericSocketAddress() OME_ALWAYS_INLINE {
        addressLength = 0;
    }

    explicit GenericSocketAddress(const struct sockaddr_in ipv4) OME_ALWAYS_INLINE {
        socketAddress.ipv4 = ipv4;
        addressLength = sizeof(struct sockaddr_in);
    }

    explicit GenericSocketAddress(const struct sockaddr_in6 ipv6) OME_ALWAYS_INLINE {
        socketAddress.ipv6 = ipv6;
        addressLength = sizeof(struct sockaddr_in6);
    }

    explicit GenericSocketAddress(const char *addressString) {
        setFromURL(addressString);
    }

    explicit GenericSocketAddress(const struct sockaddr *generic_addr, socklen_t addrLen) OME_ALWAYS_INLINE {
        // we can only hold so much...so don't try to copy more
        if (addrLen > sizeof(socketAddress)) {
            addressLength = sizeof(socketAddress);
        } else {
            addressLength = (uint8_t) addrLen;
        }
        memcpy(&socketAddress.genericAddr, generic_addr, addressLength);
    }

    GenericSocketAddress(const char *address, uint16_t port) {
        setFromHostAndPort(address, port);
    }

    GenericSocketAddress(const std::string &address, uint16_t port) {
        setFromHostAndPort(address.c_str(), port);
    }

    GenericSocketAddress &operator=(const struct sockaddr_in ipv4) OME_ALWAYS_INLINE {
        socketAddress.ipv4 = ipv4;
        addressLength = sizeof(struct sockaddr_in);
        return (*this);
    }

    GenericSocketAddress &operator=(const struct sockaddr_in6 ipv6) OME_ALWAYS_INLINE {
        socketAddress.ipv6 = ipv6;
        addressLength = sizeof(struct sockaddr_in6);
        return (*this);
    }

    /*! \brief Return the address as a URL.
     *
     * \param addrText points to the buffer into which the
     * textual representation of the address will be constructed.
     * \param bfrSize specifies the size of the buffer.
     * \param includeURLprefix is a Boolean that indicates
     * if a type-specific URL prefix, such as "ipv4://", should be
     * included before the address.
     *
     * \return The port will be returned as the integer result.
     */
    int copyAddressAndPort(char *addrText, uint32_t bfrSize, bool includeURLprefix=false) const {
        int port;

        if (bfrSize < 8) { // answer won't fit
            addrText[0] = '\0';
            return (-1);
        }

        /* Windows has flawed declarations which do not include
         * the const attribute for the incoming address
         */
        uint_fast8_t offset = 0;
        int32_t bfrLeft = (int32_t) bfrSize;
        switch (socketAddress.genericAddr.sa_family) {
        case AF_INET:
            if (includeURLprefix == true) {
                memcpy(addrText, "ipv4://", 7);
                offset += 7;
                bfrLeft -= 7;
                if (bfrLeft < 0) {
                    bfrLeft = 0;
                }
            }
            inet_ntop(AF_INET, const_cast<struct in_addr *>(&socketAddress.ipv4.sin_addr), addrText + offset, bfrLeft);
            port = ntohs(socketAddress.ipv4.sin_port);
            break;
        case AF_INET6:
            if (includeURLprefix == true) {
                memcpy(addrText, "ipv6://", 7);
                offset += 7;
                bfrLeft -= 7;
                if (bfrLeft < 0) {
                    bfrLeft = 0;
                }
            }
            inet_ntop(AF_INET6, const_cast<struct in6_addr *>(&socketAddress.ipv6.sin6_addr), addrText + offset, bfrLeft);
            port = ntohs(socketAddress.ipv6.sin6_port);
            break;
#ifndef _WIN32
        case AF_LOCAL:
            if (includeURLprefix == true) {
                memcpy(addrText, "file://", 7);
                offset += 7;
                bfrLeft -= 7;
                if (bfrLeft < 0) {
                    bfrLeft = 0;
                }
            }
            safe_strcpy(addrText + offset, socketAddress.filename.sun_path, bfrLeft);
            return (0); // there is no port with this address type
#endif
        default:
#ifndef _WIN32
            snprintf
#else
            _snprintf
#endif
                (addrText, bfrSize, "unrecognizedAF%u", socketAddress.genericAddr.sa_family);
            port = 0;
            break;
        }
        return (port);
    }

    /*! \brief Convenience cover to copyAddressAndPort() that
     * concatenates the port to the result.
     *
     * \param bfr points to the buffer into which the
     * textual representation of the address and port will be constructed.
     * \param bfrSize specifies the size of the buffer.
     * \param includeURLprefix is a Boolean that indicates
     * if a type-specific URL prefix, such as "ipv4://", should be
     * included before the address.
     *
     * \return A pointer to the result string is returned.
     */
    char *getAddressAndPort(char *bfr, uint32_t bfrSize, bool includeURLprefix=false) const {
        char addrText[64];
        int port = copyAddressAndPort(addrText, sizeof(addrText), includeURLprefix);

#ifndef _WIN32
        snprintf(bfr, bfrSize, "%s:%d", addrText, port);
#else
        _snprintf(bfr, bfrSize, "%s:%d", addrText, port);
#endif
        return (bfr);
    }

    bool operator==(const GenericSocketAddress &arg) const OME_ALWAYS_INLINE {
        if (OME_EXPECT_FALSE(addressLength != arg.addressLength)) {
            return (false);
        }
        int rc = memcmp(&socketAddress, &arg.socketAddress, addressLength);
        return (rc == 0);
    }

    bool operator!=(const GenericSocketAddress &arg) const OME_ALWAYS_INLINE {
        if (OME_EXPECT_FALSE(addressLength != arg.addressLength)) {
            return (true);
        }
        int rc = memcmp(&socketAddress, &arg.socketAddress, addressLength);
        return (rc != 0);
    }

    bool operator<(const GenericSocketAddress &arg) const OME_ALWAYS_INLINE {
        int rc;
        // rarely will we compare addresses from different families
        if (OME_EXPECT_FALSE(addressLength < arg.addressLength)) {
            rc = memcmp(&socketAddress, &arg.socketAddress, addressLength);
            if (rc <= 0) { // earlier, or matches everything but was shorter
                return (true);
            }
            return (false);
        }
        // argument (right hand side) length is same or shorter
        rc = memcmp(&socketAddress, &arg.socketAddress, arg.addressLength);
        if (rc < 0) {
            return (true);
        }
        return (false);
    }

    bool operator>(const GenericSocketAddress &arg) const OME_ALWAYS_INLINE {
        int rc;
        // normally we will compare addresses from the same family
        if (OME_EXPECT_TRUE(addressLength <= arg.addressLength)) {
            rc = memcmp(&socketAddress, &arg.socketAddress, addressLength);
            if (rc <= 0) { // earlier, or matches to our possibly shorter length
                return (false);
            }
            return (true);
        }
        // argument (right hand side) is shorter
        rc = memcmp(&socketAddress, &arg.socketAddress, arg.addressLength);
        if (rc < 0) { // earlier
            return (false);
        }
        // above or matches to their shorter length
        return (true);
    }

}; // end class GenericSocketAddress

template <typename STREAMTYPE> inline STREAMTYPE & operator<<(STREAMTYPE &os, const GenericSocketAddress &arg)
{
    os << "[GenericSocketAddress={";
    os << " address=\"" << arg.socketAddress.genericAddr << "\"";
//    os << " addressLength=" << arg.addressLength;
    os << " }]";
    return (os);
}


#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"

/*! \brief Per-session HTTP state record.
 *
 * Custom applications are realized by implementing a subclass of
 * HTTP_SessionRecord to handle the per-session behavior.
 * The most trivial implementation will make an addHTTPhandler() call
 * in the constructor and implement the corresponding method.
 */
class HTTP_SessionRecord : public Parse_And_Process_HTTP_Stream
{
public:
    struct ltCharPtrStr {
        bool operator()(const char *arg1, const char *arg2) const {
            int rc = strcmp(arg1, arg2);
            if (rc < 0) return (true);
            return (false);
        }
    }; // end struct ltCharPtrStr

    /*! \brief Signature for HTTP callback routine.
     */
    typedef ssize_t (*HTTPcallbackFP)(HTTP_SessionRecord *session,
            const char *cmd, char *url,
            unsigned char *block, size_t blockLen);

    typedef std::map<const char *, const char *, ltCharPtrStr> OptionsAttrValMap_t;

    IO_Processor            *controller;
    CircularBufferManager   *bfrMgr;
    void                    *userData;
    GenericSocketAddress    fromAddress;
    bool                    closeWhenRequestComplete;
protected:
    std::map<std::string,HTTPcallbackFP>    httpRequestHandlerTable;

protected:
    /*! \brief User-exit called when HTTP request read.
     * Not normally overridden.
     *
     * Default implementation will invoke the callback previously
     * registered by addHTTPhandler() that corresponds to
     * the incoming command.
     */
    virtual ssize_t handleHTTPrequest(unsigned char *block, size_t blockLen);

    /*! \brief Close the connection associated with this HTTP session.
     */
    void closeConnection();

    // NOTE: not static due to keep-alive state
    int_fast32_t prepareResponseHeader(void *outputBfr, size_t bfrLen,
            uint_fast32_t statusCodeValue=200, const char *statusText="OK",
            const char *extraHeaders="",
            ssize_t contentLength=0, const char *extraContent="");

    ssize_t sendResponse(const void *bfr, size_t bfrLen);
    // TODO: provide iovec equivalent

public:
    /*! \brief Create state record for current HTTP session.
     *
     * \param io points to the IO_Processor object controlling
     * the stream processing.  It should be deleted by
     * some parent object.
     * \param mgr specifies the circular buffer manager used
     * to buffer the input stream.  Ownership is transferred to
     * the HTTP_SessionRecord object.
     * \param from specifies the source of the incoming connection.
     * \param  extraData allows application-specific data to be passed.
     * \param maxDocSize specifies the maximum size of an HTTP document
     * to be pased to the Parse_And_Process_HTTP_Stream base class.
     */
    HTTP_SessionRecord(IO_Processor *io, CircularBufferManager *mgr,
        const GenericSocketAddress *from, void *extraData = nullptr,
        size_t maxDocSize=1024*128) :
        Parse_And_Process_HTTP_Stream(maxDocSize, Parse_And_Process_HTTP_Stream::READ_REQUEST),
        fromAddress(*from)
    {
        controller = io;
        bfrMgr = mgr;
        userData = extraData;
        closeWhenRequestComplete = true;
    }

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    ~HTTP_SessionRecord() {
        // controller is eventually deleted by HTTPembeddedServerBase::startHTTPacceptThread()
        delete bfrMgr;
    }
#pragma GCC diagnostic pop

    /* \brief Convert %XX hex escape into the appropriate character,
     * possibly shrinking the result string in place.
     *
     * \param bfr holds the string to be converted.  It may be modified in
     * place, if necessary.
     *
     * \return The end of the converted string is returned.
     */
    static char *convertHTTPescapes(char *bfr, int_fast32_t bfrLen=-1,
            bool nullTerminate=false);

    /*! \brief Parse an HTTP request option string.  Elements are separated
     * by "&" characters; each subelement is expected to be a "var=value"
     * pair.
     *
     * \param tbl is an attribute/value map into which the parsed attribute
     * value pairs will be stored.  Its key and value pointers will refer to
     * addresses within \a optionText, so the storage associated with
     * \a optionText must not be deleted while * the \a tbl map is accessed.
     * \param  optionText is a string of HTML option text; if the length is
     * specified, then the string content does not have to be terminated by
     * a null character.
     * \param optionLength is an optional argument that specifies the length
     * of optionText.  If set to -1, the length will be determined at runtime
     * using strlen().
     *
     * NOTE: \a optionText will be modified.
     *
     * \return The number of elements parsed is returned.
     */
    static int_fast32_t parseOptions(OptionsAttrValMap_t *tbl, char *optionText,
            int_fast32_t optionLength=-1);

    /*! \brief Add a handler routine for an HTTP request command.
     *
     * \param cmd specifies the request to be handled, such as "GET" or "POST".
     * \param routine specifies the routine to be invoked.
     */
    void addHTTPhandler(const char *cmd, HTTPcallbackFP routine) {
        httpRequestHandlerTable[cmd] = routine;
    }

    /*! \brief Set close-when-request-complete flag.
     *
     * \param doClose specifies the desired state.
     *
     * By default, if a requesting client provides a "Connection: keep-alive"
     * header, the connection will remain open after a request is processed.
     * The connection can be forced to be closed by passing a value of true.
     */
    void setCloseAfterRequestFlag(bool doClose) {
        closeWhenRequestComplete = doClose;
    }

    /*! \brief User-exit called when EOF reached on client connection
     *
     * The default implementation calls closeConnection() to clean up.
     */
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    virtual void connectionEOF() {
        LOG_COMPONENT_CERR(app,info) << "EOF from client " << fromAddress.socketAddress.genericAddr << LOG_ENDLINE;
        closeConnection();
    }
#pragma GCC diagnostic pop

    /*! \brief User-exit for receiving POSTed data.
     *
     * \param blockStart points to start of data
     * \param blockLen indicates length of provided data
     *
     * This overrides 
     */
    virtual int_fast32_t processBlock(unsigned char *blockStart, size_t blockLen) VIRTUAL_OVERRIDE {
        LOG_COMPONENT_CERR(app,info) << "from=" << fromAddress.socketAddress.genericAddr << " processBlock, len=" << blockLen << " content=" << AS_TEXT_BUFFER((const char *) blockStart, blockLen) << LOG_ENDLINE;
        int_fast32_t rc = handleHTTPrequest(blockStart, blockLen);
        return (rc);
    }

    /*! \brief User exit to process incoming HTTP header.
     *
     * Extends the underlying Parse_And_Process_HTTP_Stream routine by
     * invoking handleHTTPrequest() if the end of the headers is seen and
     * no subsequent content is to be sent.
     */
    virtual int processHTTPheader(unsigned char *line, size_t lineLen) VIRTUAL_OVERRIDE {
        if (line[lineLen - 1] == '\n') {
            lineLen -= 1;
            if (line[lineLen - 1] == '\r') {
                lineLen -= 1;
            }
        }
        if (lineLen == 0) { // end of HTTP headers
            if (contentLengthWanted <= 0) {
                // TODO: 
                LOG_COMPONENT_CERR(app,info) << "Saw end of HTTP headers from=\"" << fromAddress.socketAddress.genericAddr << "\" Process incoming request" << LOG_ENDLINE;
                int rc = handleHTTPrequest(nullptr, 0);
                return (rc);
            }
            LOG_COMPONENT_CERR(app,info) << "Saw end of HTTP headers from=\"" << fromAddress.socketAddress.genericAddr << "\" Content-Length: " << contentLengthWanted << LOG_ENDLINE;
            // remaining block of data will be read and passed to processBlock()
            return (0);
        }
        int rc = Parse_And_Process_HTTP_Stream::processHTTPheader(line, lineLen);
        return (rc);
    }

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    /*! \brief User-exit to filter out events when session has been converted
     * into use with an HTTPstatusLog object.
     */
    virtual bool interestedInEvent(const class HTTPeventStateRecord *record) {
        return (true);
    }
#pragma GCC diagnostic push

    /*! \brief Default implementation to handle OPTIONS request.
     * It emits Access-Control-Allow headers, assert GET, POST permitted
     */
    static int default_OPTIONS(HTTP_SessionRecord *session, const char *cmd,
            char *url, unsigned char *block, size_t blockLen);

}; // end class HTTP_SessionRecord

#pragma GCC diagnostic pop

template <typename STREAMTYPE> inline STREAMTYPE & operator<<(STREAMTYPE &os, const HTTP_SessionRecord &arg)
{
    os << "[HTTP_SessionRecord={";
    os << " fromAddr=" << arg.fromAddress;
    os << " closeWhenComplete=" << arg.closeWhenRequestComplete;
    os << " }]";
    return (os);
}

/*! \brief Core embedded HTTP server logic.
 *
 * Invokes the abstract function acceptNewConnection() to obtain desired
 * per-session application logic.
 *
 * The templatized HTTPembeddedServer<> class provides a convenient typed
 * interface and is how most user implementations create the HTTP server.
 */
class HTTPembeddedServerBase {
protected:
    char        serverName[64];
#ifdef _WIN32
    HANDLE      acceptThreadID;
#else
    pthread_t   acceptThreadID;
#endif

public:
    void        *userData;
    OS_SOCKET_TYPE  listenDescriptor;
    uint32_t    bufferSize;
    uint16_t    maxHTTPeventLength;
    uint16_t    listenPort;
    bool        stopAccepting;
    bool        listenActive;

protected:
    static int readHTTPstream(SharedBufferAllocRecord *rec, IO_Processor *controller);
#ifdef WIN32
    static DWORD WINAPI startHTTPacceptThread(void *arg);
#else
    static void *startHTTPacceptThread(void *arg);
#endif

public:
    /*! \brief Configures the HTTP server for listening on the indicated
     * port.
     *
     * \param port specifies the TCP port onto which incoming connections
     * will be accepted.
     * \param extraData will be passed to each of the HTTP_SessionRecord
     * objects that are created.
     * \param useServerName allows an alternate server name to be specified;
     * the default uses the host's name.
     * \param useBfrSize allows the buffer size to be specified.  If 0,
     * which is the default, the buffer size will be determined based
     * on the maximum event size.
     * \param maxEventLength allows the maximum size of a single event
     * to be specified.  If 0, the default size will be used.
     *
     * NOTE: the server is not started until startHTTPlisten() is called.
     */
    HTTPembeddedServerBase(uint16_t port, void *extraData=nullptr,
        const char *useServerName=nullptr, uint32_t useBfrSize=0,
        uint16_t maxEventLength=0)
    {
        listenPort = port;
        userData = extraData;
        bufferSize = useBfrSize;
        maxHTTPeventLength = maxEventLength;
        stopAccepting = false;
        listenActive = false;
        if (useServerName == nullptr) { // none specified
#ifndef _WIN32
            gethostname(serverName, sizeof(serverName));
#else
            DWORD nameLen = sizeof(serverName);
            GetComputerName(serverName, &nameLen);
#endif

            char *cp = serverName;
            while (*cp != '\0') {
                if (*cp == '.') { // terminate on first "."
                    *cp = '\0';
                    break;
                }
                if (isupper(*cp)) {
                    *cp = tolower(*cp);
                }
                cp += 1;
            }
        } else {
            safe_strcpy(serverName, useServerName, sizeof(serverName));
        }
    }

    virtual ~HTTPembeddedServerBase()   {}

    /*! \brief Alter buffer size used for new accepted client connections.
     */
    void setBufferSize(uint32_t bfrSize) {
        bufferSize = bfrSize;
    }

    /*! \brief Alter maximum length of an HTTP event used for new accepted connections.
     */
    void setHTTPeventLength(uint32_t maxLen) {
        maxHTTPeventLength = maxLen;
    }

    /*! \brief Returns the name associated with the HTTP server
     */
    const char *getServerName() const {
        return (serverName);
    }

    /*! \brief Start the HTTP server.
     *
     * \retval 0 indicates the server started successfully
     * \retval -1 indicates the server failed to start, usually because
     * the desired port is already in use.
     */
    int startHTTPlisten();

    /*! \brief Stop the HTTP server.
     */
    int stopHTTPlisten(const char *reason);

    /*! \brief Wait for server to cease operations.
     */
    int waitForHTTPstop();

    /*! After the HTTP server has accepted an incoming connection,
     * a new session object is created via a call to acceptNewConnection().
     *
     * NOTE: this is an abstract function which must be implemented by the
     * subclass realizing the server implementation.
     */
    virtual HTTP_SessionRecord *acceptNewConnection(IO_Processor *io, CircularBufferManager *mgr, const GenericSocketAddress *from, void *extraData) = 0;

}; // end class HTTPembeddedServerBase

/*! \brief Templatized embedded HTTP server interface; realizes a complete
 * implementation of HTTPembeddedServerBase.
 *
 *  ACCEPT_CLASS must be a subclass of HTTP_SessionRecord
 *  USER_CLASS is the type of any optional data to be provided to the new
 *  session.
 */
template <typename ACCEPT_CLASS=HTTP_SessionRecord, typename USER_CLASS=void> class HTTPembeddedServer : public HTTPembeddedServerBase
{
public:
    HTTPembeddedServer(uint16_t port, USER_CLASS *extraData=0,
        const char *useServerName=0, uint32_t useBfrSize=0,
        uint16_t maxEventLen=0) :
        HTTPembeddedServerBase(port, extraData, useServerName, useBfrSize, maxEventLen) {
    }

    virtual ~HTTPembeddedServer()   {}

    virtual HTTP_SessionRecord *acceptNewConnection(IO_Processor *io, CircularBufferManager *mgr, const GenericSocketAddress *from, void *extraData) {
        ACCEPT_CLASS *sessionRec = new ACCEPT_CLASS(io, mgr, from,
                reinterpret_cast<USER_CLASS *>(extraData));
        // NOTE: ACCEPT_CLASS pointer will be demoted to base class pointer
        return (sessionRec);
    }

    /*! \brief Returns the user-specific data as a typed value.
     */
    USER_CLASS *getUserData() const {
        return (reinterpret_cast<USER_CLASS *>(userData));
    }
}; // end class HTTPembeddedServer<>


/* NOTE: the remaining classes support sending streams of events via
 * HTTP to a client making use of an XMLHttpRequest() object.
 */

/*! \brief Internal record to describe events to be published via
 * HTTP.
 */
class HTTPeventStateRecord
{
public:
    enum { MAX_STATUS_LINE_LENGTH = 448 };
    const void  *extraData;
    uint64_t    updateTime;
    uint32_t    lineLength;
    uint32_t    eventRecordId;
    char        eventType[24];
    char        lineData[MAX_STATUS_LINE_LENGTH];

    explicit HTTPeventStateRecord(uint32_t id=0) {
        extraData = nullptr;
        updateTime = 0;
        lineLength = 0;
        eventRecordId = id;
    }
}; // end class HTTPeventStateRecord

template <typename STREAMTYPE> inline STREAMTYPE & operator<<(STREAMTYPE &os, const HTTPeventStateRecord &arg)
{
    os << "[HTTPeventStateRecord={";
    os << " updateTime=" << arg.updateTime;
    os << " eventType=\"" << arg.eventType << "\"";
    os << " eventId=" << arg.eventRecordId;
    os << " line=\"" << AS_ESCAPED_TEXT_BUFFER(arg.lineData, arg.lineLength) << "\"";
    os << " }]";
    return (os);
}

/*! \brief Key structure for imposing a partial order on HTTP status events.
 */
struct HTTPstatusRecordKey {
    const void  *object;
    char        eventType[32];
};

/*! \brief Implements Less-than for HTTPstatusRecordKey
 */
struct ltHTTPstatusRecordKey
{
    bool operator()(const HTTPstatusRecordKey &arg1, const HTTPstatusRecordKey &arg2) const {
        if (arg1.object < arg2.object) return (true);
        if (arg1.object > arg2.object) return (false);
        // must be same
        int rc = strcmp(arg1.eventType, arg2.eventType);
        if (rc < 0) return (true);
        return (false);
    }
}; // end struct ltHTTPstatusRecordKey

/*! \brief Abstract interface to receive log message for transmission
 * to HTTP-attached clients.
 *
 * NOTE:  when inheriting, this should be specified as a virtual
 * base class to avoid any ambiguity when merging several derived
 * classes via multiple inheritence.
 */
class HTTPnotificationReceiverInterface
{
public:
    HTTPnotificationReceiverInterface()  {}

    ~HTTPnotificationReceiverInterface()  {}

    /*! \brief Abstract interface to deliver formatted message
     * \param bfr points to the start of the message content
     * \param bfrLen indicates the length of the message
     * \param eventType is normally "log"
     */
    virtual void receiveNotice(const unsigned char *bfr, uint32_t bfrLen,
            const char *eventType="log") = 0;
}; // end class HTTPnotificationReceiverInterface

/*! \brief Base class for converting a LogManager into a publisher
 * that outputs log event messages to multiple attached clients.
 *
 * The actual interesting work must be done by a derived class of
 * HTTPnotificationReceiverInterface. This class really provides
 * very generic functionality and can be exploited for other purposes.
 */
class HTTPnotificationPublisher : public LogManager
{
protected:
    std::set<HTTPnotificationReceiverInterface *> interestedClients;
public:
    static int forwardToClients(SharedBufferAllocRecord *rec, IO_Processor *controller);
    SharedMemoryVariableNode    namingNode;

    explicit HTTPnotificationPublisher(const char *name) :
#ifndef _WIN32
        LogManager(0, -1, 0, name,
            LogMessageRecord::LOG_TIMESTAMP | LogMessageRecord::LOG_ERROR_LEVEL),
#else
        LogManager(0, INVALID_HANDLE_VALUE, 0, name,
            LogMessageRecord::LOG_TIMESTAMP | LogMessageRecord::LOG_ERROR_LEVEL),
#endif
        namingNode(name)
    {
        // replace ioMgr processing routine with new one
        // that forwards to interested clients.
        // NOTE: bfrMgr is from the base class
        ioMgr = new IO_Processor(&namingNode, bfrMgr, forwardToClients,
#ifndef _WIN32
                -1,
#else
                INVALID_HANDLE_VALUE,
#endif
                IO_Processor::PROCESS_DURING_READ, this);
    }

    virtual ~HTTPnotificationPublisher() {}

    void addInterestedClient(HTTPnotificationReceiverInterface *client) {
        interestedClients.insert(client);
    }
}; // end HTTPnotificationPublisher


/*! \brief Implements HTTPnotificationReceiverInterface to maintain a
 * history of events and log lines which are to be published via HTTP.
 */
// TODO: could templatize on max lines of history or make lineContent a vector
class HTTPstatusLog : public virtual HTTPnotificationReceiverInterface
{
protected:
    enum { MAX_LINES_HISTORY = 8192 };
    struct {
        const unsigned char *iov_base;
        size_t              iov_len;
        char                eventType[24];
    }                                   lineContent[MAX_LINES_HISTORY];
    time_t                              lineAddedAtTime[MAX_LINES_HISTORY];
    uint32_t                            linesAdded;
    uint32_t                            eventRecordsCreated;
    mutable TimedMutex                  sessionMutex;
    std::list<HTTP_SessionRecord *>     currentHTTPsessions;
    typedef std::map<HTTPstatusRecordKey,HTTPeventStateRecord,ltHTTPstatusRecordKey>  statusEventMap_t;
    statusEventMap_t                    statusEvents;

protected:
    /*! \brief IO_Processor-compatible processing routine used to forward
     * buffer contents to a connect client
     */
    static int forwardBuffer(SharedBufferAllocRecord *rec, IO_Processor *controller);

    int dumpLinesFromTime(HTTP_SessionRecord *rec, uint64_t fromTime);

    int dumpEventsFromTime(HTTP_SessionRecord *rec, uint64_t fromTime);

public:
    /*! \brief Static convenience routine to take over an existing HTTP
     * session and turn it into an event forwarder connected to the specified
     * HTTPstatusLog object.
     *
     * NOTE: this routine does not return until the remote client closes the
     * connection.
     */
    static void becomeEventStream(HTTP_SessionRecord *session, HTTPstatusLog *statusLog);

    HTTPstatusLog() :
        sessionMutex("HTTPstatusLog")
    {
        linesAdded = 0;
        eventRecordsCreated = 0;
        for(uint32_t i=0;i<MAX_LINES_HISTORY;i+=1) {
            lineContent[i].iov_base = nullptr;
        }
    }

    virtual ~HTTPstatusLog() {}

    /* NOTE: this implementation uses the hijacked IO_Processor. It
     * generates a copy of the formatted data into a buffer record and
     * notifies the processing thread of its availability.
     * \return the number of bytes processed.
     */
    int emitLineAsHTTPevent(IO_Processor *controller, uint64_t eventId,
            const char *eventType,
            const unsigned char *lineData, uint32_t lineLen,
            const char *additional);

    int emitLogLineAsHTTPevent(IO_Processor *controller, uint32_t relLine, const char *additional)
    {
        const unsigned char *data = const_cast<unsigned char *>(lineContent[relLine].iov_base);
        uint32_t lineLen = (uint32_t) lineContent[relLine].iov_len;
        int rc = emitLineAsHTTPevent(controller, relLine,
                lineContent[relLine].eventType,
                data, lineLen, additional);
        return (rc);
    }

    void addHTTPsession(HTTP_SessionRecord *rec, uint64_t fromTime=0);

    void removeHTTPsession(HTTP_SessionRecord *rec);

    virtual void receiveNotice(const unsigned char *bfr, uint32_t bfrLen,
        const char *eventType="log") override;

    /*! \brief Find an event state record for the indicated object
     * and event type.
     */
    const HTTPeventStateRecord *locateEventRecord(const void *obj, const char *eventType) const;

    /*! \brief Find an event state record for the indicated object
     * and event type.
     *
     * NOTE: this routine always returns a result; if no record currenty
     * exists, then a new one will created.
     */
    HTTPeventStateRecord *findOrCreateEventRecord(const void *obj, const char *eventType);

    void publishEventState(const HTTPeventStateRecord *rec);

    void publishKeepAlive();

}; // end class HTTPstatusLog


/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
