#ifndef _IO_PROCESSOR_HPP_
//! \cond
#define _IO_PROCESSOR_HPP_ "$Id: io_processor.hpp 455 2020-07-23 20:23:59Z geoff $"
//! \endcond

/* Copyright (C) 2010 - 2019 FARGOS Development, LLC
 * Author: Geoff Carpenter - http://www.fargos.net/gcc.html
 */

#include <utils/base/compiler_hints.h>
#include <utils/base/timed_mutex.hpp>
#include <utils/base/circular_bfr.hpp>
#include <utils/shared_vars/shared_variable.hpp>

/*! \file
 * \brief FARGOS I/O Processing classes
 *
 * The streaming infrastructure implemented via the IO_Processor
 * can be used in a variety of scenarios and in single-threaded mode
 * or with distinct threads handling the input and processing.
 *
 * \author Geoff Carpenter
 * \copyright 2010 - 2019 FARGOS Development, LLC
 */

/** \defgroup FARGOSio FARGOS I/O Processing Classes
 *
 * \brief The IO_Processor class is the driver of a suite of
 * classes used to process streams of data.  The OrderedInputSource class
 * and its related classes add the ability to handle multiple streams
 * of typically ordered-by-time streams data.
 * The function processPacketFromSourceUsingClass() enables
 * convenient processing of a stream of data by an object.
 * @{
 */

#ifndef OS_SOCKET_TYPE
#ifdef _WIN32
#define OS_SOCKET_TYPE SOCKET
#else
#define OS_SOCKET_TYPE int
#endif
#endif /* OS_SOCKET_TYPE */
#ifndef OS_HANDLE_TYPE
#ifdef _WIN32
#define OS_HANDLE_TYPE HANDLE
#else
#define OS_HANDLE_TYPE int
#endif
#endif /* OS_HANDLE_TYPE */
#ifndef INVALID_SOCKET
#ifdef _WIN32
#define INVALID_SOCKET ((OS_SOCKET_TYPE)(~0))
#else
#define INVALID_SOCKET (-1)
#endif
#endif /* INVALID_SOCKET */


/*! \brief Process a block of data received from an input source.
 *
 * \note The processBlock routine should update the bytesProcessed counter
 * of the statistics block.
 * \return an indicator if the block should be reclaimed or preserved.
 * \retval 0 indicates that the block can be freed/reused.  If 0 is not
 * returned, then some provision must be made to reclaim the buffer block
 * at a future point in time.
 */
typedef int (*IO_processBlockFP)(SharedBufferAllocRecord *rec,
    class IO_Processor *controller);

/*! \brief The receiveBlock routine should retrieve data and store it into
 * the indicated \c block of memory, constrained by the \c blockLen.
 *
 * \return The number of bytes read is returned.
 * \retval 0 indicates end-of-file.
 * \retval -1 indicates an error.
 */
typedef ssize_t (*IO_receiveBlockFP)(SharedBufferAllocRecord *rec,
    class IO_Processor *controller);

/*! \brief Long running routine which receives all input from
 * input source.
 *
 * \retval 0 should be returned as the value is currently ignored.
 */
typedef int (*IO_consumeFP)(class IO_Processor *controller);

/*! \brief Long running routine which processes all input made available
 * by the corresponding consume routine.
 *
 * \retval 0 should be returned as the value is currently ignored.
 */
typedef int (*IO_processFP)(class IO_Processor *controller);

typedef int (*IO_blockAllocFP)(uint32_t blockLen,
    class IO_Processor *controller);

typedef int (*IO_blockFreeFP)(unsigned char *block,
    class IO_Processor *controller);

/*! \brief Structure to hold meta-data related to a block of received data.
 */
struct IO_metaBlock_header {
    uint16_t        headerLen;              //!< length of the header
    uint16_t        fromLen;                //!< length of \c fromAddress
//! \cond
    char            _pad1[12];              //!< padding
//! \endcond
    uint64_t        receiveTimeSeconds;     //!< receive time in seconds
    uint64_t        receiveTimeNanoseconds; //!< receive time fractional nanoseconds
    char            fromAddress[32];        //!< sockaddr_t of source
};


/*! \brief Holds collected statistics for an IO_Processor object.
 */
class IO_Processor_Statistics {
public:
    SharedMemoryVariableNode        namingNode;
    // normally exposed as statistics variables
    SMV_StandaloneNumeric<uint64_t> bytesRead;          // updated by consume routine
    SMV_StandaloneNumeric<uint64_t> bytesProcessed;     // updated by processBlock routine
    // normally exposed as statistics variables
    SMV_StandaloneNumeric<uint32_t> packetsRead;        // updated by consume routine

    SMV_StandaloneNumeric<uint32_t> packetsProcessed;   // updated by process routine
    SMV_StandaloneNumeric<uint32_t> productiveReadSpins;// updated by recvRoutine
    SMV_StandaloneNumeric<uint32_t> maxReadTimeout;     // adjustable, used by recvRoutine
    //! max unproductive read attempts before thread blocks
    SMV_StandaloneNumeric<uint32_t> readAttemptsBeforeBlocking; // adjustable, used by recvRoutine

    IO_Processor_Statistics(SharedMemoryVariableNode *parentNode) :
        namingNode("io_stats", parentNode),
        bytesRead("bytesRead", &namingNode),
        bytesProcessed("bytesProcessed", &namingNode),
        packetsRead("packetsRead", &namingNode), 
        packetsProcessed("packetsProcessed", &namingNode),
        productiveReadSpins("productiveReadSpins", &namingNode),
        maxReadTimeout("maxReadTimeout", &namingNode),
        readAttemptsBeforeBlocking("readsBeforeBlock", &namingNode)
    {
    }

    ~IO_Processor_Statistics()  {}
};

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"

/*! \brief Intermediary I/O processing object for performing multi-threaded
 * receive-and-process operations on a flow of messages.
 */
class IO_Processor {
public:
    //! \brief Mask to select threading modes
    enum ThreadMode {
        NONE=0,                 //!< No threads will be spawned
        READ_THREAD=1,          //!< A read thread will be spawned
        PROCESS_THREAD=2,       //!< A processing thread will be spawned
        //! Processing will be performed by the read thread.
        PROCESS_DURING_READ=4,
        //! Both a read and a processing thread will be spawned
        BOTH_THREADS=READ_THREAD | PROCESS_THREAD,
        SEPARATE_READ_AND_PROCESS_THREADS=BOTH_THREADS,
        //! a single spawned read thread will do both read-and-processing
        READ_AND_PROCESS_ON_SAME_THREAD=READ_THREAD | PROCESS_DURING_READ 
    };

    //! Blocking mode
    enum BlockMode {
        //! packet message unit, like a datagram
        PACKET=1,
        //! bit flag indicating source information
        _PACKET_HAS_SOURCE=2,
        //! bit flag indicating arrival time
        _PACKET_HAS_RECEIVE_TIME=4,
        //! bit flags indicating meta data present
        _PACKET_HAS_META_DATA=_PACKET_HAS_SOURCE | _PACKET_HAS_RECEIVE_TIME,
        //! force use of read vs. recv().
        _NOT_SOCKET=128,
        //! contiguous byte stream from socket
        CONTIGUOUS_BYTE_STREAM=0,
        //! contiguous byte stream from file
        CONTIGUOUS_FILE_STREAM=_NOT_SOCKET,
        //! packet message unit including source address
        PACKET_WITH_SOURCE=PACKET | _PACKET_HAS_SOURCE,
        //! packet message with arrival time 
        PACKET_WITH_TIME=PACKET | _PACKET_HAS_RECEIVE_TIME,
        //! packet message with source address and arrival time
        PACKET_WITH_SOURCE_AND_TIME=PACKET | _PACKET_HAS_SOURCE | _PACKET_HAS_RECEIVE_TIME
    };

    BufferRegion            *bfrManager;             //!< buffer region
    IO_Processor_Statistics *statistics;    //!< statistics
    TimedMutex              *mutex;
    TimedCondition          *condition;
    IO_processFP            processRoutine;
    IO_processBlockFP       processBlockRoutine;
    IO_consumeFP            consumeRoutine;
    IO_receiveBlockFP       recvRoutine;
    void                    *extraData;         //!< arbitrary extra data
#ifdef _WIN32
    HANDLE                  consumeThreadID;
    HANDLE                  processThreadID;
#else
    pthread_t               consumeThreadID;
    pthread_t               processThreadID;
#endif
    OS_HANDLE_TYPE          descriptor;
    uint32_t                descriptorFlags;
    int32_t                 packetsProcessedIncrement; // signed
    char                    label[24];
    size_t                  maxPacketSize;  //!< max to receive in one go
    unsigned int            lastReadTimeout;
    unsigned char           threadStartedState;
    unsigned char           currentThreadState;
    unsigned char           desiredThreadState;
    unsigned char           stopRequested;
    unsigned char           blockingMode;

    

    /*! Normal receive/consume routine suitable for handling datagram and stream
     * sockets as well as files.
     */
    static ssize_t recvConsume(SharedBufferAllocRecord *rec,
        class IO_Processor *controller);

    /*! \brief Standard consume loop to receive incoming data.
     *
     * \retval 0 is always returned; this is subject to change.
     */
    static int doConsumeLoop(IO_Processor *controller);

    /*! \brief Standard processing loop to process data on separate thread;
     * works in conjunction with doConsumeLoop().
     *
     * \retval 0 is always returned; this is subject to change.
     */
    static int doProcessLoop(IO_Processor *controller);

    /*! \brief Submission routine that will either directly invoke
     * the processing routine or notify the processing thread
     * that it is available.
     *
     * Not normally used directly, but available for special use cases.
     */
    int submitOrProcessBlock(SharedBufferAllocRecord *rec);


    //! Constructor for IO_Processor
    IO_Processor(SharedMemoryVariableNode *parentNode,
        BufferRegion *mgr, IO_processBlockFP b_func,
        OS_HANDLE_TYPE h,
// COMMENT:  maybe these should be flipped in order...
        ThreadMode tMode=NONE, void *userData=nullptr, BlockMode bMode=PACKET,
        IO_receiveBlockFP r_func=recvConsume,
        IO_consumeFP c_func=doConsumeLoop,
        IO_processFP p_func=doProcessLoop);

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    virtual ~IO_Processor();
#pragma GCC diagnostic pop

    //! Return number of bytes reserved for meta data on each data element.
    inline size_t dataOffset() const OME_ALWAYS_INLINE {
        return ((blockingMode & _PACKET_HAS_META_DATA) ? sizeof(IO_metaBlock_header) : 0);
    }

    /*! \brief Return physical address of a buffer within the context of the
     * local process' address space.
     */
    inline unsigned char *bufferAddress(SharedBufferAllocRecord *rec,
        size_t *bufferLen=nullptr) const OME_ALWAYS_INLINE {
        const size_t data_off = dataOffset();
        if (OME_EXPECT_TRUE(bufferLen != nullptr)) { // length is wanted...
            // EOF will have usedLen == 0
            *bufferLen = OME_EXPECT_TRUE((rec->usedLen >= data_off)) ? (rec->usedLen - data_off) : 0;
        }
        return (bfrManager->blockAddress(rec) + data_off);
    }

    /*! \brief Return physical address of meta block header within the context
     * of the local process' address space.
     */
    inline IO_metaBlock_header *bufferHeaderAddress(SharedBufferAllocRecord *rec,
        size_t *headerLen=nullptr) const OME_ALWAYS_INLINE {
        if (headerLen != nullptr) {
            *headerLen = dataOffset();
        }
        return (reinterpret_cast<IO_metaBlock_header *>(bfrManager->blockAddress(rec)));
    }

    const char *getLabel() const OME_ALWAYS_INLINE NONNULL_RETURN {
        return (label);
    }

    void setLabel(const char *name) OME_ALWAYS_INLINE NONNULL_CLASS_PARAMETERS(2) {
        safe_strcpy(label, name, sizeof(label));
    }

    void setPacketsProcessedIncrement(int32_t incVal) {
        packetsProcessedIncrement = incVal;
    }

    //! Set read attempts before blocking
    void setReadAttempts(uint32_t count) OME_ALWAYS_INLINE {
        statistics->readAttemptsBeforeBlocking = count;
    }

    //! Get limit on read attempts before blocking
    uint32_t getReadAttempts() const OME_ALWAYS_INLINE {
       return ((uint32_t) statistics->readAttemptsBeforeBlocking);
    } 

    //! Set read timeout
    void setReadTimeout(uint32_t count) OME_ALWAYS_INLINE {
        statistics->maxReadTimeout = count;
    }

    //! Get limit on read attempts before blocking
    uint32_t getReadTimeout() const OME_ALWAYS_INLINE {
       return ((uint32_t) statistics->maxReadTimeout);
    } 

    //! Set extra information value
    void setExtraData(void *data) OME_ALWAYS_INLINE {
        extraData = data;
    }

    //! Retrieve extra information value
    void *getExtraData() const OME_ALWAYS_INLINE {
        return (extraData);
    }

    //! Set MTU
    void setMaxPacketSize(size_t bytes) OME_ALWAYS_INLINE {
        maxPacketSize = bytes;
    }

    //! Set processing routine.
    void setProcessRoutine(IO_processBlockFP func) OME_ALWAYS_INLINE {
        processBlockRoutine = func;
    }

    //! Set processing loop routine.
    void setProcessLoopRoutine(IO_processFP func) OME_ALWAYS_INLINE {
        processRoutine = func;
    }

    //! Set consume loop routine.
    void setConsumeLoopRoutine(IO_consumeFP func) OME_ALWAYS_INLINE {
        consumeRoutine = func;
    }

    //! Set consume routine.
    void setConsumeRoutine(IO_receiveBlockFP func) OME_ALWAYS_INLINE {
        recvRoutine = func;
    }

    //! Get requested thread modes.
    uint32_t getThreadMode() const OME_ALWAYS_INLINE {
        return (desiredThreadState);
    }
    
    //! Set threading mode
    int setThreadMode(ThreadMode mode);

    //! Set block delivery mode
    void setBlockingMode(BlockMode mode);

    //! Request stop
    int stopThread(uint32_t modes);

    //! Interrupt blocked thread
    int interruptThread(uint32_t modes, bool force=false);

    //! Wait for threads to start
    int waitForThreadStart(uint32_t mode);

    //! Wait for threads to terminate
    int waitForThreadExit(uint32_t modes);

    //! Wait for data to arrive
    int waitForDataToProcess(bool alreadyLocked=false);

    //! Wait for data to arrive or until a point in time
    int waitForDataToProcessOrUntil(const struct timespec *maxWaitUntil, bool alreadyLocked=false);

    //! Note new data has arrived
    int noteDataToProcess(bool alreadyLocked=false);


}; // end class IO_Processor

#pragma GCC diagnostic pop

#ifndef _WIN32
void setLabelForThread(pthread_t threadId, const char *label);
#else
void setLabelForThread(HANDLE threadId, const char *label);
#endif

/* NOTE:  these templated routines are marked as inline to allow
 * duplicate definitions to be collapsed at link time; as a practical
 * matter, they require an actual function body to be generated.
 */

/*! \brief Implements a processing routine that invokes a
 * processPacket() routine upon each incoming packet.
 */
template <class PARSER_TYPE> inline int processPacketUsingClass(SharedBufferAllocRecord *rec, class IO_Processor *controller)
{
    size_t len;

    unsigned char *bfr = controller->bufferAddress(rec, &len);
    PARSER_TYPE *parser = (PARSER_TYPE *) (controller->getExtraData());
    int rc = parser->processPacket(bfr, len);
    if (OME_EXPECT_FALSE(rc == -2)) {
        rec->usedLen = 0;   // force inject EOF
        return (0);
    }
    return (rc);
}

/*! \brief Implements a processing routine that invokes a
 * processPacketFromSource() routine upon each incoming packet,
 * Differs from processPacketUsingClass() by exposing the meta data header.
 */
template <class PARSER_TYPE> inline int processPacketFromSourceUsingClass(SharedBufferAllocRecord *rec, class IO_Processor *controller)
{
    size_t len, hdrLen;
    IO_metaBlock_header *metaHdr;

    if (controller->blockingMode & IO_Processor::_PACKET_HAS_META_DATA) {
        metaHdr = controller->bufferHeaderAddress(rec, &hdrLen);
    } else {
        metaHdr = nullptr;
        hdrLen = 0;
    }
    unsigned char *bfr = controller->bufferAddress(rec, &len);
    PARSER_TYPE *parser = (PARSER_TYPE *) (controller->getExtraData());
    int rc = parser->processPacketFromSource(bfr, len, metaHdr, hdrLen);
    if (OME_EXPECT_FALSE(rc == -2)) {
        rec->usedLen = 0;   // force inject EOF
        return (0);
    }
    return (rc);
}

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
