#ifndef _POLL_MONITOR_HPP_
//! \cond
#define _POLL_MONITOR_HPP_ "$Id: poll_monitor.hpp 318 2020-03-26 17:44:18Z geoff $"
//! \endcond
/* Copyright (C) 2014 - 2019 FARGOS Development, LLC
 */

#include <stdint.h>
#ifndef _WIN32
#if __MACH__ == 1 || __OpenBSD__ == 1 /* MacOS or OpenBSD */
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <vector>
#elif __linux__ == 1
#include <sys/epoll.h>
#endif
#ifndef INVALID_HANDLE_VALUE
#define INVALID_HANDLE_VALUE -1
#endif
#endif /* end ifndef _WIN32 */

#ifndef OS_HANDLE_TYPE
#ifdef _WIN32
#define OS_HANDLE_TYPE HANDLE
#else
#define OS_HANDLE_TYPE int
#endif
#endif /* OS_HANDLE_TYPE */

/*! \file
 * \brief Polling monitor
 */

/** \addtogroup FARGOSio
 * @{
 */

/*! \brief Monitor files and directories for changes.
 */
class PollMonitor {
protected:
#ifndef _WIN32
    int getInotifyResults(OS_HANDLE_TYPE *interestingFds, int *eventType);
#if __MACH__ == 1 || __OpenBSD__ == 1 /* MacOS or OpenBSD */
    int addKevent(const struct kevent &ev);
#endif
#endif
    int findFDentry(OS_HANDLE_TYPE fd);

    static uint64_t getFileSizeForDescriptor(OS_HANDLE_TYPE fd);

public:
    enum {
        NEW_FILE_DATA_AVAILABLE=1,  //!< Indicates new data available
        WRITER_CLOSED=8,            //!< Indicates writer closed file 
        POLLING_STOPPED=16,         //!< Polling stopped
        WATCHED_BY_INOTIFY=64       //!< Being watched via inotify()
    };
    //! \brief Return result from polling monitor.
    struct PollResult {
        OS_HANDLE_TYPE  fd;
        int             connectionType;
        void            *extraData;
        int             whyReady;
        int             watchId;
    };
    enum { MAX_FDS=2048 };
    uint32_t        fdsInUse;
#if __linux__ == 1 /* Linux */
    int             epoll_fd;
    int             inotify_fd;
#elif __MACH__ == 1 || __OpenBSD__ == 1 /* MacOS or OpenBSD */
    int		    kqueue_fd;
    int		    kevent_fd;
#endif
    PollResult      fdInfo[MAX_FDS];
#ifndef _WIN32
    int             watchIdToFd[MAX_FDS];
    int             fdToWatchID[MAX_FDS];
    int             fdToEntry[MAX_FDS];
#if __linux__ == 1 /* Linux */
    struct epoll_event  eventInfo[MAX_FDS];
#elif __MACH__ == 1 || __OpenBSD__ == 1 /* MacOS or OpenBSD */
    std::vector<struct kevent>   wantedEvents;
#endif
#endif /* end ifndef _WIN32 */

    PollMonitor();

    ~PollMonitor();

    /*! \brief Monitor a file for changes.
     */
    int addFile(const char *fileName, int connType=0, void *extraInfo=nullptr,
            OS_HANDLE_TYPE openedFD = INVALID_HANDLE_VALUE);

    /*! \brief Monitor an open descriptor.
     */
    int addDescriptor(OS_HANDLE_TYPE openedFD, int connType=0, void *extraInfo=nullptr) {
        int rc = addFile("", connType, extraInfo, openedFD);
        return (rc);
    }

    /*! \brief Watch directory for appearance of new files.
     */
    int addWatchDirectoryForNewFiles(const char *dirName);

    /*! \brief Watch file for increase in size.
     */
    int addWatchFileForAddedContent(const char *fileName, OS_HANDLE_TYPE openedFD);

    int dropDescriptor(OS_HANDLE_TYPE fd);

    int doPoll(PollResult interestingFds[MAX_FDS], int timeout=-1);

}; // end class PollMonitor

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
