#ifndef _READ_FILE_HPP_
//! \cond
#define _READ_FILE_HPP_ "$Id: read_file.hpp 455 2020-07-23 20:23:59Z geoff $"
//! \endcond

#include <utils/logging/logging_api.hpp>
#include <utils/io/io_processor.hpp>

/*! \file
 * \brief FARGOS file processing infrastructure.
 */
/** \addtogroup FARGOSio
 * @{
 */

/*! \brief Implements an adjustable sliding buffer that minimizes data movement
 * while enabling streams to process variable length content.
 *
 * Typical usage patterns have an outer loop whose goal is to make
 * calls to appendDataToBuffer() until all of the data has been appended.
 * Within that loop, it is typical to first test if spaceLeft() is less
 * than the amount of new data to be added.
 * If so, a call to discardInitialBytes(0) is made in an attempt to free up
 * all previously handled content in the front of the buffer.
 * Next, a call to appendDataToBuffer() is made, which returns the number
 * of bytes that were successfully appended.  That result is tested to
 * see if the number of bytes actually appended was the same as
 * requested quantity.  If not, depending on the application, this could be
 * an immediately fatal error.
 * For other scenarios, a sufficient quantity might now be in the buffer
 * to allow intermediate records to be successfully processed.
 * In those applications, there is an inner loop that processes all complete
 * records that are in the buffer. Once those records have been processed,
 * their corresponding data can be discarded from the buffer,
 * which in turn would free up space to allow more of the remaining
 * content to be appended on another pass of the outer loop.
 */
class File_Buffer {
public:
    enum {
        EXTERNAL_SEGMENT    = (1 << 0), //!< flag indicates buffer is in external segment
        IS_READ_ONLY        = (1 << 1)  //!< flag indicates buffer is read-only
    };
    unsigned char   *buffer;                //!< base of the buffer segment
    size_t          bufferLen;              //!< total length of buffer segment
    size_t          endBufferOffset;        //!< offset of last byte used in buffer
    size_t          startBufferOffset;      //!< offset of first byte used in buffer
    uint8_t         externalBufferFlags;    //!< flags indicating read-only, external

    /*! \brief Creates an empty buffer of the indicated size in the heap.
     */
    explicit File_Buffer(size_t bfrSize=65536) {
        bufferLen = bfrSize;
        endBufferOffset = 0;
        startBufferOffset = 0;
        buffer = new unsigned char[bufferLen];
        externalBufferFlags = 0;
    }

    /*! \brief Accepts the use (and potentially transfers ownership) of an
     * existing buffer that may already have some content.
     *
     * \param existingBuffer points to the existing buffer.
     * \param bfrSize indicates the size of the existing buffer.
     * \param currentLen specifies the length of the current content
     * in the buffer.
     * \param ownershipFlags specifies a mask of ownership attributes.
     * If the EXTERNAL_SEGMENT bit is set, the buffer will not be deleted.
     * This arises when the region is owned by another object or is on
     * the stack.
     * If the IS_READ_ONLY bit is set, no attempts to modify the
     * buffer will be honored.  This is useful when the region is
     * a read-only memory-mapped file that should not be modified.
     */
    File_Buffer(unsigned char *existingBuffer, size_t bfrSize,
        size_t currentLen=0, uint8_t ownershipFlags=EXTERNAL_SEGMENT) {
        bufferLen = bfrSize;
        endBufferOffset = currentLen;
        startBufferOffset = 0;
        buffer = existingBuffer;
        externalBufferFlags = ownershipFlags;
    }

    ~File_Buffer() {
        if ((externalBufferFlags & (EXTERNAL_SEGMENT | IS_READ_ONLY)) == 0) {
            delete[] buffer;
        }
    }

    /*! \brief Replace the existing buffer with a new region.
     *
     * \param existingBuffer points to the existing buffer.
     * \param bfrSize indicates the size of the existing buffer.
     * \param currentLen specifies the length of the current content
     * in the buffer.
     * \param ownershipFlags specifies a mask of ownership attributes.
     * If the EXTERNAL_SEGMENT bit is set, the buffer will not be deleted.
     * If the IS_READ_ONLY bit is set, no attempts to modify the
     * buffer will be honored.
     *
     * NOTE: should only be used if spaceUsed() is 0.
     * Obviously, caution should be used to ensure no entity expects
     * the buffer region to remain unchanged.
     */
    void useRegionAsBuffer(unsigned char *existingBuffer, size_t bfrSize,
            size_t currentLen, uint8_t ownershipFlags) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        if ((externalBufferFlags & (EXTERNAL_SEGMENT | IS_READ_ONLY)) == 0) {
            delete[] buffer; // delete prior content
        }
        bufferLen = bfrSize;
        endBufferOffset = currentLen;
        startBufferOffset = 0;
        buffer = existingBuffer;
        externalBufferFlags = ownershipFlags;
    }

    /*! \brief Returns the amount of unused space in the buffer.
     */
    inline size_t spaceLeft() const OME_ALWAYS_INLINE {
        return (bufferLen - endBufferOffset);
    }

    /*! \brief Returns the amount of space current used in the buffer.
     */
    inline size_t spaceUsed() const OME_ALWAYS_INLINE {
        return (endBufferOffset - startBufferOffset);
    }

    /*! \brief Returns the start of the active content in the buffer.
     */
    inline unsigned char *startBufferLocation() const OME_ALWAYS_INLINE {
        return (buffer + startBufferOffset);
    }

    /*! \brief Returns just past the end of the active content in the buffer.
     * Normally, this would be where new content would be appended.
     */
    inline unsigned char *endBufferLocation() const OME_ALWAYS_INLINE {
        return (buffer + endBufferOffset);
    }
    
    /*! \brief Adjusts the end of the buffer.  This is typically
     * called after new content has been added.
     *
     * \param amount specifies the number of bytes to be added.
     */
    inline void adjustBufferEnd(size_t amount) OME_ALWAYS_INLINE {
        endBufferOffset += amount;
    }

    /*! \brief Adjusts the start of the active content in the buffer.
     * 
     * \param amount specifies the number of bytes to be dropped
     * from the beginning of the active portion of the buffer.
     *
     * For efficiency, normally only the start of the buffer is
     * adjusted rather than performing a block move to reorient
     * the data in the buffer.
     *
     * See discardInitialBytes() for the routine that is used to free
     * up needed space.
     */
    inline void adjustBufferStart(size_t amount) OME_ALWAYS_INLINE {
        startBufferOffset += amount;
    }
    
    /*! \brief Discards bytes from the beginning of the active
     * portion of the buffer; typically as a result of having
     * processed them.
     */
    inline void discardInitialBytes(size_t bytes=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        size_t offset = startBufferOffset + bytes;
        // this test is done first to avoid any unneccessary writes
        if (offset == 0) { // already at start, no bytes to discard
            return; // nothing to do
        }
        if (OME_EXPECT_FALSE((externalBufferFlags & IS_READ_ONLY) != 0)) {
            if (offset == endBufferOffset) { // would be empty
                startBufferOffset = 0;  // reset
                endBufferOffset = 0;
                return;
            }
            // cannot permit buffer to be modified
            startBufferOffset += bytes;
            return;
        }
        size_t used = spaceUsed();
        if (used <= bytes) { // empty, reset to fill from start of buffer
            startBufferOffset = 0;
            endBufferOffset = 0;
            return;
        }
        size_t bytesLeft = used - bytes;
        memmove(buffer, buffer + offset, bytesLeft);
        startBufferOffset = 0;
        endBufferOffset = bytesLeft;
    }

    /*! \brief Convenience routine to append data to the buffer.
     *
     * \param newData points to the content to be added.
     * \param bytesToAdd specifies the length of the content to be added.
     * \param startFromOffset specifies an optional offset relative
     * to newData.
     *
     * The effective source address of the new content is newData + offset.
     *
     * \return The number of bytes added is returned.  This may not be the
     * same as bytesToAdd, particularly in the case where insufficient
     * space remains in the buffer.
     *
     * NOTE: this routine does not call discardInitialBytes() to free up
     * space on demand; this is the responsibility of the caller.
     */
    inline size_t appendDataToBuffer(const unsigned char *newData,
        size_t bytesToAdd, size_t startFromOffset=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3")
    {
        if (OME_EXPECT_FALSE((externalBufferFlags & IS_READ_ONLY) != 0)) {
            if (spaceUsed() == 0) { // nothing was stored at this moment
                // drop const, and set as new region, propagate existing flags
                unsigned char *bfrAddr = const_cast<unsigned char *>(newData) + startFromOffset;
                useRegionAsBuffer(bfrAddr, bytesToAdd, bytesToAdd, externalBufferFlags);
                return (bytesToAdd);
            }
            // cannot permit buffer to be modified
            return (0);
        }
        size_t byteCount = bytesToAdd;
        size_t left = spaceLeft();
        if (OME_EXPECT_FALSE(left < byteCount)) { // cannot add more than space left in the buffer
            byteCount = left; // truncate amount to store
        }
        memcpy(buffer + endBufferOffset, newData + startFromOffset, byteCount);
        endBufferOffset += byteCount;
        return (byteCount);
    }

}; // end class File_Buffer

/*! \brief Fundamental class that can read data from a memory-mapped region,
 * file or socket and process data in lines, fixed-sized blocks or as data is
 * available.
 */
class Read_And_Process_File {
public:
    enum ReadModes {
        READ_GUESS=0, READ_NORMAL=1, READ_ZLIB=2, READ_LZ4=4, READ_PCAP=8,
        WRITE=16, WRITE_APPEND=WRITE | 32, WRITE_TRUNCATE=WRITE | 64,
        READ_FROM_SOCKET=128
    };
    typedef int (*ReadDataFP)(class Read_And_Process_File *input, unsigned char *bfr, uint32_t bfrLen);

    static int defaultReadRoutine(Read_And_Process_File *input, unsigned char *bfr, uint32_t bfrLen);

    /*! \brief Describes magic numbers needed to identify a file's type
     * and the routine capable of decoding a file's contents.
     */
    struct FileTypeReaderSelector {
        ReadDataFP      readInterfaceRoutine;
        char            fileTypeDescription[24];
        enum ReadModes  modeToSelect;
        uint16_t        initialBytesToMatchMask;
        unsigned char   bytesToMatch[16];
    };

    static const FileTypeReaderSelector *findTypeOfFile(const char *fileName,
            const FileTypeReaderSelector *selectorTable);
protected:
    static uint32_t matchFileHeader(const FileTypeReaderSelector *criteria,
        const unsigned char *fileHdr, const size_t hdrLen);

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    virtual int readIntoBuffer(unsigned char *bfr, size_t bfrLen);
#pragma GCC diagnostic pop

protected:
    File_Buffer     *intermediateBuffer;
public:
    ReadDataFP      readRoutine;
    void            *auxData;
    OS_HANDLE_TYPE  descriptor;
    uint32_t        recordsProcessed;
    enum ReadModes  readMode;
    uint32_t        _explicitAlignmentPadding;

    /*! \brief Open the indicated file.
     * \param fileName specifies the file name.
     * \param mode specifies the how to open the file.
     *
     * \return A native operating system handle to
     * the open file. 
     * \retval _INVALID_DESCRIPTOR is returned if
     * the file could not be opened.
     */
    static OS_HANDLE_TYPE openFile(const char *fileName,
            ReadModes mode=READ_NORMAL);

    /*! \brief Close an native operating system file handle.
     * \param fd specifies the native file handle.
     * \param mode specifies the mode in which the file
     * was opened.
     */
    static int closeFile(OS_HANDLE_TYPE fd, ReadModes mode=READ_NORMAL);

    /*! \brief Search for a file using a combination of directory
     * roots and file suffixes.
     *
     * \param path specifies a buffer into which the found
     * file name will be stored.
     * \param pathLen specifies the length of the buffer.
     * \param searchRootPaths specifies a set of directories to be searched.
     * \param possibleFilenames specifies a set of potential
     * file name components.
     * \param possibleSuffixes specifies a set of possible file
     * name suffixes.
     *
     * Paths are separated by ":" on Unix systems and
     * by ";" under Windows.
     * \return the length of the found file is returned.
     * The \a path buffer will contain the file name.
     * \retval -1 indicates the file could not be found.
     */
    static int findFileInPathsWithSuffixes(char *path, uint_fast32_t pathLen,
        const char *searchRootPaths, const char *possibleFilenames="",
        const char *possibleSuffixes="");

    /*! \brief Construct from an existing file descriptor
     */
    explicit Read_And_Process_File(OS_HANDLE_TYPE srcDescriptor,
        ReadModes mode=READ_NORMAL,
        const FileTypeReaderSelector *selectorTable=nullptr);

    /*! \brief Construct given the name of a file to be opened.
     * \param fileName specifies the file to be opened
     * \param mode specifies the mode for the file.
     * \param selectorTable specifies a table of magic number
     * prefixes and associated routines to be used if
     * mode READ_GUESS is used.
     *
     * The special mode READ_GUESS can be used to automatically
     * select a file decoding mechanism based on the initial
     * bytes of a file.
     * This does not work on pipes or devices that cannot be
     * reopened.
     */
    explicit Read_And_Process_File(const char *fileName,
        ReadModes mode=READ_NORMAL,
        const FileTypeReaderSelector *selectorTable=nullptr);

    virtual ~Read_And_Process_File();

    /*! \brief Set a new file read routine.
     */
    void setReadRoutine(ReadDataFP altRoutine) OME_ALWAYS_INLINE {
        readRoutine = altRoutine;
    }

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    //! \brief user-exit to see original copy of any data read
    virtual void noteDataRead(const unsigned char *bfr, size_t bfrLen) const {}
#pragma GCC diagnostic pop

    //! \brief Process fixed length records
    int readAndProcessBlocksFromFile(size_t recordLength);

    /*! \brief Process next text line from buffer
     *
     * Invokes processHeaderLine() and processLine()
     */
    int findAndProcessNextLine(File_Buffer *fileBfr, bool hasHeaderLine=false);

    /*! \brief Process text lines
     *
     * Invokes beginFile(), completedFile(), readFileIntoBuffer(),
     * and findAndProcessNextLine()
     *
     * \return The number of lines processed is returned.
     */
    int readAndProcessTextLines(bool hasHeaderLine=false); // text file

    int readIntoFileBuffer(File_Buffer *bfr) {
        size_t result = readIntoBuffer(bfr->endBufferLocation(),
                bfr->spaceLeft());
        if (result > 0) {
            bfr->adjustBufferEnd(result);
        }
        return (result);
    }

    /*! \brief Process file contents with no imposed structure.
     *
     * NOTE: \ref intermediateBuffer needs to be set appropriately
     * by the subclass.
     */
    int readAndProcessFile() {
        int rc = beginFile(); // invoke user exit to do any initial work
        if (rc != 0) {
            return (rc);
        }

        ssize_t len = readIntoFileBuffer(intermediateBuffer);
        while (len > 0) {
            processBuffer(intermediateBuffer);
            len = readIntoFileBuffer(intermediateBuffer);
        }
        rc = completedFile(recordsProcessed); // invoke user routine to finish 
        return (rc);
    }


#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"

    // called after successful open, useful to output file headers
    virtual int beginFile() { return (0); }

    // called when all records read, useful to close output file
    virtual int completedFile(int recordsSeen) { return (recordsSeen); }

    // implement to do useful work when reading text lines...
    virtual int processLine(const unsigned char *line, size_t lineLen) { return (0); }

    /*! \brief Interface to handle special case processing of header lines. 
     *
     * \retval 0 = processed header, remaining lines are normal data
     * \retval 1 = processed header, still more header lines to consume
     * \retval 2 = not a header, need to process as normal data line
     */
    virtual int processHeaderLine(unsigned char *line, size_t lineLen) {
        return (2);
    }

    // implement to do useful work when reading fixed length blocks
    virtual int_fast32_t processBlock(unsigned char *block, size_t blockLen) {
//LOG_COMPONENT_CERR(io,warn) << "unexpected: Default Read_And_Process_File::processBlock() len=" << blockLen << LOG_ENDLINE;
        return (0);
    };

    // implement to do useful work when reading arbitrary-sized blocks
    virtual int processBuffer(File_Buffer *bfrState) {
//LOG_COMPONENT_CERR(io,warn) << "unexpected: Default Read_And_Process_File::processBuffer() len=" << bfrState->spaceUsed() << LOG_ENDLINE;
        return (0);
    }

#pragma GCC diagnostic pop

}; // end class Read_And_Process_File



/*! \brief Scan streams of data for start and end tags,
 * invokes processDocument() on found content.
 */
class Extract_And_Process_Document_Stream {
protected:
    enum {
        FIND_NOTHING=0,
        FIND_DOC_START=1,
        FIND_END_TAG=2,
        FIND_BLOCK_LENGTH=3,
    }   scanState, restartScanState;
    unsigned char  documentStartPrefix[64];
    unsigned char  documentEndTag[64];
    File_Buffer *pendingContent;
    ssize_t     neededBlockLen;
    size_t      documentStartLen;
    size_t      documentEndLen;

    int scanForDocument(); // TODO: invokes processDocument(doc,len) when found

public:
    Extract_And_Process_Document_Stream(const char *docStart,
            const char *endTag, size_t maxDocSize=1024*128)
    {
        pendingContent = new File_Buffer(maxDocSize);
        safe_strcpy(documentStartPrefix, docStart, sizeof(documentStartPrefix));
        safe_strcpy(documentEndTag, endTag, sizeof(documentEndTag));

        documentStartLen = strlen((const char *) documentStartPrefix);
        documentEndLen = strlen((const char *) documentEndTag);
        neededBlockLen = -1;

        if (documentStartLen > 0) {
            restartScanState = FIND_DOC_START;
        } else {
            restartScanState = FIND_END_TAG;
        }
        scanState = restartScanState;
    }

    virtual ~Extract_And_Process_Document_Stream()
    {
        delete pendingContent;
    }

    int scanAndProcessBuffer(const unsigned char *block, size_t blockLen) {
        size_t left = pendingContent->spaceLeft();
//LOG_COMPONENT_CERR(io,trace) << "scanAndProcessBuffer add=" << blockLen << " left=" << left << " rtained=" << pendingContent->spaceUsed() << LOG_ENDLINE;
        while (OME_EXPECT_FALSE(left < blockLen)) { // not enough space for all of it
            // force relocation of any remaining data to front of buffer
            pendingContent->discardInitialBytes(0);
            // then add what we can
            size_t bytesAdded = pendingContent->appendDataToBuffer(block, blockLen);
            block += bytesAdded;
            blockLen -= bytesAdded;
            // try scanning for one or more complete documents
            scanForDocument();
            if (blockLen == 0) { // added everything
                return (0);
            }
            left = pendingContent->spaceLeft();
            if (left == 0) {
                return (-1);
            }
        }
        pendingContent->appendDataToBuffer(block, blockLen);
        scanForDocument();
        return (0);
    }

    int scanAndProcessBuffer(File_Buffer *bfrState) {
        unsigned char *block = bfrState->startBufferLocation();
        size_t blockLen = bfrState->spaceUsed();
        int result = scanAndProcessBuffer(block, blockLen);
        bfrState->discardInitialBytes(blockLen); // drop content passed in
        return (result);
    }

    /*! \brief Specialized interface to add incoming block from BufferRegion
     * maintained by an IO_Processor.
     * Block content is appended and then returned before any processing
     * of the content is attempted.
     *
     * This interface exists to handle a very exotic scenario in which
     * the BufferRegion will be taken over by a child routine invoked
     * by scanForDocument() and the BufferRegion must have no blocks in use
     * at the time of the call.
     */
    int addIOblockThenProcess(SharedBufferAllocRecord *rec, IO_Processor *controller) {
        size_t newLen;
        unsigned char *newData = controller->bufferAddress(rec, &newLen);
        size_t left = pendingContent->spaceLeft();
        if (OME_EXPECT_FALSE(left < newLen)) { // not enough space for all of it
            // force relocation of any remaining data to front of buffer
            pendingContent->discardInitialBytes(0);
            left = pendingContent->spaceLeft();
            if (left < newLen) { // still not enough space
                LOG_COMPONENT_CERR(app,error) << "Could not add " << newLen <<
                    " bytes as only space for " << left << " remains" << LOG_ENDLINE;
                return (-1);
            }
        }
        // then add what we can
        size_t bytesAdded = pendingContent->appendDataToBuffer(newData, newLen);
        controller->bfrManager->returnBlock(rec); // force return of block
        scanForDocument();
        return (0);
    }

    /*! \brief User-exit to process extracted document
     *
     * \return Number of bytes of required for a subsequent block of raw data.
     * \retval 0 indicates no block of raw data is required.
     * \retval -1 indicates some error.
     */
    virtual int processDocument(unsigned char *docStart, size_t docLen) = 0;

    /*! \brief User-exit to process block of raw data.
     *
     * Will be invoked if processDocument() returns a value greater than 0.
     */
    virtual int_fast32_t processBlock(unsigned char *blockStart, size_t blockLen) {
        return (0);
    }

}; // end class Extract_And_Process_Document_Stream

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"
/*! \brief Scan HTTP and parse stream for HTTP requests/responses
 */
class Parse_And_Process_HTTP_Stream : public Extract_And_Process_Document_Stream {
public:
    enum ParseState {
        READ_RESPONSE=1,
        READ_REQUEST,
        READ_HEADER_LINE,
        READ_CONTENT_BODY
    };
protected:
    ParseState  parseState;
    ParseState  restartParseState;

public:
    char    httpVersion[16];    // typically:  HTTP/1.1
    char    statusCode[8];      // typically:  200
    char    responseReason[48]; // typically:  OK
    char    httpRequest[16];
    char    requestURL[1024];
    int32_t contentLengthWanted;
    std::map<std::string,std::string>   headerAttributes;

    explicit Parse_And_Process_HTTP_Stream(size_t maxDocSize=1024*128,
            ParseState initialState=READ_RESPONSE) :
        Extract_And_Process_Document_Stream("", "\r\n", maxDocSize)
    {
        restartParseState = initialState;
        parseState = initialState;
        contentLengthWanted = -1;

        httpVersion[0] = '\0';
        statusCode[0] = '\0';
        responseReason[0] = '\0';
    }
    
    virtual ~Parse_And_Process_HTTP_Stream() {}

    const char *findHeaderAttribute(const char *headerElement) const {
        std::map<std::string,std::string>::const_iterator i = headerAttributes.find(headerElement);
        if (i != headerAttributes.end()) {
            return (i->second.c_str());
        }
        return (nullptr);
    }

    /*! \brief User-exit for HTTP response line.
     *
     * Default implementation parses response line, which is expected to
     * appear like "HTTP/1.1 200 OK"
     */
    virtual int processHTTPresponse(unsigned char *line, size_t lineLen) {
        if (line[lineLen - 1] == '\n') {
            lineLen -= 1;
            if (line[lineLen - 1] == '\r') {
                lineLen -= 1;
            }
        }
        if (lineLen == 0) {
            LOG_COMPONENT_CERR(io,trace) << "Got null HTTP request" << LOG_ENDLINE;
            return (0);
        }
        LOG_COMPONENT_CERR(io,trace) << "Got HTTP response: len=" << lineLen <<
            " line=" << AS_TEXT_BUFFER((char *) line, lineLen) << LOG_ENDLINE;
        size_t offset = 0;
        size_t i = 0;
        while ((i < (sizeof(httpVersion) - 1)) && (offset < lineLen)) {
            if (isspace(line[offset])) break;
            httpVersion[i++] = (char) line[offset++];
        }
        httpVersion[i] = '\0';
        while (isspace(line[offset]) && (offset < lineLen)) {
            offset += 1;
        }
        i = 0;
        while ((i < (sizeof(statusCode) - 1)) && (offset < lineLen)) {
            if (isspace(line[offset])) break;
            statusCode[i++] = (char) line[offset++];
        }
        statusCode[i] = '\0';
        while (isspace(line[offset]) && (offset < lineLen)) {
            offset += 1;
        }
        i = 0;
        while ((i < (sizeof(responseReason) - 1)) && (offset < lineLen)) {
            if (isspace(line[offset])) break;
            responseReason[i++] = (char) line[offset++];
        }
        responseReason[i] = '\0';
        headerAttributes.clear();

        return (0);
    }

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-methods"
    /*! \brief User-exit for HTTP request line.
     *
     * Default implementation parses HTTP request line, which is expected to
     * appear like: "GET / HTTP/1.1"
     */
    virtual int processHTTPrequest(unsigned char *line, size_t lineLen) {
        if (line[lineLen - 1] == '\n') {
            lineLen -= 1;
            if (line[lineLen - 1] == '\r') {
                lineLen -= 1;
            }
        }
        if (lineLen == 0) {
            LOG_COMPONENT_CERR(io,trace) << "Got null HTTP request" << LOG_ENDLINE;
            return (0);
        }
        LOG_COMPONENT_CERR(io,trace) << "Got HTTP request: len=" << lineLen <<
            " line=" << AS_TEXT_BUFFER((char *) line, lineLen) << LOG_ENDLINE;
        size_t offset = 0;
        size_t i = 0;
        while ((i < (sizeof(httpRequest) - 1)) && (offset < lineLen)) {
            if (isspace(line[offset])) break;
            httpRequest[i++] = (char) line[offset++];
        }
        httpRequest[i] = '\0';
        while (isspace(line[offset]) && (offset < lineLen)) {
            offset += 1;
        }
        i = 0;
        while ((i < (sizeof(requestURL) - 1)) && (offset < lineLen)) {
            if (isspace(line[offset])) break;
            requestURL[i++] = (char) line[offset++];
        }
        requestURL[i] = '\0';
        while (isspace(line[offset]) && (offset < lineLen)) {
            offset += 1;
        }
        i = 0;
        while ((i < (sizeof(httpVersion) - 1)) && (offset < lineLen)) {
            if (isspace(line[offset])) break;
            httpVersion[i++] = (char) line[offset++];
        }
        httpVersion[i] = '\0';
        headerAttributes.clear();

        return (0);
    }

    /*! \brief User exit for HTTP header line.
     *
     * Default implementation saves lines as key/value pairs in a map.
     * Content-Length automatically recognized to retrieve block of subsequent
     * data.
     */
    virtual int processHTTPheader(unsigned char *line, size_t lineLen) {
        if (line[lineLen - 1] == '\n') {
            lineLen -= 1;
            if (line[lineLen - 1] == '\r') {
                lineLen -= 1;
            }
        }
        if (lineLen == 0) { // end of HTTP headers
            return (0);
        }
        LOG_COMPONENT_CERR(io,trace) << "Got HTTP header len=" << lineLen <<
            " line=" << AS_TEXT_BUFFER((char *) line, lineLen) << LOG_ENDLINE;
        unsigned char *colon = (unsigned char *) memchr(line, ':', lineLen);
        if (colon != nullptr) {
            size_t keyLen = (colon - line);
            size_t valLen = lineLen - (keyLen + 1);
            colon += 1; // past colon
            while (valLen > 0) { // skip initial white space
                if (isspace(*colon) == false) break;
                valLen -= 1;
                colon += 1;
            }
// LOG_COMPONENT_CERR(io,trace) << "key=" << AS_TEXT_BUFFER((char *) line, keyLen) << " val=" << AS_TEXT_BUFFER((char *) colon, valLen) << LOG_ENDLINE;
            if (memcmp(line, "Content-Length:", 15) == 0) {
                contentLengthWanted = text2uint32((char *) colon, valLen);
            } 
            headerAttributes[std::string((const char *) line, keyLen)] = std::string((const char *) colon, valLen);
        }
        return (0);
    }
#pragma GCC diagnostic pop

    // callback from Extract_And_Process_Document_Stream
    virtual int_fast32_t processBlock(unsigned char *blockStart, size_t blockLen) VIRTUAL_OVERRIDE {
// LOG_COMPONENT_CERR(io,warn) << "invoked Parse_And_Process_HTTP_Stream::processBlock() with no implementation, blockLen=" << blockLen << LOG_ENDLINE;
        return (0);
    }

    /* callback from Extract_And_Process_Document_Stream (abstract function)
     *
     * In the context of the HTTP stream, this will be invoked for each
     * HTTP request/response/header line.
     */
    int processDocument(unsigned char *docStart, size_t docLen) VIRTUAL_OVERRIDE {
//LOG_COMPONENT_CERR(io,trace) << "processDocument curState=" << parseState << " len=" << docLen << " doc=" << AS_TEXT_BUFFER((char *) docStart, docLen) << std::endl;
        int result = 0;
        switch (parseState) {
        case READ_REQUEST:
            if (docLen <= 2) { // assume just CR/LF, consume silently
                LOG_COMPONENT_CERR(io,info) << "Expected HTTP request, treating as null because length=" << docLen << LOG_ENDLINE;
                return (0);
            }
            processHTTPrequest(docStart, docLen);
            parseState = READ_HEADER_LINE;
            contentLengthWanted = -1;
            break;
        case READ_RESPONSE:
            if (docLen <= 2) { // assume just CR/LF, consume silently
                LOG_COMPONENT_CERR(io,info) << "Expected HTTP response, treating as null because length=" << docLen << LOG_ENDLINE;
                return (0);
            }
            processHTTPresponse(docStart, docLen);
            parseState = READ_HEADER_LINE;
            contentLengthWanted = -1;
            break;
        case READ_HEADER_LINE:
            processHTTPheader(docStart, docLen);
            if (docLen == 2) { // just the CR/LF
                if (contentLengthWanted <= 0) {
                    parseState = restartParseState; // do next response
                    // if Transfer-Encoding: chunked, handle change to mode here
                } else {
                    parseState = restartParseState;
                    result = contentLengthWanted;
                }
            }
            break;
        case READ_CONTENT_BODY:
            LOG_COMPONENT_CERR(io,error) << "unexpected: state=READ_CONTENT_BODY" << LOG_ENDLINE;
            parseState = restartParseState;
            break;
        }
        return (result);
    }

}; // end Parse_And_Process_HTTP_Stream

#pragma GCC diagnostic pop


#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"
/*! \brief Read and process Comma-Separated-Value (or equivalent)
 * files a record at a time.
 */
class CSV_File : public Read_And_Process_File {
public:
    enum { MAX_FIELDS = 128};
protected:
    unsigned char   *headerLine;
    unsigned char   separatorCharList[8];
    unsigned char   quoteCharList[8];
    char            *fields[MAX_FIELDS];
    int             fieldTotal;
    uint32_t        readsPerformed;
    unsigned char   separatorListLen;
    unsigned char   quoteListLen;
    bool            doHeaderLine;
public:

    /*! \brief Parse conventional CSV file.
     * \param descriptor is the descriptor/handle of an already open file.
     * \param hasHeader is an optional argument that indicates if the
     * file has an initial header describing the column names.
     * \param sepChar is an optional argument indicating the separator character
     * used between columns; it defaults to a comma.
     *
     * By default, the double quote mark can be used to group elements as a single field.
     */
    explicit CSV_File(OS_HANDLE_TYPE descriptor, bool hasHeader=true, char sepChar=',');

    /*! \brief Parse a CSV-style file that has more than one column separator
     * character in use.
     * \param descriptor is the descriptor/handle of an already open file.
     * \param separatorList is a pointer to text string listing possible
     * separator characters used between columns.
     * \param quoteChars is an optional pointer to a text string listing possible
     * quote characters.
     * \param hasHeader is an optional argument that indicates if the
     * file has an initial header describing the column names.
     */
    CSV_File(OS_HANDLE_TYPE descriptor, const char *separatorList,
            const char *quoteChars = "\"", bool hasHeader=true);


    virtual ~CSV_File() {
        delete[] headerLine;
    }

    /*! \brief Return the relative subscript for a named column.
     *
     * \param fieldHeading specifies the name of the desired column
     *
     * \return The relative index of the desired column (starting at 0) is 
     * returned.
     * \retval -1 indicates the column was not found.
     */
    int getFieldIndex(const char *fieldHeading) const;

    /*! \brief Parse text line into fields.
     *
     * Fields are separated by
     * any character specified in the separator list, unless inside a quoted
     * sequence. Quoted sequences are started by any character in the
     * quoted character list.
     *
     * \param maxFields indicates the number of elements available
     * in \a fieldStart.
     * \param fieldStart is an array of character pointers into which
     * will be stored a pointer to each field's value.
     * \param line points to the line to be parsed.  It is expected that
     * it will end in a newline character.  NOTE:  the contents of \a line
     * will be modified.
     * \param lineLen is the length of the line; the last character (which
     * would be line[lineLen]) should be a newline.
     *
     * \return The number of fields parsed is returned.  Only pointers to
     * the first \a maxFields are stored into the \a fieldStart array,
     * but the total number of fields seen is returned.
     */
    int parseIntoFields(int maxFields, char *fieldStart[],
            unsigned char *line, size_t lineLen);

    /*! \brief Top-level routine to parse CSV file.
     *
     * A trivial cover routine to invoke readAndProcessTextLines()
     * with the appropriate heading mode.
     * 
     * \return The value from readAndProcessTextLines() is returned.
     */
    int readAndProcessCSV_File() { // CSV text file with headings
        int result = readAndProcessTextLines(doHeaderLine);
        return (result);
    };

    /*! \brief Streaming interface equivalent to readAndProcessCSV_File().
     * API is compatible with processPacketUsingClass<> convenience
     * interface for the IO_Processor class processing routine.
     *
     * NOTE: \ref intermediateBuffer needs to be set appropriately
     * by the subclass.
     */
    int processPacket(unsigned char *data, size_t len) {
        if (readsPerformed == 0) {
            beginFile();
        }
        readsPerformed += 1;
        if (len == 0) {
            completedFile(recordsProcessed);
            return (0);
        }
        noteDataRead(data, len);
        intermediateBuffer->appendDataToBuffer(data, len);
        int bytesProcessed;
        do {
            bytesProcessed = findAndProcessNextLine(intermediateBuffer, doHeaderLine);
            if ((doHeaderLine != false) && (bytesProcessed > 0)) {
                doHeaderLine = false;
            }
        } while (bytesProcessed != 0);
        intermediateBuffer->discardInitialBytes(0);
        return (0);
    }

    /*! \brief Process initial header line in file.
     *
     * Normally called by readAndProcessTextLines().
     */
    virtual int processHeaderLine(unsigned char *line, size_t lineLen) VIRTUAL_OVERRIDE;

    /*! \brief User-exit invoked when initial header line is parsed.
     *
     * Usually, this is used to make several calls to \ref getFieldIndex()
     * to obtain the corresponding subscripts of the desired columns,
     * thus performing the expensive work only once per file.
     *
     * \retval 0 for now, always return 0.  The value is currently ignored.
     */
    virtual int parsedHeadingLine() { return (0); }

    /*! \brief User-exit to notify that file has been
     * completely read.
     * \retval 0 for now, always return 0.  The value is currently ignored.
     */
    virtual int completedFile(int recordsSeen) VIRTUAL_OVERRIDE;
}; // end class CSV_File

#pragma GCC diagnostic pop

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
