#ifndef _SHARED_VARIABLE_HPP_
//! \cond
#define _SHARED_VARIABLE_HPP_ "$Id: shared_variable.hpp 456 2020-07-23 20:25:49Z geoff $"
//! \endcond

/* Copyright (C) 2011 - 2020 FARGOS Development, LLC
 * Author:  Geoff Carpenter - http://www.fargos.net/gcc.html
 */

#include <utils/base/compiler_hints.h>
#include <stdint.h>
#include <string.h>
#include <list> //needed for deferredList
#include <typeinfo>

#include <utils/base/circular_bfr.hpp>
#include <utils/io/mapped_file.h> // to pick up MAP_FILE defines for createSegment modes

/*! \file
 * \brief FARGOS Shared Memory Variable routines.
 *
 * \author Geoff Carpenter
 * \copyright 2011 - 2020 FARGOS Development, LLC
 */
/** \defgroup FARGOS_shared FARGOS Shared Memory Functions
 *
 * \brief Implementation of integer, floating-point and string variables
 * within shared memory segments.
 * @{
 */

//! \brief Magic number used for Shared Memory Variable segments
#define SMV_MAGIC_NUMBER "SHMVAR01"

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-final-types"
/*! \brief Shared memory variable which allows statistics and operational
 * controls to be exposed to other processes with minimal runtime overhead.
 */
class SharedMemoryVariable {
friend class SharedMemoryVariableManager;
public:
    enum SharedMemoryVariableType {
        SMV_TYPE_NONE=0,
        SMV_TYPE_UNSIGNED=1,
        SMV_TYPE_LARGE=2,
        SMV_TYPE_INT32=4,
        SMV_TYPE_UINT32=SMV_TYPE_INT32|SMV_TYPE_UNSIGNED,
        SMV_TYPE_INT64=SMV_TYPE_INT32|SMV_TYPE_LARGE,
        SMV_TYPE_UINT64=SMV_TYPE_INT64|SMV_TYPE_UNSIGNED,
        SMV_TYPE_FLOAT=8,
        SMV_TYPE_DOUBLE=SMV_TYPE_FLOAT|SMV_TYPE_LARGE,
        SMV_TYPE_FIXED=SMV_TYPE_FLOAT|SMV_TYPE_UNSIGNED,
        SMV_TYPE_TINY_STRING=16,    // 7 or less chars + null 
        SMV_TYPE_TINY_BINARY_STRING=SMV_TYPE_TINY_STRING|SMV_TYPE_UNSIGNED,
        SMV_TYPE_STRING=SMV_TYPE_TINY_STRING|SMV_TYPE_LARGE,
        SMV_TYPE_BINARY_STRING=SMV_TYPE_STRING|SMV_TYPE_UNSIGNED,
        SMV_TYPE_NAME_NODE=32
    };

    enum {
        SMV_DISPLAY_NAME=1,
        SMV_DISPLAY_VALUE=2,
        SMV_DISPLAY_TYPE=4,
        SMV_DISPLAY_EQUALS=8
    };

    /*! \brief Storage layout for shared memory variable record within
     * a memory mapped segment.
     *
     * Note: all data in this record must be able to be interpreted regardless
     * of address space.
     */
    struct SharedMemoryVariable_Record final {
        /* WARNING:  it is MANDATORY that numericValue be at offset 0 */
        union {
            uint64_t    u64;
            int64_t     i64;
            int32_t     i32;
            uint32_t    u32;
            float       f;
            double      d;
            char        tinyString[8];
        } numericValue; /* 8 bytes, must be at offset 0 */
        uint64_t        parentNodeOffset;
        uint32_t        allocatedLength;    //!< allocated length
        uint32_t        usedLength;         //!< used length, always <= allocated length
        uint16_t        recordType;         //!< type of record, see \ref SharedMemoryVariableType
        uint16_t        varNameLength;      //!< strlen() + 1; includes null in count
        char            varName[12 + sizeof(SharedBufferAllocRecord)]; //!< variable name

        //! \brief Format variable into text buffer.
        char *formatAsText(char *bfr, uint_fast32_t bfrLen, uint_fast32_t displayMode,
                const char *pathPrefix=nullptr, bool isNative=true) const;

        //! \brief Get parent path name of variable.
        int getPathPrefix(char *bfr, uint_fast32_t bfrLen, unsigned char *base, bool isNative=true) const;

    };
protected:
    static SharedMemoryVariable_Record  unallocated_data;

    SharedMemoryVariable_Record     *dataPointer;
    class SharedMemoryVariableNode  *parentNodeRecord;
    class SharedMemoryVariableManager *variableManager;
    const char      *variableName;
    uint32_t        minLength;
    unsigned char   dataType;
    unsigned char   registered;
    unsigned char   _pad1[2];

public:
    SharedMemoryVariableNode *getParentNode() const {
        return (parentNodeRecord);
    }

    SharedMemoryVariableManager *getVariableManager() const {
        return (variableManager);
    }

    //! \brief Return type as text string.
    static const char *typeName(unsigned int t) OME_CONST_FUNCTION;

    //! \brief Return type of variable, see \ref SharedMemoryVariableType
    unsigned int getType() const OME_ALWAYS_INLINE {
        return (dataType);
    }

    //! \brief Return minimum number of bytes required for variable record.
    unsigned int getMinLength() const OME_ALWAYS_INLINE {
        return (minLength);
    }

    //! \brief Get variable name.
    const char *getName(uint_fast32_t *retNameLen=nullptr) const OME_ALWAYS_INLINE {
        if (OME_EXPECT_TRUE(registered)) {
            if (retNameLen != nullptr) {
                *retNameLen = dataPointer->varNameLength;
            }
            return (dataPointer->varName);
        } else { // not yet allocated
            if (retNameLen != nullptr) {
                *retNameLen = strlen(variableName);
            }
            return (variableName);
        }
    }

    //! \brief Set new variable name
    int setName(const char *newName) {
        variableName = newName;
        if (registered) {
            strncpy(dataPointer->varName, newName, dataPointer->varNameLength);
        }
        return (registered);
    }

    //! \brief Get full path name of variable
    uint_fast32_t getFullName(char *bfr, uint_fast32_t bfrLen) const;

    /*! \brief Construct an shared memory variable and attach to the
     * indicated manager for the collection of variables.
     */
    SharedMemoryVariable(class SharedMemoryVariableManager *mgr,
            const char *varName, int varType, size_t length);

    SharedMemoryVariable(class SharedMemoryVariableNode *parentNode,
            const char *varName, int varType, size_t length);

    virtual ~SharedMemoryVariable();

    /*! \brief Extension mechanism to allow derived classes to be informed when
     * they are about to be assigned storage; called at most once per lifetime
     * of object.
     */
    virtual void noteNowRegistered(SharedMemoryVariable_Record *newArea) {}

    void makeAliasFor(SharedMemoryVariable *v) {
        dataPointer = v->dataPointer;
    }

}; // end class SharedMemoryVariable
#pragma GCC diagnostic pop

/*! \brief Manager for collection of shared memory variables.
 *
 * \note One of the significant features of the SharedMemoryVariableManager
 * is that it provides support for registering SharedMemoryVariables with it
 * prior to the shared memory segment being created.  This allows static
 * variables to be defined and the segment to be created at runtime after
 * a file name is computed.
 */
class SharedMemoryVariableManager {
protected:
    typedef std::list<SharedMemoryVariable *> smv_list_t;

    smv_list_t              *deferredList;
    BufferRegion            *bfrMgr;
    unsigned char           *regionBase;
    size_t                  regionLength;

    /*! \brief Internal routine to allocate storage within the segment
     * for a shared memory variable.
     */
    void allocateVariable(SharedMemoryVariable *var);

    /*! \brief Internal routine to allocate previously registered
     * SharedMemoryVariables within a newly allocated segment.
     */
    void allocateDeferredVariables();

public:
    /*! \brief Open a file and map it to a memory segment for variable
     *  storage.
     *
     *  \param fileName specifies the file to be opened.
     *  \param segmentLengthPtr points to an area from which the desired
     *  segment length will be taken or will be filled in with the size
     *  of the opened file.
     *  \param initMode is a mode for createMappedFile(), 
     *  typically MAP_FILE_INIT.
     *  \param desiredMagicNumber points to a sequence of character
     *  that will be used as the magic number for the segment.
     *  Typically this is SMV_MAGIC_NUMBER.
     */
    static unsigned char *createSegment(const char *fileName,
        size_t *segmentLengthPtr, int initMode = MAP_FILE_INIT,
        const char *desiredMagicNumber = SMV_MAGIC_NUMBER);

    /*! \brief Use an existing, already initialized, memory segment
     * for variable storage.
     *
     * This routine does not destroy the existing contents of the
     * region and is typically used by an external process to access
     * a segment.
     */
    void attachExistingSegment(unsigned char *segment,
            size_t segmentLength);

    /*! \brief Overloaded variant of attachExistingSegment() that uses
     * an BufferRegion object in control of an already initialized segment.
     *
     * \param bfr specifies the BufferRegion manager owning the
     * segment.  Ownership of the BufferRegion object is transferred.
     *
     * This routine does not destroy the existing contents of the region.
     */
    void attachExistingSegment(BufferRegion *bfr);

    /*! \brief Initialize a memory segment for use as variable storage.
     *
     * \param segment points to the base of the segment
     * \param segmentLength specifies the length of the segment
     * \param componentName specifies the name of the component
     * initializing the segment.
     */
    void initializeSegment(unsigned char *segment, size_t segmentLength,
            const char *componentName);

    /*! \brief Overloaded variant of initializeSegment that
     * initializes a memory segment already under control of a
     * BufferRegion for use as variable storage.
     *
     * \param bfr points to an already established BufferRegion 
     * whose region will be initialized for use.  Ownership of the BufferRegion
     * object is transferred.
     */
    void initializeSegment(BufferRegion *bfr);

    /*! \brief Convenience function to create a file and map it to 
     * segment for variable storage.  The filename will include the
     * component name, date, and process Id.
     */
    unsigned char *createAndAttachStandardSegment(const char *componentName,
            size_t *segmentLength);

    /*! \brief Constructor for a shared memory variable manager.
     *
     * \param inDataSegment should be set to false if the object is
     * constructed in the heap (e.g., via the new operator).
     */
    explicit SharedMemoryVariableManager(bool inDataSegment=true);

    ~SharedMemoryVariableManager();

    int takeDeferredRegistrations(SharedMemoryVariableManager *otherMgr);

    SharedMemoryVariable::SharedMemoryVariable_Record *iterateOver32BitSegment(SharedBufferAllocRecord_32 **iterator);

    SharedMemoryVariable::SharedMemoryVariable_Record *iterateOver64BitSegment(SharedBufferAllocRecord_64 **iterator);

    size_t offsetInSegment(const SharedMemoryVariable::SharedMemoryVariable_Record *rec) const {
        size_t result = reinterpret_cast<const unsigned char *>(rec) -
            bfrMgr->getBufferBase();
        return (result);
    }

    inline bool initialized() const OME_ALWAYS_INLINE { return (bfrMgr != nullptr); }

    //! \brief Register a single variable with the segment manager.
    bool registerVariable(SharedMemoryVariable &var);

    //! \brief Register an array of variables with the segment manager.
    void registerVariables(int count, SharedMemoryVariable *vars[]);
}; // end class SharedMemoryVariableManager


//! \brief Default shared memory variable manager.
extern SharedMemoryVariableManager DEFAULT_sharedMemoryVariableManager;

/*! \brief Intermediate naming node for supporting variable naming
 * hierarchies.
 *
 * Each individual variable only stores its dedicated suffix;
 * an arbitrary path can be constructed using a linked list of parent naming
 * nodes and elements in the logical path name hierarchy are only stored once.
 */
class SharedMemoryVariableNode : public SharedMemoryVariable
{
public:
    explicit SharedMemoryVariableNode(const char *varName,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager) :
        SharedMemoryVariable(mgr, varName,
                SMV_TYPE_NAME_NODE, strlen(varName)) {}

    SharedMemoryVariableNode(const char *varName,
        SharedMemoryVariableNode *parentNode) :
        SharedMemoryVariable(parentNode, varName,
                SMV_TYPE_NAME_NODE, strlen(varName)) {}
}; // end class SharedMemoryVariableNode



/*! \brief Convenience template for creating shared memory variables of
 * one of the supported numeric types.
 */
template<typename NUMTYPE> class SMV_Numeric:
    public SharedMemoryVariable
{
public:
    explicit SMV_Numeric(const char *varName,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager) NONNULL_CLASS_PARAMETERS(2) :
        SharedMemoryVariable(mgr, varName,
                typeid(NUMTYPE) == typeid(int32_t) ? SMV_TYPE_INT32 :
                typeid(NUMTYPE) == typeid(uint32_t) ? SMV_TYPE_UINT32 :
                typeid(NUMTYPE) == typeid(int64_t) ? SMV_TYPE_INT64 :
                typeid(NUMTYPE) == typeid(uint64_t) ? SMV_TYPE_UINT64 :
                typeid(NUMTYPE) == typeid(float) ? SMV_TYPE_FLOAT :
                typeid(NUMTYPE) == typeid(double) ? SMV_TYPE_DOUBLE :
                ~0, /* default if unknown */
                sizeof(NUMTYPE)) {}

    SMV_Numeric(const char *varName,
        SharedMemoryVariableNode *parentNode) NONNULL_CLASS_PARAMETERS(2) :
        SharedMemoryVariable(parentNode, varName,
                typeid(NUMTYPE) == typeid(int32_t) ? SMV_TYPE_INT32 :
                typeid(NUMTYPE) == typeid(uint32_t) ? SMV_TYPE_UINT32 :
                typeid(NUMTYPE) == typeid(int64_t) ? SMV_TYPE_INT64 :
                typeid(NUMTYPE) == typeid(uint64_t) ? SMV_TYPE_UINT64 :
                typeid(NUMTYPE) == typeid(float) ? SMV_TYPE_FLOAT :
                typeid(NUMTYPE) == typeid(double) ? SMV_TYPE_DOUBLE :
                ~0, /* default if unknown */
                sizeof(NUMTYPE)) {}

    /* WARNING:  these all depend on SharedMemoryVariable_Record having
     * the numeric field at offset 0
     */
    //! \brief Cast operator for a numeric shared memory variable.
    inline operator NUMTYPE() const OME_ALWAYS_INLINE {
        return (*reinterpret_cast<const NUMTYPE *>(dataPointer));
    }

    //! \brief Get address of shared memory variable.
    inline NUMTYPE *getAddress() const OME_ALWAYS_INLINE {
        return (reinterpret_cast<NUMTYPE *>(dataPointer));
    }

    //! \brief Cast operator for a pointer to a numeric shared memory variable.
    inline explicit operator const NUMTYPE *() const OME_ALWAYS_INLINE {
        return (reinterpret_cast<const NUMTYPE *>(dataPointer));
    }

    //! \brief Assignment operator for a numeric shared memory variable.
    inline NUMTYPE &operator=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) = arg);
    }

    //! \brief Addition operator for a numeric shared memory variable.
    inline NUMTYPE &operator+=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) += arg);
    }

    //! \brief Subtraction operator for a numeric shared memory variable.
    inline NUMTYPE &operator-=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) -= arg);
    }

    //! \brief Multiplication operator for a numeric shared memory variable.
    inline NUMTYPE &operator*=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) *= arg);
    }

    //! \brief Division operator for a numeric shared memory variable.
    inline NUMTYPE &operator/=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) /= arg);
    }

    //! \brief Prefix increment operator for a numeric shared memory variable.
    inline NUMTYPE &operator++() OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) += 1);
    }
    //! \brief Prefix decrement operator for a numeric shared memory variable.
    inline NUMTYPE &operator--() OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) -= 1);
    }

    //! \brief Postfix increment operator for a numeric shared memory variable.
    inline NUMTYPE operator++(int) OME_ALWAYS_INLINE {
        NUMTYPE result = *reinterpret_cast<NUMTYPE *>(dataPointer);
        *reinterpret_cast<NUMTYPE *>(dataPointer) += 1;
        return (result);
    }
    //! \brief Postfix decrement operator for a numeric shared memory variable.
    inline NUMTYPE operator--(int) OME_ALWAYS_INLINE {
        NUMTYPE result = *reinterpret_cast<NUMTYPE *>(dataPointer);
        *reinterpret_cast<NUMTYPE *>(dataPointer) -= 1;
        return (result);
    }
}; // end class SMV_Numeric

/*! \brief Convenience template for creating shared memory variables of
 * one of the supported numeric types.  At the cost of additional
 * storage, these standalone variables can operate when the variable has
 * not been registered and allocated with a SharedMemoryVariableManager.
 */
template<typename NUMTYPE> class SMV_StandaloneNumeric:
    public SharedMemoryVariable
{
protected:
    SharedMemoryVariable::SharedMemoryVariable_Record     initialCopy;
public:
    explicit SMV_StandaloneNumeric(const char *varName,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager,
        NUMTYPE initVal=0) NONNULL_CLASS_PARAMETERS(2) :
        SharedMemoryVariable(mgr, varName,
                typeid(NUMTYPE) == typeid(int32_t) ? SMV_TYPE_INT32 :
                typeid(NUMTYPE) == typeid(uint32_t) ? SMV_TYPE_UINT32 :
                typeid(NUMTYPE) == typeid(int64_t) ? SMV_TYPE_INT64 :
                typeid(NUMTYPE) == typeid(uint64_t) ? SMV_TYPE_UINT64 :
                typeid(NUMTYPE) == typeid(float) ? SMV_TYPE_FLOAT :
                typeid(NUMTYPE) == typeid(double) ? SMV_TYPE_DOUBLE :
                ~0, /* default if unknown */
                sizeof(NUMTYPE))
    {
        // NOTE:  dataPointer will be set by base class SharedMemoryVariable,
        // which gets constructed first
        if (dataPointer == nullptr) {
            dataPointer = &initialCopy;
        }
        *this = initVal;
    }

    SMV_StandaloneNumeric(const char *varName,
        SharedMemoryVariableNode *parentNode, NUMTYPE initVal=0) :
        SharedMemoryVariable(parentNode, varName,
                typeid(NUMTYPE) == typeid(int32_t) ? SMV_TYPE_INT32 :
                typeid(NUMTYPE) == typeid(uint32_t) ? SMV_TYPE_UINT32 :
                typeid(NUMTYPE) == typeid(int64_t) ? SMV_TYPE_INT64 :
                typeid(NUMTYPE) == typeid(uint64_t) ? SMV_TYPE_UINT64 :
                typeid(NUMTYPE) == typeid(float) ? SMV_TYPE_FLOAT :
                typeid(NUMTYPE) == typeid(double) ? SMV_TYPE_DOUBLE :
                ~0, /* default if unknown */
                sizeof(NUMTYPE))
    {
        if (dataPointer == nullptr) {
            dataPointer = &initialCopy;
        }
        *this = initVal;
    }

    /* WARNING:  these all depend on SharedMemoryVariable_Record having
     * the numeric field at offset 0
     */
    //! \brief Cast operator for a numeric shared memory variable.
    inline operator NUMTYPE() const OME_ALWAYS_INLINE {
        return (*reinterpret_cast<const NUMTYPE *>(dataPointer));
    }

    //! \brief Get address of shared memory variable.
    inline NUMTYPE *getAddress() const OME_ALWAYS_INLINE {
        return (reinterpret_cast<NUMTYPE *>(dataPointer));
    }

    //! \brief Cast operator for a pointer to a numeric shared memory variable.
    inline explicit operator const NUMTYPE *() const OME_ALWAYS_INLINE {
        return (reinterpret_cast<const NUMTYPE *>(dataPointer));
    }

    //! \brief Assignment operator for a numeric shared memory variable.
    inline NUMTYPE &operator=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) = arg);
    }

    //! \brief Addition operator for a numeric shared memory variable.
    inline NUMTYPE &operator+=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) += arg);
    }

    //! \brief Subtraction operator for a numeric shared memory variable.
    inline NUMTYPE &operator-=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) -= arg);
    }

    //! \brief Multiplication operator for a numeric shared memory variable.
    inline NUMTYPE &operator*=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) *= arg);
    }

    //! \brief Division operator for a numeric shared memory variable.
    inline NUMTYPE &operator/=(const NUMTYPE arg) OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) /= arg);
    }

    //! \brief Prefix increment operator for a numeric shared memory variable.
    inline NUMTYPE &operator++() OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) += 1);
    }
    //! \brief Prefix decrement operator for a numeric shared memory variable.
    inline NUMTYPE &operator--() OME_ALWAYS_INLINE {
        return (*reinterpret_cast<NUMTYPE *>(dataPointer) -= 1);
    }

    //! \brief Postfix increment operator for a numeric shared memory variable.
    inline NUMTYPE operator++(int) OME_ALWAYS_INLINE {
        NUMTYPE result = *reinterpret_cast<NUMTYPE *>(dataPointer);
        *reinterpret_cast<NUMTYPE *>(dataPointer) += 1;
        return (result);
    }
    //! \brief Postfix decrement operator for a numeric shared memory variable.
    inline NUMTYPE operator--(int) OME_ALWAYS_INLINE {
        NUMTYPE result = *reinterpret_cast<NUMTYPE *>(dataPointer);
        *reinterpret_cast<NUMTYPE *>(dataPointer) -= 1;
        return (result);
    }

    virtual void noteNowRegistered(SharedMemoryVariable_Record *newArea) override {
        // copy any existing counts into allocated storage area
        *reinterpret_cast<NUMTYPE *>(newArea) =
            *reinterpret_cast<NUMTYPE *>(dataPointer);
    }
}; // end class SMV_StandaloneNumeric


/*! \brief Convenience class for string variables.
 */
class SMV_String:
    public SharedMemoryVariable
{
public:
    SMV_String(const char *varName, size_t reservedLength,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager) NONNULL_CLASS_PARAMETERS(2) :
        SharedMemoryVariable(mgr, varName, SMV_TYPE_STRING, reservedLength) {};

    SMV_String(const char *varName, size_t reservedLength,
        SharedMemoryVariableNode *parentNode) NONNULL_CLASS_PARAMETERS(2) :
        SharedMemoryVariable(parentNode, varName, SMV_TYPE_STRING,
            reservedLength) {};

    //! \brief Cast operator for a shared memory string variable.
    inline operator const char *() const OME_ALWAYS_INLINE {
        return (dataPointer->varName + dataPointer->varNameLength);
    };

    //! \brief Assignment operator into a shared memory string variable.
    inline char *operator=(const char *val) OME_ALWAYS_INLINE {
        // TODO:  either externalize the overhead of a record or make
        // allocatedLength not include any recording keeping overhead.
        // The FixedBufferManager prefixes the data block with the allocation
        // record, whereas the CircularBufferManager keeps allocation records
        // separate from data records.
        const uint32_t headerOverhead = (sizeof(SharedMemoryVariable_Record) -
            sizeof(dataPointer->varName)) +
            sizeof(SharedBufferAllocRecord); // TODO:  this depends on storage mode
        size_t avail = dataPointer->allocatedLength -
            (headerOverhead + dataPointer->varNameLength);
        char *d = dataPointer->varName + dataPointer->varNameLength;
        char *n = (char *) memccpy(d, val, '\0', avail);
        if (OME_EXPECT_FALSE(n == nullptr)) {
            dataPointer->usedLength = avail;
        } else {
            dataPointer->usedLength = (n - d);
        }
        return (d);
    }
}; // end class SMV_String


/*! \brief Convenience class for string variables
 * that can be also be used if a shared memory segment
 * is never attached.
 *
 * \param MAXLEN specifies the maximum length of the string.
 *
 * \note This implementation is templated to eliminate the need for any
 * heap access.  Users should minimize the number of distinct sizes
 * in play to prevent a possible explosion of implementation routines.
 */
template <size_t MAXLEN> class SMV_StandaloneString:
    public SharedMemoryVariable
{
protected:
    union {
        SharedMemoryVariable::SharedMemoryVariable_Record     initialCopy;
        char reservedSpace[MAXLEN + sizeof(SharedMemoryVariable::SharedMemoryVariable_Record)];
    } space;
public:
    explicit SMV_StandaloneString(const char *varName,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager) :
        SharedMemoryVariable(mgr, varName, SMV_TYPE_STRING, MAXLEN)
    {
        if (dataPointer == nullptr) {
            memset(&space, 0, sizeof(space));
            dataPointer = &space.initialCopy;
            safe_strcpy(dataPointer->varName, varName, sizeof(dataPointer->varName));
			dataPointer->varNameLength = static_cast<uint16_t>(strlen(varName) + 1);
            dataPointer->allocatedLength = sizeof(space);
        }
    }

    SMV_StandaloneString(const char *varName, SharedMemoryVariableNode *parentNode) :
        SharedMemoryVariable(parentNode, varName, SMV_TYPE_STRING, MAXLEN)
    {
        if (dataPointer == nullptr) {
            memset(&space, 0, sizeof(space));
            dataPointer = &space.initialCopy;
            dataPointer->varNameLength = strlen(varName) + 1;
            dataPointer->allocatedLength = sizeof(space);
        }
    }

    //! \brief Cast operator for a shared memory string variable.
    inline operator const char *() const OME_ALWAYS_INLINE {
        return (dataPointer->varName + dataPointer->varNameLength);
    };

    //! \brief Assignment operator into a shared memory string variable.
    inline char *operator=(const char *val) OME_ALWAYS_INLINE {
        // TODO:  either externalize the overhead of a record or make
        // allocatedLength not include any recording keeping overhead.
        // The FixedBufferManager prefixes the data block with the allocation
        // record, whereas the CircularBufferManager keeps allocation records
        // separate from data records.
        const uint32_t headerOverhead = (sizeof(SharedMemoryVariable_Record) -
            sizeof(dataPointer->varName)) +
            sizeof(SharedBufferAllocRecord); // TODO:  this depends on storage mode
        size_t avail = dataPointer->allocatedLength -
            (headerOverhead + dataPointer->varNameLength);
        char *d = dataPointer->varName + dataPointer->varNameLength;
        char *n = (char *) memccpy(d, val, '\0', avail);
        if (OME_EXPECT_FALSE(n == nullptr)) {
            dataPointer->usedLength = avail;
        } else {
            dataPointer->usedLength = (n - d);
        }
        return (d);
    }

    virtual void noteNowRegistered(SharedMemoryVariable_Record *newArea) override {
        // copy any string into allocated storage area
        const uint32_t headerOverhead = (sizeof(SharedMemoryVariable_Record) -
            sizeof(dataPointer->varName)) +
            sizeof(SharedBufferAllocRecord); // TODO:  this depends on storage mode
        size_t avail = newArea->allocatedLength -
            (headerOverhead + newArea->varNameLength);
        char *d = newArea->varName + newArea->varNameLength;
        char *s = dataPointer->varName + dataPointer->varNameLength;
        char *n = (char *) memccpy(d, s, '\0', avail);
        if (OME_EXPECT_FALSE(n == nullptr)) {
            newArea->usedLength = avail;
        } else {
            newArea->usedLength = (n - d);
        }
    }
}; // end class SMV_StandaloneString:

/*! \brief Subclass of SharedMemoryVariableNode that reserves heap
 * space to hold data when a shared memory segment is not in use.
 *
 * \param MAXLEN specifies the maximum length of the node name.
 *
 * \note This implementation is templated to eliminate the need for any
 * heap access.  Users should minimize the number of distinct sizes
 * in play to prevent a possible explosion of implementation routines.
 */
template <size_t MAXLEN> class SMV_StandaloneNode :
    public SharedMemoryVariableNode
{
protected:
    char reservedSpace[MAXLEN];

public:
    explicit SMV_StandaloneNode(const char *varName,
        SharedMemoryVariableManager *mgr=&DEFAULT_sharedMemoryVariableManager) :
        SharedMemoryVariableNode(varName, mgr)
    {
        safe_strcpy(reservedSpace, varName, sizeof(reservedSpace));
        setName(reservedSpace);
    }

    SMV_StandaloneNode(const char *varName,
        SharedMemoryVariableNode *parentNode) :
        SharedMemoryVariableNode(varName, parentNode)
    {
        safe_strcpy(reservedSpace, varName, sizeof(reservedSpace));
        setName(reservedSpace);
    }
}; // end class SMV_StandaloneNode

/*! \brief Convenience macro to declare a shared memory counter variable.
 *
 * Note:  no support for alternate segment managers or parent naming nodes.
 * Best used for very simple instrumentation.
 */
#define DECLARE_SMV_COUNTER(name) SMV_Numeric<uint32_t> name(#name)


/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
