#ifndef _TIME_POINT_HPP_
//! \cond
#define _TIME_POINT_HPP_ "$Id: time_point.hpp 456 2020-07-23 20:25:49Z geoff $"
//! \endcond

/* Copyright (C) 2010 - 2019 FARGOS Development, LLC */

#include <utils/base/compiler_hints.h>
#include <utils/base/text2int.h>
#include <utils/base/timed_mutex.hpp>
#include <utils/time/tick_time.h>
#include <utils/time/get_time.h>
#include <iostream>
#include <stdio.h>
#include <time.h>
#include <stdint.h>
#include <string.h>
#include <map>

/*! \file
 * \brief FARGOS Time Functions.
 *
 * The routines implemented by time_point.hpp are able to deal with
 * arbitrary granularity of clocks and convert between locales.
 *
 * When done correctly, and it rarely is, the time representation
 * used by an application should be an unambiguous value.
 * Ambiguity should only be introduced when converting to a locale.
 * For example, 9pm Eastern in the United States represents the same
 * moment in time as 6pm Pacific, but are named differently.
 *
 * \note Never choose a coarser granularity than needed.  It is expensive
 * to convert from finer-granularity precision to a coarser precision,
 * but very cheap to scale a coarse granularity clock to an apparently
 * finer precision.  For that reason, nanoseconds is the suggested
 * normal precision that should be used.
 *
 * \copyright 2010 - 2019 FARGOS Development, LLC
 * \author Geoff Carpenter
 */
/** \defgroup FARGOStime FARGOS Time Functions
 *
 * \brief High-performance unambiguous time representations of arbitrary
 * granularity, cache-assisted conversions from native representation to
 * locale-specific interpretation.
 * @{
 */


#ifdef _WIN32
#define CALL_TZSET()  if (TimePointCache::TimePoint_setTZ == false) TimePointCache::doSetTZ()
#else
#define CALL_TZSET()  do { } while (0) /* null statement */
#endif

template <unsigned int UNITS> extern CONSTEXPR unsigned int int_digitsNeeded() OME_CONST_FUNCTION OME_ALWAYS_INLINE;

/*! \brief Via compile-time evaluation, compute log10().
 * Used to determine number digits required to display precision.
 */
template <unsigned int UNITS> inline CONSTEXPR unsigned int int_digitsNeeded()
{
    return (1 + int_digitsNeeded<UNITS / 10>());
}

//! \\brief Special case for unit precision of 1--no decimal point needed
template <> inline CONSTEXPR unsigned int int_digitsNeeded<1>()
{
    return (0);
}

//! \brief Termination condition for safety
template <> inline CONSTEXPR unsigned int int_digitsNeeded<0>()
{
    return (0);
}

/*! \brief Cache locale-specific time data around a particular
 * point in time.
 *
 * TimePointCache objects maintain information about the local midnight,
 * end of day, and 3am of a specific day to enable
 * efficient manipulation of times that occur within that day.
 * While their use is optional, appropriately constructed TimePointCache
 * objects enable expensive library calls to be avoided when dealing with
 * times that occur within a particular day.  Since more than one
 * TimePointCache can be created, multiple threads can be handling
 * information from different days without creating interference.
 *
 * \sa TimePointCacheTable
 */
class TimePointCache {
public:
#ifdef _WIN32
    static bool TimePoint_setTZ;    //!< flag to indicate if _tzset() was called

    //! \brief Call _tzset() if needed
    static void doSetTZ();
#endif
    struct tm   localMidnight;
    struct tm   localEndOfDay;
    struct tm   local3am;   // sufficiently skewed to avoid any DST issues
    struct tm   lastTime;
    struct tm   utc3am;
    time_t      localMidnightEpoch;
    time_t      localEndOfDayEpoch;
    time_t      local3amEpoch;
    time_t      lastTimeEpoch;
    int32_t     utcDifferenceSeconds;
    
    // not needed ~TimePointCache()   {}

    //! \brief Null constructor that does no work
    TimePointCache()    {
        localEndOfDayEpoch = 0;  // force isForDay() to be false
    }

    /*! \brief Create a TimePointCache based around a specific point in time.
     *
     * \param asOfPOSIXepoch indicates the desired point in time based
     * on the POSIX epoch.  If the value of 0 is specified, the current
     * time is retrieved using clock_gettime_via_tick().
     */
    explicit TimePointCache(const time_t asOfPOSIXepoch)
    {
        localEndOfDayEpoch = 0;  // force isForDay() to be false
        initializeForDay(asOfPOSIXepoch);
    }

    /*! \brief Create a TimePointCache around a time expressed as
     * a \c timespec, which is represented using the POSIX epoch.
     * The resulting expanded times will be relative to the local
     * time.
     *
     * \note The \a asIfOnDay structure is passed by reference since
     * the address of the \a tv_sec member is taken.
     */
    explicit TimePointCache(const struct timespec &asIfOnDay)
    {
        initializeForDay(asIfOnDay);
    }

    /*! \brief Create a TimePointCache based around a local expanded
     * time.
     */
    explicit TimePointCache(const struct tm &asIfOnDay)
    {
        initializeForDay(asIfOnDay);
    }

    /*! \brief Initialize a TimePointCache around an already expanded
     * time, which is intended to reflect a local time.
     */
    void initializeForDay(const struct tm &asIfOnDay)
    {
        lastTime = asIfOnDay;
        localMidnight = asIfOnDay;
        localMidnight.tm_sec = 0;
        localMidnight.tm_min = 0;
        localMidnight.tm_hour = 0;
        localMidnight.tm_isdst = -1;    // unknown
        localEndOfDay = localMidnight;
        localEndOfDay.tm_hour = 24;
        local3am = localMidnight;
        local3am.tm_hour = 3;
    
        localMidnightEpoch = mktime(&localMidnight);
        localEndOfDayEpoch = mktime(&localEndOfDay);
        local3amEpoch = mktime(&local3am);
        lastTimeEpoch = mktime(&lastTime);
        // convert our 3am time to UTC
#ifdef _WIN32
        gmtime_s(&utc3am, &local3amEpoch);
#else
        gmtime_r(&local3amEpoch, &utc3am);
#endif
        // use standard semantics of zones like Eastern being negative
        utcDifferenceSeconds = (local3am.tm_hour - utc3am.tm_hour) * (60 * 60);
        // TODO:  handle fractional zone if such a deployment scenario arises
    }

    /*! \brief Initialize a TimePointCache around a time specified
     * using a \c timespec, which is based on the POSIX epoch.  The
     * resulting expanded times will be relative to the local time.
     *
     * \note The \a asIfOnDay structure is passed by reference since
     * the address of the \a tv_sec member is taken.
     */
    void initializeForDay(const struct timespec &asIfOnDay)
    {
        struct tm   expandedFields;
#ifdef _WIN32
        CALL_TZSET();
        localtime_s(&expandedFields, &(asIfOnDay.tv_sec));
#else
        localtime_r(&(asIfOnDay.tv_sec), &expandedFields);
#endif
        initializeForDay(expandedFields);
    }

    bool isForDay(const time_t t) const OME_ALWAYS_INLINE {
        /* NOTE:  localEndOfDayEpoch == 0 is used to force
         * this to return false.  In practice, most times
         * will be before the end-of-day, so we check it last.
         */
        if ((t >= localMidnightEpoch) && (t < localEndOfDayEpoch)) {
            return (true);
        }
        return (false);
    }

    /*! \brief Initialize a TimePointCache based around a
     * specific point in time.
     *
     * \param asOfPOSIXepoch indicates the desired point in time based
     * on the POSIX epoch.  If the value of 0 is specified, the current
     * time is retrieved using clock_gettime_via_tick().
     *
     * The heavy overhead of initialization will not be performed
     * again if the TimePointCache object is already representing the
     * indicated day.
     */
    void initializeForDay(const time_t asOfPOSIXepoch)
    {
        if (isForDay(asOfPOSIXepoch) == true) {
            return; // assume already initialized
        }
        struct timespec   todayTime;
        if (asOfPOSIXepoch == 0) { // was not set...
            // These objects should be infrequently created, so we force
            // a kernel call to retrieve the time and thus avoid any dependency
            // on having the clock rate determined prior to this call.
            clock_gettime_via_tick(&todayTime, true);
        } else {
            todayTime.tv_sec = asOfPOSIXepoch;
            todayTime.tv_nsec = 0;
        }
        initializeForDay(todayTime);
    }

    int32_t offsetFromUTC() const OME_ALWAYS_INLINE {
       return (utcDifferenceSeconds);
    }

    bool isForDay(const struct tm &dayOfInterest) const {
        if ((local3am.tm_mday == dayOfInterest.tm_mday) &&
            (local3am.tm_mon == dayOfInterest.tm_mon) &&
            (local3am.tm_year == dayOfInterest.tm_year) ) {
            return (true);
        }
        return (false);
    }

}; // end class TimePointCache

//! \brief Cached TimePoint information representing today
extern TimePointCache TimePointCacheForToday;

/*! \brief Convenience class to maintain a table of
 * TimePointCache objects.
 *
 * Many applications make repeated references to times covering
 * a small population of dates.  Initialization of a TimePointCache
 * object is relatively expensive, often requiring a trip into the
 * kernel to examine the filesystem. A TimePointCacheTable can be
 * used to avoid creating redundant TimePointCache objects along
 * with the unproductive repeated work.
 */
class TimePointCacheTable {
private:
    TimedMutex                      tblLock;
    std::map<time_t,TimePointCache *> timeTable;
public:
    explicit TimePointCacheTable(const char *tableName="timeCache") :
        tblLock(tableName, (SharedMemoryVariableManager *) nullptr) // never allocate statistics for this 
    {}

    ~TimePointCacheTable()  {}

    /*! \brief Returns a TimePointCache entry for the day
     * indicated by the number of seconds since the POSIX epoch.
     *
     * \param t specifies the time in question
     *
     * \return a reference to the corresponding record is returned.
     *
     * NOTE: this routine is thread-safe.
     * Records are automatically created as needed.  The address
     * of the returned record will not change during the lifetime of the
     * table.
     */
    const TimePointCache &getCacheEntry(time_t t) {
        tblLock.LOCK_HERE();
        std::map<time_t,TimePointCache *>::const_iterator i = timeTable.find(t);
        if (i != timeTable.end()) {
            const TimePointCache *result = i->second;
            tblLock.unlock();
            return (*result);
        }
        TimePointCache *newEntry = new TimePointCache(t);
        timeTable[t] = newEntry; // save
        tblLock.unlock();
        return (*newEntry);
    }
}; // end class TimePointCacheTable

/*! \brief Representation of UNITS-since-POSIX-epoch.  This is an unambiguous
 * time reference; there is no locale-specific attribute.
 *
 * \note:  this template exists to be able to handle arbitrary
 * granularity of an incoming time unit.
 */
template <unsigned int UNITS> class POSIXtimeInUnits {
protected:
    uint64_t    time_in_units;  //!< holds total UNITS since POSIX epoch

    inline void assignFromTimeWithNanoseconds(const TimeWithNanoseconds time_ns,
            const int skewByMinutes=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        enum { one_billion = 1000000000UL }; // for readability
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
        if (UNITS == one_billion) { // test done at compile time
            time_in_units = (time_ns.time_sec * UNITS) + time_ns.time_nanosec +
                (static_cast<int64_t>(skewByMinutes) * (static_cast<int64_t>(60) * UNITS));
        } else {
            enum { NS_PER_UNIT = one_billion / UNITS }; // compile-time constant
            int64_t fractionalUnits;
            if (time_ns.time_nanosec < NS_PER_UNIT) { // avoid the division if we can
                fractionalUnits = 0;
            } else {
                fractionalUnits = time_ns.time_nanosec / NS_PER_UNIT;
            }
            uint64_t totalSeconds = time_ns.time_sec + (skewByMinutes * 60);
            time_in_units = (totalSeconds * UNITS) + fractionalUnits;
        }
    }

public:
    /* NOTE: functions are defined before they are used to ensure
     * they can be inlined.
     */
    
    /*! \brief Converts from a POSIXtimeInUnits to an arbitrary granularity.
     *
     * \param desired_granularity specifies the desired granularity.
     *
     * \note Common usage is for \a desired_granularity to be passed as
     * a constant at compile-time, which eliminates much of the code.
     */
    inline uint64_t convertToCustomEpochUnits(uint64_t desired_granularity=UNITS) const OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
        if (desired_granularity == UNITS) { // test done at compile-time
            return (time_in_units);
        }
        if (desired_granularity > UNITS) { // test done at compile time
            /* NOTE: assuming desired_granularity is a constant, then
             * if conversion is possible by via a simple integer
             * multiplication, the resulting code will be comprised of just
             * a multiplication by a constant; otherwise this code block
             * disappears entirely.
             */
            const uint64_t intRatio = desired_granularity / UNITS; // computed at compile time
            if ((desired_granularity - (UNITS * intRatio)) == 0) { // test done at compile time
               return (time_in_units * intRatio); 
            }
        } else { // intend to decrease precision
            /* NOTE: assuming desired_granularity is a constant and
             * the intent was to decrease the resolution by an integral
             * divisor, the code emitted will be comprised of just the
             * steps needed to do that logical division by an integer
             * constant.  An actual division operation may not be
             * emitted when the compiler knows of special sequences of
             * shifts, adds and multiplies that yield the intended
             * value without the need for an expensive division (which
             * can be 20-100 times slower).
             */
            const uint64_t intRatio = UNITS / desired_granularity; // computed at compile time
            if ((UNITS - (desired_granularity * intRatio)) == 0) { // test done at compile time
               return (time_in_units / intRatio); 
            }
        }
        /* This code will be emitted only when needed to handle arbitrary
         * or unknown-at-compile-time ratios.
         * It avoids overflow by extracting the number of seconds and then
         * converting the remaining fractional portion.
         */
        uint64_t secs = time_in_units / UNITS;
        uint64_t fraction = time_in_units - (secs * UNITS);
        uint64_t result = (secs * desired_granularity);
        if (fraction != 0) { // only bother with conversion if needed
            // now split out seconds and fractional units
            uint64_t new_fraction = (fraction * desired_granularity) / UNITS;
            result += new_fraction;
        }
        return (result);
    }


    /* NOTE:  this class intentionally has no destructor nor copy constructor.
     *
     * This is crucial for performance so that the values can be passed in
     * registers with no intervening code being invoked.
     *
     * Tempting as it may be to implement a templated assignment operator
     * to permit conversions from different unit of granularity,
     * this must be avoided.
     *
     * The workaround is to provide a templated cast operator.
     */

    /*! \brief Construct a POSIXtimeInUnits from seconds since the POSIX epoch.
     *
     * \param seconds specifies the number of seconds since the POSIX epoch.
     * \param fractional specifies the number of UNITS per second.
     * \param skewByMinutes is an optional parameter to allow the time to
     * be skewed by the specified number of minutes.  This is typically used
     * to shift timezones.
     *
     * \note If \a seconds is 0, \a fractional can specify the total number
     * of UNITS since the POSIX epoch.
     */
    POSIXtimeInUnits(time_t seconds, int64_t fractional, int skewByMinutes=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        /* most common usage is to pass seconds as 0
         * and fractional as total nanoseconds.  When inlined,
         * skewByMinutes * 60 becomes 0, the add of 0 seconds is still 0,
         * and 0 * UNITS is 0, resulting in secondsAsUnits being 0.
         * Adding 0 to fractional allows this to reduce to just an assignment.
         */
        uint64_t secondsAsUnits = (static_cast<uint64_t>(seconds) + (skewByMinutes * 60)) * UNITS;
        time_in_units = fractional + secondsAsUnits;
    }

    /*! \brief Construct a POSIXtimeInUnits from a TimeWithNanoseconds
     * structure.
     *
     * \param time_ns specifies the time as a TimeWithNanoseconds structure.
     * \param skewByMinutes is an optional parameter to allow the time to
     * be skewed by the specified number of minutes.  This is typically used
     * to shift timezones.
     */
    explicit POSIXtimeInUnits(const TimeWithNanoseconds time_ns, int skewByMinutes=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        assignFromTimeWithNanoseconds(time_ns, skewByMinutes);
    }


    /*! \brief Cast operator to convert to alternate unit precision.
     */
    template <unsigned int ALT_UNITS> operator POSIXtimeInUnits<ALT_UNITS>() const OME_ALWAYS_INLINE;

#if 0
    // specialization for when target units matches implementation
    inline operator POSIXtimeInUnits<UNITS>() const OME_ALWAYS_INLINE {
        return (*this);
    }
#endif

    /*! \brief Assignment operator from TimeWithNanoseconds structure.
     *
     * \param time_ns specifies the time as a TimeWithNanoseconds structure.
     */
    inline POSIXtimeInUnits<UNITS> & operator=(const TimeWithNanoseconds time_ns) OME_ALWAYS_INLINE {
        assignFromTimeWithNanoseconds(time_ns, 0);
        return (*this);
    }

    /*! \brief Convert a POSIXtimeInUnits to nanoseconds since the POSIX epoch.
     */
    inline uint64_t convertToNanosecondsSinceEpoch() const OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        enum { one_billion = 1000000000UL }; // for readability
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
        if (UNITS == one_billion) { // test done at compile time
            return (time_in_units);
        }
        uint64_t secs = time_in_units / UNITS;
        uint64_t fraction = time_in_units - (secs * UNITS);
        // now split out seconds and fractional units
        uint64_t new_fraction = (fraction * one_billion) / UNITS;
        uint64_t result = (secs * one_billion) + new_fraction;
        return (result);
    }

    /*! \brief Convert a POSIXtimeInUnits to seconds since the POSIX epoch.
     */
    inline time_t getSeconds() const OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3") {
        uint64_t seconds = time_in_units / UNITS;
        return (static_cast<time_t>(seconds));
    }

    /*! \brief Equality comparison against a POSIXtimeInUnits of
     * arbitrary granularity
     *
     * \param arg specifies the time to be compared
     */
    template <unsigned int ALT_UNITS> bool operator==(const POSIXtimeInUnits<ALT_UNITS> arg) const {
        return ((ALT_UNITS == UNITS) ? (time_in_units == arg.time_in_units) :
            (UNITS < ALT_UNITS) ? ((powerOf10Constant<ALT_UNITS - UNITS>() * time_in_units) == arg.time_in_units) :
            (time_in_units == (powerOf10Constant<UNITS - ALT_UNITS>() * arg.time_in_units)));
    }

    /*! \brief Inequality comparison against a POSIXtimeInUnits of
     * arbitrary granularity
     *
     * \param arg specifies the time to be compared
     */
    template <unsigned int ALT_UNITS> bool operator!=(const POSIXtimeInUnits<ALT_UNITS> arg) const {
        return ((ALT_UNITS == UNITS) ? (time_in_units != arg.time_in_units) :
            (UNITS < ALT_UNITS) ? ((powerOf10Constant<ALT_UNITS - UNITS>() * time_in_units) != arg.time_in_units) :
            (time_in_units != (powerOf10Constant<UNITS - ALT_UNITS>() * arg.time_in_units)));
    }

    /*! \brief Less-than-or-equal comparison against a POSIXtimeInUnits of
     * arbitrary granularity
     *
     * \param arg specifies the time to be compared
     */
    template <unsigned int ALT_UNITS> bool operator<=(const POSIXtimeInUnits<ALT_UNITS> arg) const {
        return ((ALT_UNITS == UNITS) ? (time_in_units <= arg.time_in_units) :
            (UNITS < ALT_UNITS) ? ((powerOf10Constant<ALT_UNITS - UNITS>() * time_in_units) <= arg.time_in_units) :
            (time_in_units <= (powerOf10Constant<UNITS - ALT_UNITS>() * arg.time_in_units)));
    }

    /*! \brief Greater-than-or-equal comparison against a POSIXtimeInUnits of
     * arbitrary granularity
     *
     * \param arg specifies the time to be compared
     */
    template <unsigned int ALT_UNITS> bool operator>=(const POSIXtimeInUnits<ALT_UNITS> arg) const {
        return ((ALT_UNITS == UNITS) ? (time_in_units >= arg.time_in_units) :
            (UNITS < ALT_UNITS) ? ((powerOf10Constant<ALT_UNITS - UNITS>() * time_in_units) >= arg.time_in_units) :
            (time_in_units >= (powerOf10Constant<UNITS - ALT_UNITS>() * arg.time_in_units)));
    }

    /*! \brief Less-than comparison against a POSIXtimeInUnits of
     * arbitrary granularity
     *
     * \param arg specifies the time to be compared
     */
    template <unsigned int ALT_UNITS> bool operator<(const POSIXtimeInUnits<ALT_UNITS> arg) const {
        return ((ALT_UNITS == UNITS) ? (time_in_units < arg.time_in_units) :
            (UNITS < ALT_UNITS) ? ((powerOf10Constant<ALT_UNITS - UNITS>() * time_in_units) < arg.time_in_units) :
            (time_in_units < (powerOf10Constant<UNITS - ALT_UNITS>() * arg.time_in_units)));
    }

    /*! \brief Greater-than comparison against a POSIXtimeInUnits of
     * arbitrary granularity
     *
     * \param arg specifies the time to be compared
     */
    template <unsigned int ALT_UNITS> bool operator>(const POSIXtimeInUnits<ALT_UNITS> arg) const {
        return ((ALT_UNITS == UNITS) ? (time_in_units > arg.time_in_units) :
            (UNITS < ALT_UNITS) ? ((powerOf10Constant<ALT_UNITS - UNITS>() * time_in_units) > arg.time_in_units) :
            (time_in_units > (powerOf10Constant<UNITS - ALT_UNITS>() * arg.time_in_units)));
    }

    /*! \brief Get the current time as units-since-the-POSIX epoch.
     *
     * \note This is a static routine.  Uses getCurrentTime() from get_time.h.
     */
    static inline POSIXtimeInUnits<UNITS> getCurrentTime() OME_ALWAYS_INLINE {
        TimeWithNanoseconds curTime;

        getCurrentTime(&curTime, false);
        return (POSIXtimeInUnits<UNITS>(curTime));
    }
}; // end class POSIXtimeInUnits<>


template <unsigned int UNITS> template<unsigned int ALT_UNITS> inline POSIXtimeInUnits<UNITS>::operator POSIXtimeInUnits<ALT_UNITS>() const
{
    return POSIXtimeInUnits<ALT_UNITS>((int64_t) 0, this->convertToCustomEpochUnits(ALT_UNITS));
}

/*! \brief Convenience typedef for nanosecond-resolution time.
 */
typedef POSIXtimeInUnits<1000000000U> POSIXtimeInNanoseconds;

/*! \brief Convenience typedef for conventional 1-second resolution time.
 *
 * \note This typedef is directly equivalent to a standard time_t.
 *
 */
typedef POSIXtimeInUnits<1U> POSIXtimeInSeconds;

/*! \brief Representation of localized expanded UNITS-since-epoch.
 *
 * \note This template exists only to be able to handle arbitrary
 * granularity of an incoming time unit.
 * Ideally, it would be implemented to consume nanoseconds and variable
 * granularity would not be an issue.
 */
template <unsigned int UNITS> class TimePointOfGranularity {
public:
    enum {
        OUTPUT_DATE=1,  // corresponds to LOG_DATESTAMP
        OUTPUT_TIME=2   // corresponds to LOG_TIMESTAMP
    };
    struct tm   expandedTime;       //!< Locale-specific time, conforms to POSIX, C89 and onwards
    uint64_t    fractionalUnits;    //!< Fractional seconds in granularity of UNITS-per-second

    // utility routines are defined before they are used
    // to ensure inlining is possible

    /*! \brief Initialize with seconds from POSIX epoch,
     * fractional component and optional skew from localtime.
     *
     * \param secondsSinceEpoch provides the number of seconds since
     * the POSIX epoch.
     * \param fractionInUnits specifies the number of fractional units in
     * a partial second.  Values greater than a whole second are
     * accepted.  Defaults to 0.
     * \param skewByMinutes indicates the number of minutes the expanded
     * time should be shifted from localtime.  Defaults to 0.
     * \param cachedTimeInfo provides an optional TimePointCache object
     * to enable the avoidance of expensive calculations if the time
     * in question is within the range supported by the TimePointCache object.
     */
    void initializeWithUnitsFromEpoch(time_t secondsSinceEpoch,
        uint64_t fractionInUnits, int skewByMinutes=0,
        const TimePointCache *cachedTimeInfo=&TimePointCacheForToday)
    {
        uint64_t extraSeconds, fraction;
        // only pay cost of division if needed...
        if (fractionInUnits < UNITS) {
            extraSeconds = 0;
            fraction = fractionInUnits;
        } else {
            extraSeconds = fractionInUnits / UNITS;
            fraction = fractionInUnits - (extraSeconds * UNITS);
        }
        time_t skewedTime = secondsSinceEpoch + extraSeconds + (skewByMinutes * 60);
        // if between 3am and midnight, hours move monotonically
        if ((cachedTimeInfo != nullptr) &&
            (skewedTime >= cachedTimeInfo->local3amEpoch) &&
            (skewedTime < cachedTimeInfo->localEndOfDayEpoch)) {
            // start with 3am time
            expandedTime = cachedTimeInfo->local3am;
            time_t total_seconds = skewedTime - cachedTimeInfo->local3amEpoch;
            time_t hours = total_seconds / (60 * 60); // 60 secs/min, 60 min/hour
            time_t totalMinSecs = total_seconds - (hours * (60 * 60));
            time_t mins = totalMinSecs / 60;
            time_t secs = totalMinSecs - (mins * 60);
/* Generic code:
            expandedTime->tm_hour += hours;
            expandedTime->tm_min += mins;
            expandedTime->tm_secs += secs;
*/
            expandedTime.tm_hour = hours + 3;  // specific for 03:00:00 offset
            expandedTime.tm_min = mins;
            expandedTime.tm_sec = secs;
        } else { // not in our cached window, do the heavy work to expand
#ifdef _WIN32
            CALL_TZSET();
            localtime_s(&expandedTime, &skewedTime);
#else
            localtime_r(&skewedTime, &expandedTime);
#endif
        }
        fractionalUnits = fraction;
    }

    /*! \brief Initialize with time specified as nanoseconds
     * from POSIX epoch.
     *
     * \param nanosecondsSinceEpoch provides the number of nanoseconds since
     * the POSIX epoch.  This can be the total number of nanoseconds or 
     * only the sub-second fraction.
     * \param secondsSinceEpoch provides the number of seconds since the
     * POSIX epoch.  If the complete number of nanoseconds was provided
     * by \a nanosecondsSinceEpoch, this will be 0.
     * \param cachedTimeInfo provides a TimePointCache object
     * to enable the avoidance of expensive calculations if the time
     * in question is within the range supported by the TimePointCache object.
     * \param skewByMinutes indicates the number of minutes the expanded
     * time should be shifted from localtime.  Defaults to 0.
     */
    inline void initializeWithNanosecondsFromEpoch(uint64_t nanosecondsSinceEpoch,
            time_t secondsSinceEpoch=0,
            const TimePointCache *cachedTimeInfo=&TimePointCacheForToday,
            int skewByMinutes=0) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3")
    {
        enum { one_billion = 1000000000UL }; // for readability

        // NOTE:  this test is evaluated at compile-time
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
        if (UNITS == one_billion) { // avoid work if possible...
            initializeWithUnitsFromEpoch(secondsSinceEpoch,
                    nanosecondsSinceEpoch, skewByMinutes, cachedTimeInfo);
        } else {
            enum { NS_PER_UNIT = one_billion / UNITS }; // compile-time constant
            uint64_t extraSeconds, nanoseconds;
            // only pay cost of division if needed...
            if (nanosecondsSinceEpoch < one_billion) {
                extraSeconds = 0;
                nanoseconds = nanosecondsSinceEpoch;
            } else {
                extraSeconds = nanosecondsSinceEpoch / one_billion;
                nanoseconds = nanosecondsSinceEpoch - (extraSeconds * one_billion);
            }
            time_t totalSeconds = secondsSinceEpoch + extraSeconds;

            uint64_t unit_fraction = nanoseconds / NS_PER_UNIT;
            initializeWithUnitsFromEpoch(totalSeconds, unit_fraction,
                skewByMinutes, cachedTimeInfo);
        }
    }


    //! \brief Null constructor, leaves in undefined state
    TimePointOfGranularity()    {}

    /*! \brief Construct from standard broken-out time structure
     *
     * \param explodedTime provides the desired locale-specific time.
     * \param fraction is an optional parameter that specifies the fraction of
     * a second in a granularity of UNITS.
     */
    explicit TimePointOfGranularity(const struct tm &explodedTime, uint64_t fraction=0)
    {
        expandedTime = explodedTime;
        fractionalUnits = fraction;
    }

    /*! \brief Construct from arbitrary POSIXtimeInUnits<> object.
     *
     * \param ptTime specifies an unambiguous time specified as
     * POSIXtimeInUnits with granularity PT_UNITS.
     * \param cachedTimeInfo provides an TimePointCache object
     * to enable the avoidance of expensive calculations if the time
     * in question is within the range supported by the TimePointCache object.
     * \param skewByMinutes is an optional parameter that allows skewing
     * the provided time by the specified number of minutes.
     */
    template <unsigned int PT_UNITS>
    explicit TimePointOfGranularity(const POSIXtimeInUnits<PT_UNITS> ptTime,
            const TimePointCache *cachedTimeInfo=&TimePointCacheForToday,
            int skewByMinutes=0)
    {
        // NOTE:  when UNITS == PT_UNITS, this just a copies a uint64_t
        uint64_t fract = ptTime.convertToCustomEpochUnits(UNITS);
        // NOTE: "this->" usage needed since this is a template
        this->initializeWithUnitsFromEpoch(0, fract, skewByMinutes, cachedTimeInfo);
    }

#if 0 // a non-default copy constructor creates unwanted overhead
    //! \brief Copy constructor
    TimePointOfGranularity(const TimePointOfGranularity<UNITS> &arg)
    {
        expandedTime = arg.expandedTime;
        fractionalUnits = arg.fractionalUnits;
    }
#endif

    /*! \brief Construct from text string specifying date and time
     *
     * \param dayAndTime specifies the date and time as a single contiguous
     * string.
     * \param cachedTimeInfo points to an optional TimePointCache that can be
     * used to accelerate conversions if the date being converted corresponds
     * to the TimePointCache.
     */
    explicit TimePointOfGranularity(const char *dayAndTime,
            const TimePointCache *cachedTimeInfo=&TimePointCacheForToday) {
        convertDayAndTime(dayAndTime, cachedTimeInfo);
    }

    /*! \brief Construct from text string specifying date and time
     *
     * \param day specifies the date
     * \param timeOfDay specifies the time of day (hours, minutes, seconds)
     * \param cachedTimeInfo is an optional TimePointCache that can be
     * used to accelerate conversions if the date being converted corresponds
     * to the TimePointCache.
     */
    TimePointOfGranularity(const char *day, const char *timeOfDay,
            const TimePointCache *cachedTimeInfo=&TimePointCacheForToday) {
        convertDayAndTime(day, timeOfDay, cachedTimeInfo);
    }

    /*! \brief Construct from seconds from POSIX epoch and optional
     * fractional component and skew from localtime.
     *
     * \param secondsSinceEpoch provides the number of seconds since
     * the POSIX epoch.
     * \param fraction specifies the number of fractional units in
     * a partial second.  Defaults to 0.
     * \param skewByMinutes indicates the number of minutes the expanded
     * time should be shifted from localtime.  Defaults to 0.
     * \param cachedTimeInfo provides an optional TimePointCache object
     * to enable the avoidance of expensive calculations if the time
     * in question is within the range supported by the TimePointCache object.
     */
    explicit TimePointOfGranularity(time_t secondsSinceEpoch,
        uint64_t fraction=0, int skewByMinutes=0,
        const TimePointCache *cachedTimeInfo=&TimePointCacheForToday)
    {
        initializeWithUnitsFromEpoch(secondsSinceEpoch, fraction,
            skewByMinutes, cachedTimeInfo);
    }


    //! \brief assignment operator from a TimePointOfGranularity object.
    template <unsigned int INCOMING_UNITS=UNITS> TimePointOfGranularity<UNITS> &operator=(const TimePointOfGranularity<INCOMING_UNITS> &arg) {
        expandedTime = arg.expandedTime;
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
        if (INCOMING_UNITS == UNITS) { // NOTE: evaluated at compile time
            fractionalUnits = arg.fractionalUnits;
        } else {
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
            if (UNITS >= INCOMING_UNITS) { // evaluated at compile time
                // scale up
                fractionalUnits = arg.fractionalUnits * powerOf10Constant<UNITS - INCOMING_UNITS>();
            } else { // reduce resolution
                fractionalUnits = arg.fractionalUnits / powerOf10Constant<INCOMING_UNITS - UNITS>();
            }
        }
        return (*this);
    }

    //! \brief assignment operator from a \b tm structure.
    TimePointOfGranularity<UNITS> &operator=(const struct tm &POSIXexpandedDate) {
        expandedTime = POSIXexpandedDate;
        fractionalUnits = 0;
        return (*this);
    }

    /*! \brief Cast operator to return reference to \a tm structure.
     */
    explicit operator const struct tm &() const OME_ALWAYS_INLINE {
        return (expandedTime);
    }

    /*! \brief Convert expanded time to nanoseconds since POSIX epoch
     *
     * \param cachedTimeInfo provides an optional TimePointCache object
     * to enable the avoidance of expensive calculations if the time
     * in question is within the range supported by the TimePointCache object.
     */
    uint64_t convertToNanosecondsSinceEpoch(const TimePointCache *cachedTimeInfo=&TimePointCacheForToday) const {
        enum { one_billion = 1000000000UL }; // for readability
        // if cached info provided and for exact same day and after 3am
        if ((cachedTimeInfo != nullptr) &&
            (expandedTime.tm_mday == cachedTimeInfo->local3am.tm_mday) &&
            (expandedTime.tm_mon == cachedTimeInfo->local3am.tm_mon) &&
            (expandedTime.tm_year == cachedTimeInfo->local3am.tm_year) &&
            (expandedTime.tm_hour >= 3)) {  // 3 = 3am
            // NOTE: for performance, we use the true constant values as
            // there's no need to read them from memory and the subtract zero
            // operations just collapse to logical assignments.
            // Implied field documented in comment to right.
            int hours = expandedTime.tm_hour - 3; // cachedTimedInfo->local3am.tv_hour;
            int mins = expandedTime.tm_min - 0; // cachedTimedInfo->local3am.tv_min;
            int secs = expandedTime.tm_sec - 0; // cachedTimedInfo->local3am.tv_sec;
#if 0 // By definition, 03:00:00 AM can't have these cases
            if (secs < 0) {
                secs += 60;
                mins -= 1;
            }
            if (mins < 0) {
                hours -= 1;
                mins += 60;
            }
#endif
            uint64_t seconds = hours * (60 * 60) + (mins * 60) + secs +
                cachedTimeInfo->local3amEpoch;
            uint64_t nanoseconds = (seconds * one_billion);
            if (fractionalUnits == one_billion) { // evaluated at compile-time
                nanoseconds += fractionalUnits;
            } else {
                nanoseconds += ((fractionalUnits * one_billion) / UNITS) ;
            }
            return (nanoseconds);
        }
        // otherwise, no valid cache help available, so resort to mktime().
#ifdef _WIN32
        CALL_TZSET();
#endif
        // NOTE:  in theory, mktime can fiddle with the tm_wday and
        // tm_yday fields, so we cast away the constness
        time_t seconds = mktime(const_cast<struct tm *>(&expandedTime));
        uint64_t nanoseconds = (seconds * one_billion);
        if (UNITS == one_billion) { // evaluated at compile-time
            nanoseconds += fractionalUnits;
        } else {
            nanoseconds += (fractionalUnits * one_billion / UNITS) ;
        }
        return (nanoseconds);
    }

    /*! \brief Convert expanded time to custom units since POSIX epoch.
     *
     * \param desired_granularity specifies the granularity of the
     * converted result.  It does not have to match the units used by
     * the TimePointOfGranularity representation.
     * \param cachedTimeInfo specifies a TimePointCache object that can be
     * used to accelerate the conversion of times within the date covered
     * by the cache object.
     *
     * \return the corresponding units of the desired granularity are returned.
     */
    uint64_t convertToCustomEpochUnits(uint64_t desired_granularity=UNITS,
            const TimePointCache *cachedTimeInfo=&TimePointCacheForToday) const {
        // if cached info provided and for exact same day and after 3am
        if ((cachedTimeInfo != nullptr) &&
            (expandedTime.tm_mday == cachedTimeInfo->local3am.tm_mday) &&
            (expandedTime.tm_mon == cachedTimeInfo->local3am.tm_mon) &&
            (expandedTime.tm_year == cachedTimeInfo->local3am.tm_year) &&
            (expandedTime.tm_hour >= 3)) {  // 3 = 3am
            // NOTE: for performance, we use the true constant values as
            // there's no need to read them from memory and the subtract zero
            // operations just collapse to logical assignments.
            // Implied field documented in comment to right.
            int hours = expandedTime.tm_hour - 3; // cachedTimedInfo->local3am.tv_hour;
            int mins = expandedTime.tm_min - 0; // cachedTimedInfo->local3am.tv_min;
            int secs = expandedTime.tm_sec - 0; // cachedTimedInfo->local3am.tv_sec;
#if 0 // By definition, 03:00:00 AM can't have these cases
            if (secs < 0) {
                secs += 60;
                mins -= 1;
            }
            if (mins < 0) {
                hours -= 1;
                mins += 60;
            }
#endif
            uint64_t seconds = hours * (60 * 60) + (mins * 60) + secs +
                cachedTimeInfo->local3amEpoch;
            uint64_t units = (seconds * desired_granularity);
#ifdef _MSC_VER
#pragma warning(suppress : 4127)
#endif
            if (desired_granularity == UNITS) { // if inlined, possibly evaluated at compile-time
                units += fractionalUnits;
            } else {
                units += ((fractionalUnits * desired_granularity) / UNITS) ;
            }
            return (units);
        }
#ifdef _WIN32
        CALL_TZSET();
#endif
        // NOTE:  in theory, mktime can fiddle with the tm_wday and tm_yday fields
        time_t seconds = mktime(const_cast<struct tm *>(&expandedTime));
        uint64_t units = (seconds * desired_granularity) +
            (fractionalUnits * desired_granularity / UNITS) ;
        return (units);
    }

    /*! \brief Get relative units within a day */
    uint64_t getRelativeUnitsInDay() const {
        uint64_t seconds = expandedTime.tm_sec + 
            expandedTime.tm_min * 60 +
            expandedTime.tm_hour * 60 * 60;
        uint64_t result = (seconds * UNITS) + fractionalUnits;
        return (result);
    }

    /*! \brief Get relative microseconds within a day */
    uint64_t getRelativeMicrosecondsInDay() const {
        enum { one_million = 1000000 }; // for readability
        uint64_t seconds = expandedTime.tm_sec + 
            expandedTime.tm_min * 60 +
            expandedTime.tm_hour * 60 * 60;
        uint64_t result = (seconds * one_million) +
            ((fractionalUnits * one_million) / UNITS);
        return (result);
    }

    /*! \brief Get relative nanoseconds within a day */
    uint64_t getRelativeNanosecondsInDay() const {
        enum { one_billion = 1000000000UL }; // for readability
        uint64_t seconds = expandedTime.tm_sec + 
            expandedTime.tm_min * 60 +
            expandedTime.tm_hour * 60 * 60;
        uint64_t result = (seconds * one_billion) +
            ((fractionalUnits * one_billion) / UNITS);
        return (result);
    }

    /*! \brief Get units per second represented by fractional component */
    static CONSTEXPR uint64_t getUnitsPerSecond() OME_CONST_FUNCTION OME_ALWAYS_INLINE {
        return (UNITS); // a constant, used for generic code
    }

    /*! \brief Convert string representing day and time into expanded time
     * structure.
     * \param dayAndTime points to a text string that holds a day and time.
     * \param cachedTimeInfo points to a TimePointCache object that can
     * be used to accelerate conversion of times that are in UTC.
     *
     * \note The following formats are recognized for days:
     * Y/M/D, Y-M-D, YYYYMMDD, Mon D Y
     *
     * The following formats are recognized for time:
     * HHMMSS[.fraction][Z] [{AM|PM}], H:M:S[.fraction][Z] [{AM|PM}]
     *
     * \todo Could tolerate asctime() output of "Day Mon D HH:MM:SS YYYY".
     * Except for the fractional component, this is generic and could
     * be converted to a standalone function that took the desired precision
     * as an argument.
     *
     * \return a pointer to the character after the last character processed
     * by the routine will be returned.  Often the last character processsed
     * will be the null character, but it could be whitespace.
     */
    const char *convertDayAndTime(const char *dayAndTime,
            const TimePointCache *cachedTimeInfo=&TimePointCacheForToday) {
        memset(&expandedTime, 0, sizeof(expandedTime)); // clear
        expandedTime.tm_isdst = -1; // assert we don't have a clue
        fractionalUnits = 0;
        const char *cp = dayAndTime;
        while (*cp == ' ') ++cp;    // skip any initial whitespace
        const char *fieldStart = cp;
        while ((*cp >= '0') && (*cp <= '9')) ++cp;
        uint32_t fieldLen = (uint32_t) (cp - fieldStart);
        // OTHER DATABASE TIME:  Mon N YYYY HH:MM:SS:fract{AM|PM}
        if (fieldLen == 8) { // YYYYMMDD
            expandedTime.tm_year = text2uint32(fieldStart, 4) - 1900;
            expandedTime.tm_mon = text2uint32(fieldStart + 4, 2) - 1;
            expandedTime.tm_mday = text2uint32(fieldStart + 6, 2);
            fieldStart = cp;
        } else if ((*cp == '/') || (*cp == '-')) { // YYYY-MM-DD
            expandedTime.tm_year = text2uint32(fieldStart, fieldLen);
            if (expandedTime.tm_year < 100) expandedTime.tm_year += 2000;
            expandedTime.tm_year -= 1900;
            fieldStart = ++cp;
            while ((*cp >= '0') && (*cp <= '9')) ++cp;
            fieldLen = (uint32_t) (cp - fieldStart);
            expandedTime.tm_mon = text2uint32(fieldStart, fieldLen) - 1;
            fieldStart = ++cp;
            while ((*cp >= '0') && (*cp <= '9')) ++cp;
            fieldLen = (uint32_t) (cp - fieldStart);
            expandedTime.tm_mday = text2uint32(fieldStart, fieldLen);
            fieldStart = cp;
        } else if (fieldLen == 0) { // hit characters at first...
            while ((*cp != ' ') && (*cp != '\0')) ++cp;
            fieldLen = (uint32_t) (cp - fieldStart);
            if (fieldLen < 3) { // don't recognize pattern
                return (fieldStart);
            }
            // NOTE:  AND with ~0x20 forces ASCII lowercase to uppercase
            // THIS WILLL NOT WORK ON EBCDIC; ~0x40 would be used and the
            // compare would be against the lowercase variant or 0x40 would
            // be ORd in to permit the compare against the uppercase letter.
            expandedTime.tm_mon = 127; // invalid value as flag
            switch (fieldStart[0] & ~0x20) {
            case 'A': // APR,AUG
                switch (fieldStart[1] & ~0x20) {
                case 'P': // APR
                    expandedTime.tm_mon = 4 - 1;
                    break;
                case 'U':   // AUG
                    expandedTime.tm_mon = 8 - 1;
                    break;
                }
                break;
            case 'D': // DEC
                expandedTime.tm_mon = 12 - 1;
                break;
            case 'F': // FEB
                expandedTime.tm_mon = 2 - 1;
                break;
            case 'J': // JAN,JUN,JUL
                switch (fieldStart[1] & ~0x20) {
                case 'A': // JAN
                    expandedTime.tm_mon = 1 - 1;
                    break;
                case 'U': // JUN,JUL
                    switch (fieldStart[2] & ~0x20) {
                        case 'L': // JUL
                            expandedTime.tm_mon = 7 - 1;
                            break;
                        case 'N': // JUN
                            expandedTime.tm_mon = 6 - 1;
                            break;
                    } // end switch third character
                    break;
                } // end switch second character
                break;
            case 'M': // MAR,MAY
                switch (fieldStart[1] & ~0x20) {
                case 'A':
                    switch (fieldStart[2] & ~0x20) {
                    case 'R':
                        expandedTime.tm_mon = 3 - 1;
                        break;
                    case 'Y':
                        expandedTime.tm_mon = 5 - 1;
                        break;
                    }
                    break;
                } // end switch third character
                break;
            case 'N': // NOV
                expandedTime.tm_mon = 11 - 1;
                break;
            case 'O': // OCT
                expandedTime.tm_mon = 10 - 1;
                break;
            case 'S': // SEP
                expandedTime.tm_mon = 9 - 1;
                break;
            } // end switch first character
            if (expandedTime.tm_mon == 127) { // failed to parse month name
                return (fieldStart);
            }
            while (*cp == ' ') ++cp;    // skip any whitespace after month name
            fieldStart = cp;
            while ((*cp >= '0') && (*cp <= '9')) ++cp;
            fieldLen = (uint32_t) (cp - fieldStart);
            expandedTime.tm_mday = text2uint32(fieldStart, fieldLen);
            while (*cp == ' ') ++cp;    // skip any initial whitespace
            fieldStart = cp;
            while ((*cp >= '0') && (*cp <= '9')) ++cp;
            fieldLen = (uint32_t) (cp - fieldStart);
            expandedTime.tm_year = text2uint32(fieldStart, fieldLen);
            expandedTime.tm_year -= 1900;
            fieldStart = cp;
        }

        if (cp == fieldStart) { // find end of field
            if ((*cp == '@') || (*cp == 'T')) {
                ++cp;   // skip @ in YYYY/MM/DD@HH:MM:SS.mmmm
            } else {
                while (*cp == ' ') ++cp;    // skip any initial whitespace
            }
            fieldStart = cp;
            while ((*cp >= '0') && (*cp <= '9')) ++cp;
        }
        fieldLen = (uint32_t) (cp - fieldStart);
        if (fieldLen == 6) { // HHMMSS
            expandedTime.tm_hour = text2uint32(fieldStart, 2);
            expandedTime.tm_min = text2uint32(fieldStart + 2, 2);
            expandedTime.tm_sec = text2uint32(fieldStart + 4, 2);
        } else if (fieldLen == 4) { // HHMM
            expandedTime.tm_hour = text2uint32(fieldStart, 2);
            expandedTime.tm_min = text2uint32(fieldStart + 2, 2);
            expandedTime.tm_sec = 0;    // implied
        } else if (fieldLen == 2) { // at least 3 fields with interfield separators
            expandedTime.tm_hour = text2uint32(fieldStart, fieldLen);
            if (*cp == ':') {
                fieldStart = ++cp; // skip over ":"
                while ((*cp >= '0') && (*cp <= '9')) ++cp;
                fieldLen = (uint32_t) (cp - fieldStart);
                expandedTime.tm_min = text2uint32(fieldStart, fieldLen);
                if (*cp == ':') {
                    fieldStart = ++cp; // skip over ":"
                    while ((*cp >= '0') && (*cp <= '9')) ++cp;
                    fieldLen = (uint32_t) (cp - fieldStart);
                    expandedTime.tm_sec = text2uint32(fieldStart, fieldLen);
                } else {
                    expandedTime.tm_sec = 0;
                }
            } else {
                    expandedTime.tm_min = 0;
                    expandedTime.tm_sec = 0;
            }
        }
        // One database format used HH:MM:SS:fractional{A|P}M
        if ((*cp == '.') || (*cp == ':')) { // fractional units specified
            fieldStart = ++cp;
            while ((*cp >= '0') && (*cp <= '9')) ++cp;
            fieldLen = (uint32_t) (cp - fieldStart);
            uint32_t fraction = text2uint32(fieldStart, fieldLen);
            int tensPlaces = fieldLen;
            const int digitsWanted = int_digitsNeeded<UNITS>();
            while (tensPlaces < digitsWanted) { // not enough precision
                fraction *= 10;
                tensPlaces += 1;
            }
            while (tensPlaces > digitsWanted) { // too much, drop precision
                fraction /= 10;
                tensPlaces -= 1;
            }
            fractionalUnits = fraction;
        }
        if (*cp == 'Z') { // was in UTC
            int32_t utcSkew;
            if (cachedTimeInfo->isForDay(expandedTime)) {
                utcSkew = cachedTimeInfo->offsetFromUTC();
            } else {
                TimePointCache infoForDay(expandedTime);
                utcSkew = infoForDay.offsetFromUTC();
            }

            int32_t hours = utcSkew / (60 * 60);
            expandedTime.tm_hour += hours;
            utcSkew -= hours * (60 *  60);
            int32_t minutes = utcSkew / 60;
            expandedTime.tm_min += minutes;
            // TODO:  check to see if tm_hour < 0 or > 23 and adjust tm_mday
            if (expandedTime.tm_hour < 0) {
                expandedTime.tm_hour += 24;
                expandedTime.tm_mday -= 1;
                if (expandedTime.tm_mday < 1) {
                    // TODO:  need to handle feb, etc.
                    expandedTime.tm_mday = 31;
                    expandedTime.tm_mon -= 1;
                    if (expandedTime.tm_mon < 0) {
                        expandedTime.tm_mon += 12;
                        expandedTime.tm_year -= 1;
                    }
                }
            }
            cp += 1; // skip 'Z'
            return (cp);
        }
        while (*cp == ' ') ++cp; // move to start of next field
        // Check for AM/PM suffix
        char c1 = *(cp) & ~0x20; // force to ASCII uppercase
        if ((c1 == 'P') && 
            ((*(cp + 1) & ~0x20) == 'M')) { // PM
            expandedTime.tm_hour += 12;
            cp += 2;
        } else if ((c1 == 'A') &&
            ((*(cp + 1) & ~0x20) == 'M')) { // AM
            cp += 2;
        }
        return (cp);
    }

    /*! \brief Convert strings representing date and time into expanded time
     * structure.
     * \param day points to a text string that holds a date.
     * \param timeOfDay points to a text string that holds a time.
     * \param cachedTimeInfo points to a TimePointCache object that can
     * be used to accelerate conversion of times that are in UTC.
     */
    int convertDayAndTime(const char *day, const char *timeOfDay,
        const TimePointCache *cachedTimeInfo=&TimePointCacheForToday) {
        char bfr[128];

        char *fieldEnd = (char *) memccpy(bfr, day, '\0', 32);
        if (fieldEnd == nullptr) {
            return (0);
        }
        *(fieldEnd - 1) = ' ';
        fieldEnd = (char *) memccpy(fieldEnd, timeOfDay, '\0', 32);
        if (fieldEnd == nullptr) {
            return (0);
        }

        const char *lastChar = convertDayAndTime(bfr, cachedTimeInfo);
        if (lastChar == bfr) {
            return (0);
        }
        return (1);
    }

    /*! \brief Format into text string.
     *
     * \param bfr specifies the output buffer,
     * \param bfrLen specified the length of available space
     * \param resultLen is an option pointer to a location into which
     * the length of the result can be stored.
     * \param fieldsWanted is a bit mask indicating the desired elements.
     * \param fieldSeparators is a 4-character string indicating the
     * separator elements between the year/month/day, hour:min:sec, fractional
     * and date vs. time groupings.  It defaults to "/:. ".  If one
     * of the individual elements is a null character, no separator will
     * be output.
     */
    char *formatIntoString(char *bfr, uint_fast8_t bfrLen,
        uint_fast8_t *resultLen=nullptr,
        uint_fast8_t fieldsWanted=OUTPUT_DATE|OUTPUT_TIME,
        const char *fieldSeparators="/:. ") const
    {
        uint_fast8_t offset;
        uint_fast8_t field2Offset, field3Offset;

        if (fieldsWanted & OUTPUT_DATE) {
            int y = expandedTime.tm_year; // years from 1900
            if (fieldSeparators[0] != '\0') {
                field2Offset = 5;
                field3Offset = 8;
                offset = 10;
                bfr[4] = fieldSeparators[0];
                bfr[7] = fieldSeparators[0];
            } else {
                field2Offset = 4;
                field3Offset = 6;
                offset = 8;
            }
            memcpy(bfr, numberAs4Digits[y + 1900].digits, 4);
            memcpy(bfr+field2Offset, numberAsTwoDigits[expandedTime.tm_mon + 1], 2);
            memcpy(bfr+field3Offset, numberAsTwoDigits[expandedTime.tm_mday], 2);
        } else {
            offset = 0;
        }
        if (fieldsWanted & OUTPUT_TIME) {
            uint32_t fractBfr[6];
            const int fractionalLen = int_digitsNeeded<UNITS>(); // compile-time constant
            if ((offset != 0) && (fieldSeparators[3] != '\0')) {
                // output date and want a separator
                bfr[offset] = fieldSeparators[3];
                offset += 1;
            }
            uint_fast8_t fieldLen;
            if (fieldSeparators[1] != '\0') {
                field2Offset = offset + 3;
                field3Offset = offset + 6;
                fieldLen = 8;
                bfr[offset+2] = fieldSeparators[1];
                bfr[offset+5] = fieldSeparators[1];
            } else {
                field2Offset = offset + 2;
                field3Offset = offset + 4;
                fieldLen = 6;
            }
            memcpy(bfr+offset, numberAsTwoDigits[expandedTime.tm_hour], 2);
            memcpy(bfr+field2Offset, numberAsTwoDigits[expandedTime.tm_min], 2);
            memcpy(bfr+field3Offset, numberAsTwoDigits[expandedTime.tm_sec], 2);
            offset += fieldLen;

            if (fieldSeparators[2] != '\0') {
                bfr[offset] = fieldSeparators[2];
                offset += 1;
            }

            const char *fractText = uint_to_ascii_right_justified(fractBfr,
                    sizeof(fractBfr), fractionalUnits, fractionalLen, '0');
            memcpy(bfr+offset, fractText, fractionalLen);
            offset += fractionalLen;
        }
        bfr[offset] = '\0';
        if (resultLen != nullptr) {
            *resultLen = offset;
        }
        return (bfr);
    }

    /*! \brief Compare the date portion of a TimePointOfGranularity
     *
     * \param arg specifies a TimePointOfGranularity of arbitrary granularity.
     *
     * The year, month, and day of the month are compared, not the time-of-day
     * component.
     */
    template <unsigned int ALT_UNITS> int compareDate(const TimePointOfGranularity<ALT_UNITS> &arg) const {
        int rc = expandedTime.tm_year - arg.expandedTime.tm_year;
        if (rc != 0) {
            return (rc);
        }
        rc = expandedTime.tm_mon - arg.expandedTime.tm_mon;
        if (rc != 0) {
            return (rc);
        }
        rc = expandedTime.tm_mday - arg.expandedTime.tm_mday;
        return (rc);
    }

}; // end class TimePointOfGranularity<>

//! \brief Desired standard granularity with nanosecond resolution.
typedef TimePointOfGranularity<1000000000> NanosecondTimePoint;

//! \brief Convenience typedef using conventional 1-second resolution.
typedef TimePointOfGranularity<1> POSIXEpochTimePoint;

/*! \brief Output TimePointOfGranularity as text onto a stream-like object */
template<typename STREAMTYPE, unsigned int UNITS>
inline STREAMTYPE & operator<<(STREAMTYPE &o, const TimePointOfGranularity<UNITS> &t)
{
    char    bfr[64];

    o << t.formatIntoString(bfr, sizeof(bfr));

    return (o);
}

/*! \brief Output POSIXtimeInUnits as text onto a stream-like object */
template <typename STREAMTYPE, unsigned int UNITS>
inline STREAMTYPE & operator<<(STREAMTYPE &o, const POSIXtimeInUnits<UNITS> pt)
{
    TimePointOfGranularity<UNITS> tp(pt);
    o << tp;
    return (o);
}

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
