#ifndef _OME_OID_H
//! \cond
#define _OME_OID_H "$Id: OMEoid.h 406 2020-07-12 05:07:33Z geoff $"
//! \endcond

/*! \file
 * \brief OME object identifier implementation.
 */

/** \addtogroup OMEtypes
 * @{
 */

#include <iostream> /* for debugging */
#    include <memory.h>


#    include <OMErefCount.h>
#    include <OMEassoc.h>
#    include <OMEprocInfo.h>

typedef class OMEbaseType *OMEbaseTypeP;


/*! \brief Fundamental elements of an OME object identifier.
 */
struct OMEoidID {
    uint32_t instanceIndex;
    uint32_t classVersionID;
    uint32_t classHashIndex;
    OMEentityID birthplace;
};

template <typename STREAMTYPE> inline STREAMTYPE &operator<<(STREAMTYPE &o, const OMEoidID &data)
{
    o << "[" << data.instanceIndex << ":";
    o << data.classVersionID << ":";
    o << data.classHashIndex << ":";
    o << data.birthplace;
    o << "]";
    return (o);
}

// defines abstract interface for OMEoid storage, hiding internal/external
/*! \brief Abstract interface for storage of OMEoid objects that
 * hides the internal vs. external implementation.
 */
class OMEoidStorageBase {
protected:
    OMEassoc acl;

    explicit OMEoidStorageBase(const OMEassoc &aclData) : acl(aclData)
    {
    }
public:

    virtual ~OMEoidStorageBase() {}

    virtual OMEstring *encodeOID(const class OMEencodeBuffer *bfr) const = 0;

    virtual void setProxyObject(const class OMEoid &proxy, const OMEstring &method) = 0;;

    virtual const class OMEoid *getProxyObject(const OMEstring **method) const = 0;

    virtual void getObjectOID(OMEoidID *result) const = 0;

    virtual const class OMEobject *getInstanceInMemory() const = 0;

    // NOTE: this was virtual, but never exploited
    const OMEassoc &getObjectACL() const
    {
        return (acl);
    }

}; // end class OMEoidStorageBase

/*! \brief Reference-counted link to OMEoidStorageBase object.
 */
class OMEoidStorage final : public OMEreferenceCount
{
    friend class OMEoid;
    friend class OMEoidStorageInternal;
    friend class OMEoidStorageExternal;
    friend class OMEreferenceToData<OMEoidStorage>;

public:
    enum OMErefType {
        _NOT_SET = 0,
        internal = 1,
        external = 5
    };
    template <typename STREAMTYPE> static STREAMTYPE &OMEdumpOIDchain(STREAMTYPE &outputStream);

    static OMEstring *OMEencodeOID(const OMEoidID &, const OMEassoc &,
                                   int encodingVersion);

    static int OMEdecodeOID(uint_fast32_t version,
                            const OMEstring *encodedData, size_t *offset,
                            OMEoidID *resultOID, OMEassoc **resultACL);

    // findReference implemented OMEoidInt.cpp/OMEoidIntStub.cpp
    static OMEoidStorage *findReference(const OMEoidID *id,
        const OMEassoc *acl = nullptr, OMEoidStorage **sameEntity = nullptr,
        bool externalOnly = false);

    static int convertExternalReferences(const OMEoidID *id, class OMEobject *o);

private:
    static OMEoidStorage *startOfChain;
    OMEoidStorage *nextInChain;
    OMEoidStorage *prevInChain;

    void addLink();

    void removeLink();

    // A null initializer is an illegal call, but is needed to complete
    // template for OMErefCount as enforced by Visual C++.
    OMEoidStorage()
    {
        // bogus call needed for VC++
        std::cerr << "OMEoidStorage default init\n";
        nextInChain = nullptr;
        prevInChain = nullptr;
    }

protected:
    class OMEoidStorageBase *storageRef;

    void convertInternalToExternalRef(const OMEoidID *entity,
                                      const class OMEoidStorageInternal *checkPtr);

public:
    OMErefType referenceType;

    // default initializer
    explicit OMEoidStorage(class OMEoidStorageInternal *ptr);

    explicit OMEoidStorage(class OMEoidStorageExternal *ptr);

    // destructor
    ~OMEoidStorage();

    OMEstring *encodeOID(const class OMEencodeBuffer *bfr) const
    {
        return (storageRef->encodeOID(bfr));
    }

    void getObjectOID(OMEoidID *arg) const
    {
        storageRef->getObjectOID(arg);
    }

    const OMEassoc &getObjectACL() const
    {
        return (storageRef->getObjectACL());
    }

    const class OMEoid *getProxyObject(const OMEstring **method) const
    {
        return (storageRef->getProxyObject(method));
    }

    void setProxyObject(const class OMEoid &proxy, const OMEstring &method)
    {
        storageRef->setProxyObject(proxy, method);
    }

    const class OMEobject *getInstanceInMemory() const
    {
        return (storageRef->getInstanceInMemory());
    }

    // SHOULD NEVER BE CALLED:  OID's are immutable...
    virtual OMEreferenceCount *deepCopy() const override
    {
        std::cerr << "OIDstorage::deepCopy called!\n";
        return (const_cast<OMEoidStorage *>(this));
    }

    bool operator==(const OMEoidStorage &arg) const;

    bool operator!=(const OMEoidStorage &arg) const;

    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0,
            bool fullInfo = false) const;

    bool objectIsLocal() const; // not implemented

    void convertExternalToInternalRef(OMEobject *b);
}; // end class OMEoidStorage


/*! \brief Public interface to an OME Object Identifier.
 */
class OME_DLL_EXPORT OMEoid final : public OMEreferenceToData<OMEoidStorage>
{
    friend class OMEapi;

public:
    static OMEoid *decodeOID(uint32_t ver, const OMEstring *encodedData,
                             size_t *offset);

private:
    explicit OMEoid(OMEoidStorage *s);

public:
    // NO DEFAULT INITIALIZER
    // copy initializer
    explicit OMEoid(const OMEoid &org) : OMEreferenceToData<OMEoidStorage>(org.data)
    {
//        data = org.data;
//        addReadOnlyReference();
    }

    // implemented in OMEoidInt.cpp
    OMEoid(class OMEobject *o, const OMEassoc &acl);

    ~OMEoid();

    OMEstring *encodeOID(const class OMEencodeBuffer *bfr) const
    {
        return (data->encodeOID(bfr));
    }

    // ASSIGNMENT
    OMEoid &operator=(const OMEoid &arg)
    {
        if (this->data == arg.data) {
            return (*this); // asignment to self...
        }
        arg.data->addReadOnlyReference();
        dropReference();
        data = arg.data;
        return (*this);
    }

    OMEoid &operator=(const OMEoid *arg)
    {
        (*this) = (*arg);
        return (*this);
    }

    // COMPARISON
    bool operator==(const OMEoid &arg) const
    {
        // if pointing at the same entity, must be equal...
        if (data == arg.data) {
            return (true);
        }
        return (*data == *arg.data);
    }

    bool operator!=(const OMEoid &arg) const
    {
        // if pointing at the same entity, must be equal...
        if (data == arg.data) {
            return (false);
        }
        return (*data != *arg.data);
    }

    OMEoid *deepCopy() const
    {
        // OIDs are immutable, so copy ourselves...
        OMEoid *copy = new OMEoid(*this);
        return (copy);
    }

    const class OMEoid *getProxyObject(const OMEstring **method) const
    {
        return (data->getProxyObject(method));
    }

    void setProxyObject(const OMEoid &proxy, const OMEstring &method)
    {
        data->setProxyObject(proxy, method);
    }

    const class OMEobject *getInstanceInMemory() const
    {
        return (data->getInstanceInMemory());
    }

    const OMEassoc &getObjectACL() const
    {
        return (data->getObjectACL());
    }

    void getObjectOID(OMEoidID *result) const
    {
        data->getObjectOID(result);
    }

    bool refersTo(const class OMEobject *obj) const OME_ALWAYS_INLINE
    {
        return (getInstanceInMemory() == obj);
    }

    bool isExternal() const OME_ALWAYS_INLINE
    {
        if (data->referenceType != OMEoidStorage::internal) {
            return (true);
        }
        return (false);
    }

    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0,
            bool fullInfo = false) const
    {
        data->outputOnStream(outputStream, indent, fullInfo);
        return (outputStream);
    }
}; // end class OMEoid

template <typename STREAMTYPE> inline STREAMTYPE &operator<<(STREAMTYPE &outputStream, const OMEoid &data)
{
    data.outputOnStream(outputStream, 0);
    return (outputStream);
}

/*! \brief Interface to an internal OME object.
 */
class OME_DLL_EXPORT OMEoidStorageInternal final :
    public OMEoidStorageBase
{
    friend class OMEobject;
    friend class OMEoidStorage;

private:
    const class OMEobject *objectRecord;
    // all internal references are linked in a chain originating from
    // the object they reference...
    OMEoidStorageInternal *nextRefToSameObject;
    OMEoidStorage *ownedBy;

    void setOwningRecord(OMEoidStorage *rec) OME_ALWAYS_INLINE
    {
        ownedBy = rec;
    }

    void noteReferencedObjectDeleted(OMEoidID *entity);

public:
    void doDestructorImplementation() OME_USED;

    OMEoidStorageInternal(OMEobject *o, const OMEassoc &a);

    ~OMEoidStorageInternal()
    {
        doDestructorImplementation();
    }

    virtual OMEstring *encodeOID(const class OMEencodeBuffer *bfr) const override;

    // paged out records will have objectProxy != InstanceInMemory
    virtual const OMEoid *getProxyObject(const OMEstring **method) const override;

    virtual void setProxyObject(const OMEoid &proxy, const OMEstring &method) override;
    virtual void getObjectOID(OMEoidID *) const override;

    bool operator==(const OMEoidStorageInternal &arg) const
    {
        if (objectRecord != arg.objectRecord) {
            return (false);
        }
        if (acl != arg.acl) {
            return (false);
        }
        return (true);
    }

    bool operator!=(const OMEoidStorageInternal &arg) const
    {
        if (objectRecord != arg.objectRecord) {
            return (true);
        }
        if (acl != arg.acl) {
            return (true);
        }
        return (false);
    }

    bool operator==(const OMEoidStorage &arg) const
    {
        if (arg.referenceType == OMEoidStorage::external) {
            return (false);
        }
        const OMEoidStorageInternal *p = dynamic_cast<const OMEoidStorageInternal *>(arg.storageRef);
        return (*this == *p);
    }

    bool operator!=(const OMEoidStorage &arg) const
    {
        if (arg.referenceType == OMEoidStorage::external) {
            return (true);
        }
        const OMEoidStorageInternal *p = dynamic_cast<const OMEoidStorageInternal *>(arg.storageRef);
        return (*this != *p);
    }

    virtual const class OMEobject *getInstanceInMemory() const override
    {
        return (objectRecord);
    }


    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0,
            bool fullInfo = false) const;
}; // end class OMEoidStorageInternal


/*! \brief Interface to an external OME object.
 */
class OME_DLL_EXPORT OMEoidStorageExternal final :
    public OMEoidStorageBase
{
    friend class OMEoidStorage;

private:
    OMEoidID oidData;
    OMEstring encapsulationMethodName;
    OMEoid *pagerObject;

public:
    OMEoidStorageExternal(const OMEoidID *data, const OMEassoc &acl);

    OMEoidStorageExternal(const OMEoidStorageExternal *existing, const OMEassoc &acl);

    ~OMEoidStorageExternal();

    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0,
            bool fullInfo = false) const;

    virtual OMEstring *encodeOID(const class OMEencodeBuffer *bfr) const override;

    virtual void setProxyObject(const OMEoid &pager, const OMEstring &method) override
    {
        OME_DELETE_OBJECT(pagerObject);
        pagerObject = new OMEoid(pager);
        encapsulationMethodName = method;
    }

    virtual const OMEoid *getProxyObject(const OMEstring **method) const override
    {
        *method = &encapsulationMethodName;
        return (pagerObject);
    }

    virtual void getObjectOID(OMEoidID *result) const override
    {
        *result = oidData;
    }

    virtual const class OMEobject *getInstanceInMemory() const override
    {
        return (nullptr);
    }

    bool operator==(const OMEoidStorage &arg) const
    {
        if (arg.referenceType != OMEoidStorage::external) {
            return (false);
        }
        const OMEoidStorageExternal *p = dynamic_cast<OMEoidStorageExternal *>(arg.storageRef);
        int rc = memcmp(&oidData, &(p->oidData), sizeof(oidData));
        return (rc == 0);
    }

    bool operator!=(const OMEoidStorage &arg) const
    {
        if (arg.referenceType != OMEoidStorage::external) {
            return (true);
        }
        const OMEoidStorageExternal *p = dynamic_cast<OMEoidStorageExternal *>(arg.storageRef);
        int rc = memcmp(&oidData, &(p->oidData), sizeof(oidData));
        return (rc != 0);
    }


}; // end class OMEoidStorageExternal

inline OMEoidStorage::OMEoidStorage(class OMEoidStorageExternal *ptr)
{
    referenceType = external;
    storageRef = ptr;
    addLink();
}

/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
